 (* Copyright (c) ENS de Lyon and Inria. All rights reserved. *)

Require Import Reals  Psatz.
From Flocq Require Import Core Relative Sterbenz Operations Mult_error  Plus_error.
Require Import mathcomp.ssreflect.ssreflect.
From Flocq Require Import Pff2Flocq.

(* Require Import F2SumFLT (* TwoSum_overflow *) . *)
From Flocq Require Import Div_sqrt_error Sqrt.
Require Import Bayleyaux F2Sum DWPlus.

Open Scope R_scope.



Set Implicit Arguments.

Section Prelim.
Local Notation two := radix2 (only parsing).
Notation pow e := (bpow two e).

Fact pow1 : pow 1 = 2 . by []. Qed.
Fact pow2 : pow 2 = 4 . by []. Qed.
Fact pow0: pow 0 = 1 . by []. Qed.
Fact powm1: pow (-1) = /2 .  by []. Qed.
Fact pow3 : pow 3 = 8 . by []. Qed.
Fact pow4 : pow 4 = 16 . by []. Qed.

Variable p emin emax: Z.
Context { prec_gt_0_ : Prec_gt_0 p }.
Notation u := (pow (-p)).

Variable choice : Z -> bool.
Hypothesis choice_sym: forall x, 
        choice x  = negb (choice (- (x + 1))).
Notation Emin:=  (emin - p + 1)%Z.
Notation eta:= (Rpower 2 (IZR (emin + p)/2)).

(***********************   FLX  ************************************)
Notation fexp := (FLX_exp p).

Notation format := (generic_format two fexp ).
Notation RN :=(round two  fexp (Znearest choice)).
Hypothesis pGt1: (1 < p)%Z.
Hypothesis emin_neg: (emin <0)%Z.

Notation ulp := (ulp two fexp).

Lemma rel_error_bound_flx  x :
           Rabs ((RN x -x)/x)<= u/(1+u).
Proof.
have upos: 0 < u by rewrite /u ; apply/bpow_gt_0.
case :(Req_dec x 0)=>[-> |h0].
  rewrite round_0 Rminus_0_r {1}/Rdiv Rmult_0_l Rabs_R0.
  move:(u_rod1pu_ro_pos two p); rewrite /u_ro bpow_plus pow1.
  field_simplify (/ 2 * (u * 2) / (1 + / 2 * (u * 2))); try lra.
    by   rewrite (Rplus_comm u);  lra.

rewrite Rabs_mult Rabs_Rinv //.
 apply/(Rmult_le_reg_r (Rabs x)).
  by apply/Rabs_pos_lt.
rewrite Rmult_assoc Rinv_l ?Rmult_1_r; last by apply/Rabs_no_R0.
move:(relative_error_N_FLX' two p  prec_gt_0_ choice x).
rewrite /u_ro.
suff ->:  / 2 * pow (- p + 1) = u by [].
by rewrite bpow_plus /u pow1 ; field.
Qed.



(*Prop   1.1*)
Lemma rel_err_flx_ulp t (t0: t <> 0):
Rabs (RN t  -t) <= (Rabs ((RN t -t)/t))/u * ulp t.
Proof.
have upos : 0 < u    by rewrite /u; apply/bpow_gt_0.
rewrite Rabs_mult -[X in X <= _]Rmult_1_r.
rewrite !Rmult_assoc; apply/Rmult_le_compat_l; first by apply/Rabs_pos.
apply/(Rmult_le_reg_l (Rabs t)); first by apply/Rabs_pos_lt.
rewrite Rmult_1_r -Rmult_assoc -Rabs_mult Rinv_r //(Rabs_pos_eq 1);
  last by lra.
rewrite Rmult_1_l; apply/(Rmult_le_reg_r u)=>//.
 have -> :  / u * ulp t * u = ulp t by field ; lra.
 by apply/ulp_FLX_ge.
Qed.

Lemma ex_mul_flx z k : (pow k)  <= Rabs z  -> format z ->  
   exists (mz:Z), z = IZR mz * (2 * (pow (- p))  * (pow k)).
Proof.
move=>  hle Fz.
case:(ex_shift two fexp z (1 -p + k)) =>//.
  rewrite /cexp /fexp.
  suff:( 1 + k <= mag radix2 z)%Z by lia.
  by apply:mag_ge_bpow; ring_simplify (1 + k -1)%Z.
move=>mz zE.
by exists mz; rewrite zE !bpow_plus.
Qed.



Lemma square_error_FLX  x (Fx: generic_format two (FLX_exp p) x)
             (xge_eta: eta <= x): generic_format two  (FLX_exp p)
    (round two  (FLX_exp p) (Znearest choice) (Rsqr x) - (Rsqr x)).
Proof.
have x2pos: 0 < Rsqr x.
  apply/Rlt_0_sqr.
  suff :0 <x by lra.
  apply/(Rlt_le_trans _ eta)=>//.
  by apply/exp_pos.
by rewrite /Rsqr; apply/mult_error_FLX.
Qed.


Notation format_flx:= (generic_format two (FLX_exp p)).



(* Lemma 1.5 *)
Lemma ro1_exact_flx  x (Fx: format x): 
   format (x - (Rsqr (RN (sqrt x)))).
Proof.
 by apply/ sqrt_error_FLX_N.
Qed.
End Prelim.

Section Main.
Notation two  := radix2.
Notation pow e := (bpow two e).


Variable p: Z.
Context { prec_gt_0_ : Prec_gt_0 p }.
Variable choice : Z -> bool.
Hypothesis choice_sym: forall x, choice x  = negb (choice (- (x + 1))).



(***********************   FLX ************************************)
Notation fexp := (FLX_exp  p).
Notation format := (generic_format two fexp ).
Notation RN :=(round two  fexp (Znearest choice) ).

Hypothesis pGt1: (1 < p)%Z.


Let u := pow (-p).

Local Notation ulp := (ulp two fexp).




Lemma upos: 0 < u.
Proof.
move: (bpow_gt_0 two (-p)); rewrite /u; lra.
Qed.

(* Double-word arthmetic *)


Notation double_word xh xl :=  (((format xh) /\ (format xl)) /\ 
                                 ( xh = RN (xh + xl))%R).




Lemma RN_sym x: RN (-x) = - (RN  (x)).
Proof.
suff : - RN (- x) =  RN x by lra.
by rewrite round_N_opp Ropp_involutive /round  /Znearest -choice_sym.
Qed.

Lemma RN_abs x: RN (Rabs x) = Rabs (RN x).
Proof.
case:(Rlt_le_dec x 0)=> x0.
  have: RN x <= 0.
    have -> : 0 = RN 0 by rewrite round_0.
    apply:round_le; first lra.
  by rewrite Rabs_left // => h ; rewrite Rabs_left1 // RN_sym.
move/(round_le radix2 fexp (Znearest choice)):(x0); rewrite round_0=> h.
by rewrite !Rabs_pos_eq.
Qed.



Fact diff_max x y (xpos: 0 <= x) (ypos : 0 <= y) : Rabs (x -y) <= Rmax x y.
Proof.
case:(Rle_lt_dec x y)=> xy.
  have ->: (x -y) = -(y -x) by ring.
  rewrite Rabs_Ropp Rabs_pos_eq ?Rmax_right; lra.
by rewrite  Rabs_pos_eq ?Rmax_left; lra.
Qed.

Notation TwoSum := (TwoSum p choice).

Notation TwoSum_err := (TwoSum_err p choice).
Notation TwoSum_sum := (TwoSum_sum p choice).



Lemma dw_ulpx xh xl: double_word xh xl -> Rabs xl <= /2 * ulp (xh +xl).
Proof.
case=>[[Fxh Fxl] xhe].
have {1} -> : xl = - (RN (xh + xl)  - (xh + xl)) by rewrite -xhe; ring.
rewrite Rabs_Ropp;
  by apply: error_le_half_ulp.
Qed.

Fact Rabs_0 x: x = 0 -> Rabs x = 0.
Proof. by move ->; rewrite Rabs_R0. Qed.

Lemma error_le_half_ulp' (t:R) (k:Z): 
        ((Rabs t) <= pow k)%R -> 
                       (Rabs ((RN t) - t)) <= /2 * u  * (pow k).
Proof.
have rge0: 0 <=  / 2 * u * pow k.
  by rewrite  /u Rmult_assoc -bpow_plus; move:(bpow_ge_0 two (- p + k)); lra.
case:(Req_dec t 0)=>[-> |t0].
  by rewrite round_0 Rminus_0_r Rabs_R0. 
case=>Rteq.
  apply: (Rle_trans _  (/2 * Ulp.ulp two fexp t)); 
    first exact:error_le_half_ulp.
  rewrite Rmult_assoc; apply:Rmult_le_compat_l; first lra.
  rewrite /u ulp_neq_0 // /cexp /fexp Rmult_comm.  
  rewrite bpow_plus.
  apply:Rmult_le_compat_r; first exact: bpow_ge_0.
  apply/bpow_le.
  by apply/mag_le_bpow=>//.
rewrite round_generic.
  by ring_simplify ( t - t); rewrite Rabs_R0 //.
move:Rteq; rewrite -(Rabs_pos_eq (pow k)); last by apply/bpow_ge_0.


case/Rabs_eq_Rabs => ->; try apply/generic_format_opp;

apply/format_bpow; lia.
Qed.

Lemma sterbenz' x y (Fx: format x) (Fy : format y):
  ( x/2 <= y <= 2*x)%R  -> format  (x - y).
Proof.
move=> hxy; replace (x - y)%R with (- (y - x))%R by ring.
by apply:generic_format_opp; apply:sterbenz.
Qed.


Lemma scale_12  x  (xpos:  0 < x):(exists e, 1 <= x * pow e < 2 ).
Proof.
have xn0: x <> 0 by lra.
case: (mag two x)=>ex /(_ xn0).
rewrite  Rabs_pos_eq;( try lra) => [[hex1 hex2]].
exists ( 1 - ex)%Z.
split.
  apply:(Rmult_le_reg_r (pow (ex -1))); first by apply: bpow_gt_0.
  rewrite Rmult_assoc -bpow_plus.
  ring_simplify (1 - ex + (ex - 1))%Z.
  have ->: pow 0 = 1 by [].
  lra.
apply:(Rmult_lt_reg_r (pow (ex -1))); first by apply: bpow_gt_0.
rewrite Rmult_assoc -bpow_plus.
ring_simplify (1 - ex + (ex - 1))%Z.
rewrite bpow_plus.
change (x * 1 < 2 * ((pow ex) * (/2))); lra.
Qed.

Section DWPlus.

Section DWPlus_pre.
Variables xh xl y: R.
Notation  x := (xh + xl)%R.

Notation  sh := (TwoSum_sum  xh y).
Notation  sl :=  (TwoSum_err xh y).
Notation  v := (RN (xl + sl)).

Hypothesis Fxh: format xh.
Hypothesis Fxl : format xl.
Hypothesis Fy : format y.
Hypothesis y0 : 0 <= y .
Hypothesis xl_ub : Rabs xl <= / 2 * ulp xh.
Hypothesis xhy : y <= xh.
Hypothesis xl0 : xl <> 0 .



Property  xl_ub': -/2 * ulp xh <= xl <= /2 * ulp xh.
Proof.  by move/Rabs_le_inv: xl_ub;lra. Qed.

 

Property xy_ge0: 0 <= xh + xl + y.
Proof.
have  xl_ub' :=  xl_ub'.
suff : 0 <= xh + xl by lra.
apply/(Rle_trans _ (- (/ 2 * ulp xh) + xh));last lra.
suff: (/ 2 * ulp xh) <=  xh by lra.
apply/(Rle_trans _ (ulp xh)).
  suff: 0 <= ulp xh by lra.
  by apply/ulp_ge_0.
set uxh := ulp xh; rewrite -(Rabs_pos_eq xh); try lra.
case:(Req_dec xh 0)=> x0.
rewrite /uxh x0 ulp_FLX_0 Rabs_R0; lra.
by apply/ulp_le_abs.  
Qed.

Property xpos : y<> 0 ->  0 < xh + xl + y.
Proof.
move=> yn0.
have  xl_ub' :=  xl_ub'.
apply/(Rlt_le_trans _ (- (/ 2 * ulp xh) + xh));last lra.
suff: (/ 2 * ulp xh) <  xh by lra.
apply/(Rlt_le_trans _ (ulp xh)).
  suff: 0 < ulp xh by lra.
  rewrite ulp_neq_0; try lra.
  by apply/bpow_gt_0.
set uxh := ulp xh; rewrite -(Rabs_pos_eq xh); try lra.
by  apply/ulp_le_abs =>// ; lra.
Qed.

Property xipos: y <> 0 ->   0 < / (xh + xl + y).
Proof. move=> yn0; by apply/Rinv_0_lt_compat/xpos. Qed.



Fact v_le_sh: Rabs v <= Rabs sh.
Proof.
have Hp1:= pGt1.
rewrite -!RN_abs//;apply/round_le.
apply :(Rle_trans _ _  _  (Rabs_triang _ _)); try lra.
rewrite( Rabs_pos_eq (xh + y)); last lra.
apply/(Rle_trans _ xh); try lra.
apply/(Rle_trans _ (3 * /2 * ulp xh)).
  suff: Rabs sl <= ulp xh by lra.
  rewrite TwoSum_correct  //.
  rewrite - (Rabs_Ropp).
  have ->: (- (xh + y - sh )) =  sh   - (xh + y) by ring.
  rewrite TwoSum_sumE.
  apply/(Rle_trans _ (/2 * ulp (xh + y))).
    by apply/error_le_half_ulp.
  apply/(Rle_trans _ (/2 * ulp (xh + xh))).
    apply/Rmult_le_compat_l; first lra.
    by apply/ulp_le; rewrite !Rabs_pos_eq;lra.
  have ->: xh + xh = xh * 2 by ring.
  rewrite -pow1 ulp_FLX_exact_shift; try lra; try lia.
  by rewrite pow1; lra.
suff : ulp xh <= xh * 2/3 by lra.
apply/(Rle_trans _ ((Rabs xh) * pow ( 1 -p))).
apply/ulp_FLX_le.
  rewrite Rabs_pos_eq ; last lra.
rewrite bpow_plus pow1.
rewrite /Rdiv Rmult_assoc.
apply/Rmult_le_compat_l; try  lra.
apply/Rmult_le_compat_l; first  lra.
apply/(Rle_trans _ (/4)); last lra.
change (pow (-p) <= pow (-2)); apply/bpow_le; lia.
Qed.


Fact errorDWFPE_W: Rabs (errorDWFP  p choice xh xl y)= Rabs (v - (xl + sl)).
Proof.
have Hp1 := pGt1.
have v_le_sh := v_le_sh.

rewrite /errorDWFP F2Sum_correct_abs //;
       try apply/generic_format_round; try apply/RN_sym.
have->: sh =xh + y - sl by  rewrite TwoSum_correct //;  ring.
have ->: v -(xl + sl) = - (sl + xl  -v) by ring.
rewrite Rabs_Ropp;congr Rabs;ring.
Qed.

(* Th 3.1 *)


Lemma DWPlusFP_rele_bound_W:   (relative_errorDWFP p choice xh xl y) <= u ^ 2.
Proof.
have upos :=  upos.
case:(Req_dec y 0) => [ye0|yne0].
have F0: format 0 by apply/generic_format_0.
  rewrite ye0 /relative_errorDWFP TwoSum_correct // TwoSum_sumE !Rplus_0_r.
have->: RN xh = xh by rewrite round_generic.
  ring_simplify (xl + (xh - xh)).
have->: RN xl = xl by rewrite round_generic.
rewrite Rabs_mult Rabs0 ?Rmult_0_l; first  nra.
rewrite F2Sum_correct_abs =>//; first ring.
exact RN_sym.
apply/(Rle_trans _ (ulp xh)).
suff : 0 <= ulp xh by lra.
apply/ulp_ge_0.
case:(Req_dec xh 0)=> xh0.
rewrite xh0 ulp_FLX_0 Rabs_R0; lra.
  by apply/ulp_le_abs.
have xhn0: xh <>  0 by lra.
case:(mag two xh ) => magxh /(_ xhn0).
set k := (magxh -1)%Z.
rewrite Rabs_pos_eq; last lra.
have magxhE: (magxh = k +1)%Z by lia.
move=> xh_b.
  have xhne0: 0 <> xh by lra.
have kE: ( magxh= mag two xh)%Z by rewrite (mag_unique_pos _ _  magxh) // -/k. 
rewrite magxhE in xh_b.
(* have kub: (emin +1 <= k)%Z. *)
(*   move:(alpha_minFP Fxh xhne0 ); rewrite Rabs_pos_eq ; last lra. *)
(*   move/(mag_le two);move /(_ (bpow_gt_0 two Emin)). *)
(*   rewrite mag_bpow. *)
(*   move: hmagxh; by lia. *)
have Hp1 := pGt1.
have u2pos : 0 < u^2 by nra.
have ult1: u < 1  by change (pow (-p) < pow 0); apply/bpow_lt ;lia.
(* have emin_neg : (emin < 0)%Z by  lia. *)
have uxh:  ulp xh = 2* u * pow k.
  rewrite ulp_neq_0; last lra.
  rewrite /cexp /fexp (mag_unique_pos _ _  (k+1)).
    rewrite !bpow_plus pow1 /u; ring.
  by have->: (k + 1 -1 = k)%Z by ring.
have xh_ub : xh <= 2* (pow k) * (1 - u).
  have ->: 2 * pow k *( 1 - u) =  pred two fexp (pow (k + 1)).
    rewrite pred_bpow /fexp.  
    rewrite (bpow_plus _ _ (-p)) -/u (bpow_plus _ k) pow1; ring.
  apply/pred_ge_gt; rewrite ?pow1 //; try lra.
  apply/generic_format_bpow; rewrite /fexp; lia.
move/Rabs_le_inv :xl_ub; rewrite uxh -!Rmult_assoc Rinv_l ?Rmult_1_l; try lra;
    move=> xl_u.
have x_lbub : (1 -u)* pow k <= x <= (2 - u)* pow k by lra.
have xy_lbub: (1 -u)* pow k < x + y <= (4 - 3*u) * pow k by lra.
pose e := v - (xl + sl).


have Fpk: format (pow k)  by apply/format_bpow.
have succkpE:  succ two fexp (pow k)  = (1 + 2*u)* pow k.
  rewrite succ_eq_pos ?ulp_neq_0; try lra.
  rewrite /cexp /fexp (mag_unique_pos _  _ (k + 1)); last first.
    have -> : (k+1 -1 =k)%Z by ring.
    rewrite bpow_plus pow1 ; lra.
  rewrite !bpow_plus pow1 /u; ring.

have xpos:= xpos.
have xipos:= xipos.
have RN_sym:= RN_sym.

rewrite   /relative_errorDWFP  Rabs_mult F2Sum_correct_abs //;
  try apply/generic_format_round; last by  apply/v_le_sh.

have->: sh =xh + y - sl  by rewrite TwoSum_correct //;ring.

(* have->:  (- sl + v - xl)  = v - (xl + sl) by ring. *)
rewrite (Rabs_pos_eq (/_)); last lra.
(* case :(Rle_lt_dec  (Rabs (xl + sl)) (pow (emin + 1)))=> Rs_ub. *)
(*   rewrite Rabs_0 ?Rmult_0_l; first lra. *)
(*   rewrite round_generic; first ring. *)
(*   by apply/Hauser=>//; first by apply/generic_format_round. *)
apply/(Rmult_le_reg_r (x + y)); first lra.
rewrite Rmult_assoc Rinv_l  ?Rmult_1_r; last lra.
case :(Rlt_le_dec ( x + y) (pow k)) => hxy1.
(* x + y < pow k *)
  have xlu0: -u * pow k  <= xl < 0 by lra.
  have y0xl: 0 < y < Rabs xl.
    rewrite -Rabs_Ropp Rabs_pos_eq;  lra.
  have : pow k <= xh by lra.
  case=> xh1.
    have:  succ two fexp (pow k) <= xh.
      by apply/succ_le_lt.
    rewrite succkpE.
    lra.
  have rnd_floor_1y: round two fexp Zfloor (pow k + y) = pow k.
    apply/round_DN_eq=>//; rewrite succkpE;lra.
  have shkp : sh = pow k.
    rewrite TwoSum_sumE -xh1 round_N_eq_DN // round_UP_DN_ulp.
      rewrite  rnd_floor_1y.
      suff ->: ulp (pow k +y) = 2 * u * (pow k) by lra.
      rewrite ulp_neq_0; last lra.
      rewrite /cexp /fexp (mag_unique_pos  _ _ (k + 1)).
        by rewrite 2!bpow_plus -/u pow1; lra.
      have ->: (k +1-1 = k )%Z by ring.
      by rewrite bpow_plus pow1; lra.
    apply/(generic_format_discrete _ _ _ (2^( p -1))); rewrite /F2R /=.
    have ->: (cexp two fexp (pow k  + y) = k+1-p)%Z .
      rewrite /cexp /fexp (mag_unique_pos  _ _  (k +1)) //.
      have ->: (k+ 1-1 = k )%Z by ring.
      rewrite bpow_plus pow1; lra.
    rewrite plus_IZR (IZR_Zpower two);last lia.
    rewrite -bpow_plus.
    ring_simplify (p - 1 + (k+ 1 - p))%Z.
    ring_simplify((pow (p - 1) + 1) * pow (k+ 1 - p)).
    rewrite -bpow_plus.
    ring_simplify (p - 1 + (k+ 1 - p))%Z; rewrite (bpow_plus _ _ (-p)) -/u.
    rewrite bpow_plus pow1; lra.
  have slE: sl = y by rewrite TwoSum_correct //; lra.
  have : -(u*pow k) <= xl + sl <= u*pow k by lra.
  move/Rabs_le=> xlsl_ub.
  have eu2: Rabs e <= /2 * u^2 * pow k.
    rewrite /e.
    have ->: / 2 * u ^ 2 * pow k= /2 * pow (-p) * pow (k-p).
      by rewrite (bpow_plus _ k)  -/u; ring.
    have haux :  Rabs (xl + sl) <= pow (k - p).
      by rewrite bpow_plus -/u ; lra.
    by apply/error_le_half_ulp'=>//.
  apply/(Rle_trans _ (/ 2 * u ^ 2 * pow k))=>//.
    apply/(Rle_trans _ (Rabs e))=>//.
    apply/Req_le; rewrite  /e.
    by congr Rabs; lra.
  rewrite (Rmult_comm (/2)) Rmult_assoc; apply/Rmult_le_compat;  lra.
(*cas pow k <= x + y*)
have F2: format 2.
  by change (format (pow 1)); apply/format_bpow.
case (Rle_lt_dec (x+y) (2 * (pow k)))=> xy2.
(* x + y <= 2 * pow k *)
  have xhy2: xh + y <= (2 + u) * (pow k)  by lra.
  have : pow k  <=  sh <= 2 * (pow k) .
    split.
      have ->:  pow k = RN (pow k) by rewrite round_generic //.
      apply/round_le;lra.
    have : sh <=  RN ((2 + u) * (pow k)) by apply/round_le.
    suff ->: RN ((2 + u) * pow k) = 2 * pow k by [].
    rewrite round_bpow.
      (* rewrite (mag_unique_pos _ _ 2); last first. *)
      (*   have -> : (2 -1 = 1)%Z by ring. *)
      (*   rewrite pow1; have ->: pow 2 = 4 by []. *)
      (*   lra. *)
      (* apply/Z.min_glb; lia. *)
    apply/Rmult_eq_compat_r.
    have succ2E:  succ two fexp 2 = 2 + 4*u.
      rewrite succ_eq_pos ?ulp_neq_0; try lra.
      rewrite /cexp /fexp (mag_unique_pos _ _ 2); last first.
        have -> : (2 -1 = 1)%Z by ring.
        change (2 <= 2 <4); lra.
      by rewrite bpow_plus -/u; have->: pow 2 = 4 by [].
    have haux: 2 < 2 + u < succ two fexp 2 by lra.
    have RDN2pu: round two  fexp Zfloor (2 + u) = 2.
      apply/round_DN_eq=>//;lra.
    rewrite round_N_eq_DN //.
    rewrite round_UP_DN_ulp.
      rewrite    RDN2pu.
      suff ->: ulp (2+u) = 4 * u by lra.
      rewrite ulp_neq_0; last lra.
      rewrite /cexp /fexp (mag_unique_pos  _ _ 2).
        by rewrite bpow_plus /u.
      have ->: (2-1 = 1 )%Z by ring.
      change(2 <= 2+u  < 4); lra.
    apply/(generic_format_discrete _ _ _ (2^(p -1))); rewrite /F2R /=.
    rewrite /cexp /fexp (mag_unique_pos _ _ 2).
      rewrite plus_IZR (IZR_Zpower two); last  lia.
      rewrite Rmult_plus_distr_r Rmult_1_l -!bpow_plus.
      ring_simplify (p - 1 + (2 - p))%Z.
      rewrite bpow_plus.
      have ->: pow 1 = 2 by [].
      have ->: pow 2 = 4 by [].
      rewrite -/u; lra.
    have -> : (2 -1 = 1) %Z by ring.
    change( 2 <= 2 + u < 4) ; by lra.
  move=> sh12.
  have:  -(u*pow k) <= sl <= u * pow k.
    rewrite TwoSum_correct //.
    case (Rle_lt_dec (2*pow k) (xh + y))=> xhy2'.
      have sh2 : sh = 2* pow k.
        apply/Rle_antisym; first lra.
        have ->: 2 * pow k = RN (2 *pow k).
          rewrite round_bpow.
            (* rewrite (mag_unique_pos _ _ 2); last first. *)
            (*   have -> : (2 -1 = 1)%Z by ring. *)
            (*   rewrite pow1; have ->: pow 2 = 4 by []. *)
            (*   lra. *)
            (* apply/Z.min_glb; lia. *)  
          by rewrite round_generic.
        by apply/round_le.
      lra.
    apply/Rabs_le_inv.
    have->:  (xh + y - sh) = - ( sh - (xh + y)) by ring.
    rewrite Rabs_Ropp.
    have ->: u * pow k = /2 * u * pow (k + 1).
      rewrite bpow_plus pow1;lra.
    apply/error_le_half_ulp'.
    by rewrite Rabs_pos_eq 1?bpow_plus ?pow1; try lra.
  move/Rabs_le=>  slu.
  have v2u': Rabs (xl + sl)<= 2*u * pow k.
    apply :(Rle_trans _ _  _  (Rabs_triang _ _)); lra.
  have eu2: Rabs e <= u^2 * pow k.
    have ->: u^2 * pow k= /2 *u *pow (k + 1 -p).
      by rewrite (bpow_plus _ (k +1)) (bpow_plus _ _ 1) pow1 -/u; lra.
    (* case (Z_lt_le_dec (emin ) (k - p))=> kemin. *)
    have haux :  Rabs (xl + sl) <= pow (k +1 - p).
      by rewrite 2!bpow_plus -/u pow1;lra.
    by apply/error_le_half_ulp'=>//.
  apply/(Rle_trans _ (Rabs e))=>//.
    apply/Req_le; rewrite  /e.
    by congr Rabs; lra.
   apply/(Rle_trans _ ( u ^ 2 * pow k)) =>//.
  apply/Rmult_le_compat_l; lra.
(*  2 * pow k < x + y *) 
have xhyb: (2 -u) * pow k< xh + y<= (4 - 4* u)* pow k by  lra.
have : 2 * pow k<= sh .
  rewrite -pow1  -bpow_plus; apply/round_N_ge_midp.
    by apply/format_bpow.
  rewrite (pred_bpow _ _ (1 + k)) /fexp.  
  have ->: (pow (1 + k) + (pow (1 + k) - pow (1 + k - p))) = 
           (2 -u) * ( pow (1 + k)).
    by rewrite (bpow_plus _ (1+k)) -/u; ring.
  rewrite bpow_plus pow1; lra.

move=> sh2.
have sh4 : sh <= (4 -4*u)* pow k.
  suff -> : (4 -4 *u)*pow k  = RN ((4 - 4*u)*pow k)  by apply/round_le;lra.
  rewrite round_bpow.
  rewrite round_generic //.
  suff->: 4 -4*u = pred two fexp 4.
    apply/generic_format_pred.
    by change (format (pow 2));apply/format_bpow.
  by rewrite (pred_bpow _ _ 2) /fexp bpow_plus pow2 /u. 
  (* rewrite (mag_unique_pos _ _ 2);last first. *)
  (*   rewrite pow1 pow2; split; try lra. *)
  (*   suff : 2 *u <= 1 by lra. *)
  (*   rewrite -pow0  -pow1 /u -bpow_plus; apply/bpow_le;lia. *)
  (* apply/Z.min_glb; lia. *)  
have slub: Rabs sl <= 2*u*pow k.
  have ->: 2*u * pow k= /2 * u * (pow (k + 2)) by rewrite (bpow_plus _ k) pow2 ;lra.
  rewrite -Rabs_Ropp TwoSum_correct //  Ropp_minus_distr.
  apply/error_le_half_ulp'; try lia.
  by rewrite bpow_plus pow2 Rabs_pos_eq; lra.
have xlsl_ub: Rabs (xl +sl) <= 3 *u * pow k.
  by apply :(Rle_trans _ _  _  (Rabs_triang _ _)); lra.
have eub: Rabs e <= 2*u^2 * pow k.
  have ->: 2 * u^2 * pow k = /2 * u * pow (k + 2 -p).
    rewrite (bpow_plus _ (k + 2)) (bpow_plus _ k)  -/u pow2; field.
    (* case (Z_lt_le_dec (emin +1 ) (k - p +2))=> kemin. *)
  have haux :  Rabs (xl + sl) <= pow (k +2 - p).
   rewrite 2!bpow_plus -/u pow2; apply/(Rle_trans _ (3 * u * pow k)); lra.
  by apply/error_le_half_ulp'=>//.
apply/(Rle_trans _  (2 * u ^ 2 * pow k))=>//.
  apply/(Rle_trans _ (Rabs e))=>//.
  by apply/Req_le; rewrite /e; congr Rabs; lra.
rewrite Rmult_assoc Rmult_comm Rmult_assoc; apply/Rmult_le_compat_l; try lra.
Qed.

Hypothesis xh_lbub:  1 <= xh  <= 2 - 2 * u.

Lemma DWPlusFP_rele_bound: (0 < y)->  relative_errorDWFP p choice xh xl y  <= u ^ 2.
Proof.
move=> ypos.
have Hp1 := pGt1.
have RN_sym:= RN_sym.

have upos : 0 < u by rewrite /u; apply/bpow_gt_0.
have u2pos : 0 < u^2 by nra.
have u_ulp1: u =  / 2 * ulp xh.
  rewrite ulp_neq_0; last lra.
  rewrite /cexp /fexp (mag_unique_pos _ _ 1).
    by rewrite bpow_plus pow1 /u; lra.
  have->: (1-1 = 0 )%Z by ring.
  change(1 <= xh < 2); lra.
have uu1: u = /2 * ulp 1.
  rewrite ulp_neq_0 ; last lra.
  rewrite /cexp /fexp  (mag_unique_pos _ _ 1) ?Z.max_l; try lia.
       by rewrite bpow_plus; change(pow (-p) = /2 *( 2 * (pow (-p)))); field.
    have->: (1-1 = 0 )%Z by ring.
  change(1 <= 1 < 2); lra.
have F1 : format 1.
  change (format (pow  0)).
  by apply/format_bpow. 
have succ1E:  succ two fexp 1 = 1 + 2*u.
  rewrite succ_eq_pos ?ulp_neq_0; try lra.
  rewrite /cexp /fexp (mag_unique_pos _  1 1); last first.
    have -> : (1 -1 =0)%Z by ring.
    change (1 <= 1 <2); lra.
  by rewrite bpow_plus -/u; have->: pow 1 = 2 by [].
move: (xl_ub); rewrite -u_ulp1.
move/Rabs_le_inv=> xl_u.
have x_lbub : 1 -u <= x <= 2 - u by lra.
have xy_lbub: 1 -u < x + y <= 4 - 3*u by lra.
pose e := v - (xl + sl).
case :(Rlt_le_dec ( x + y) 1) => hxy1.
  have : 1 <= xh by lra.
  case=> xh1.
    have:  succ two fexp 1 <= xh.
      apply/succ_le_lt =>//.
    rewrite succ1E.
    have xlb: -u <= xl < 0 by lra.
    have hyxl : 0 < y < Rabs xl.
      rewrite Rabs_left; last lra.
      lra.
    lra.
  have xlu0: -u <= xl < 0 by lra.
  have y0xl: 0 < y < Rabs xl.
    rewrite -Rabs_Ropp Rabs_pos_eq;  lra.
  have rnd_floor_1y: round two fexp Zfloor (1 + y) = 1.
    apply/round_DN_eq=>//; rewrite succ1E;lra.
  have sh1 : sh = 1.
    rewrite TwoSum_sumE -xh1.
    rewrite round_N_eq_DN //.
    rewrite round_UP_DN_ulp.
      rewrite  rnd_floor_1y.
      suff ->: ulp (1+y) = 2 * u by lra.
      rewrite ulp_neq_0; last lra.
      rewrite /cexp /fexp (mag_unique_pos  _ _ 1).
        by rewrite bpow_plus /u.
      have ->: (1-1 = 0 )%Z by ring.
      change(1 <= 1+y  <2); lra.
    apply/(generic_format_discrete _ _ _ (2^(p -1))); rewrite /F2R /=.
    have ->: (cexp two fexp (1 + y) = 1-p)%Z .
    rewrite /cexp /fexp (mag_unique_pos  _ _  1) //.        
      have ->: (1-1 = 0 )%Z by ring.
      change(1 <= 1+y  <2); lra.
    rewrite plus_IZR (IZR_Zpower two); last lia.
    rewrite -bpow_plus.
    ring_simplify (p - 1 + (1 - p))%Z.
    ring_simplify((pow (p - 1) + 1) * pow (1 - p)).
    rewrite -bpow_plus.
    ring_simplify (p - 1 + (1 - p))%Z; rewrite bpow_plus.
    change(1 < 1+y  <1 + 2 * u).
    lra.  
  have slE: sl = y by rewrite TwoSum_correct //; lra.
  have : -u <= xl + sl <= u by lra.
  move/Rabs_le=> xlsl_ub.
  have: RN ( Rabs (xl + sl)) <= RN u by apply/round_le.
  rewrite RN_abs // (round_generic _ _ _ u); last first.
    by change (format (pow (-p))); apply/format_bpow. 
  move=> vub.
  have eu2: Rabs e <= /2 * u^2.
    rewrite /e.
    have ->: / 2 * u ^ 2 = /2 * pow (-p) * pow (-p) by rewrite /u; ring.
    apply/error_le_half_ulp'.
    by rewrite -/u.
  rewrite /relative_errorDWFP F2Sum_correct_abs //;
     try apply/generic_format_round =>//; first last.
    rewrite sh1 ( Rabs_pos_eq 1);  lra.
  set zh := fst _.
  have->:  ((zh + (sh + v - zh) - (x + y))) = e by rewrite /e; lra.
  apply/(Rle_trans _ ((u^2 /2) * (/(1 -u)))).
    rewrite Rabs_mult; apply/Rmult_le_compat;  try apply/Rabs_pos; first lra.
    have xyi : 0< /(x + y) by apply/Rinv_0_lt_compat; lra.
    rewrite Rabs_pos_eq; try apply/Rinv_le;lra.
  apply/(Rmult_le_reg_r (1 -u)); first lra.
  rewrite Rmult_assoc Rinv_l; last lra.
  rewrite Rmult_assoc; apply/Rmult_le_compat_l; lra.

(*cas 1 <= x + y*)
have F2: format 2 by change (format (pow 1)); apply/format_bpow.
case (Rle_lt_dec (x+y) 2)=> xy2.
  have xhy2: xh + y <= 2+u  by lra.
  have : 1 <=  sh <= 2.
    split.
      have ->:  1 = RN 1 by rewrite round_generic //.
      apply/round_le;lra.
    have : sh <= RN (2 +u) by apply/round_le.
    suff ->: (RN (2 + u) = 2) by [].
    have succ2E:  succ two fexp 2 = 2 + 4*u.
      rewrite succ_eq_pos ?ulp_neq_0; try lra.
      rewrite /cexp /fexp (mag_unique_pos _ _ 2); last first.
        have -> : (2 -1 = 1)%Z by ring.
        change (2 <= 2 <4); lra.
      by rewrite bpow_plus -/u; have->: pow 2 = 4 by [].
    have haux: 2 < 2 + u < succ two fexp 2 by lra.
    have RDN2pu: round two  fexp Zfloor (2 + u) = 2.
      apply/round_DN_eq=>//; lra.
    rewrite round_N_eq_DN //.
    rewrite round_UP_DN_ulp.
      rewrite    RDN2pu.
      suff ->: ulp (2+u) = 4 * u by lra.
      rewrite ulp_neq_0; last lra.
      rewrite /cexp /fexp (mag_unique_pos  _ _ 2).
        by rewrite bpow_plus /u.
      have ->: (2-1 = 1 )%Z by ring.
      change(2 <= 2+u  < 4); lra.
    apply/(generic_format_discrete _ _ _ (2^(p -1))); rewrite /F2R /=.
    rewrite /cexp /fexp (mag_unique_pos _ _ 2).
      rewrite plus_IZR (IZR_Zpower two); last  lia.
      rewrite Rmult_plus_distr_r Rmult_1_l -!bpow_plus.
      ring_simplify (p - 1 + (2 - p))%Z.
      rewrite bpow_plus.
      have ->: pow 1 = 2 by [].
      have ->: pow 2 = 4 by [].
      rewrite -/u; lra.
    have -> : (2 -1 = 1) %Z by ring.
    change( 2 <= 2 + u < 4) ; by lra.
  move=> sh12.
  have:  -u <= sl <= u.
  rewrite TwoSum_correct //.
    case (Rle_lt_dec 2 (xh + y))=> xhy2'.
      have sh2 : sh = 2.
        apply/Rle_antisym; try lra.
        have ->: 2 = RN 2 by rewrite round_generic.
        by apply/round_le.
      lra.
    apply/Rabs_le_inv.
    have->:  (xh + y - sh) = - ( sh - (xh + y)) by ring.
    rewrite Rabs_Ropp.
    have ->: u = /2 * u * pow 1.
      change (u = /2 * u * 2); lra.
    apply/error_le_half_ulp'.
    have->:  pow 1 = 2 by [].
    rewrite Rabs_pos_eq; lra.
  move/Rabs_le=>  slu.
  have v2u': Rabs (xl + sl)<= 2*u.
    apply :(Rle_trans _ _  _  (Rabs_triang _ _)); lra.
  have v2u: Rabs v <= 2*u.
    rewrite -RN_abs //.
    have ->: 2 * u = RN (2 * u).
      rewrite round_generic //.
      have -> : 2 *u = pow (1 -p) by rewrite bpow_plus.
      by apply/format_bpow.
    by apply/round_le.
  have eu2: Rabs e <= u^2.
    have ->: u^2 = /2 *u *pow (1 -p).
      rewrite bpow_plus.
      have->:  pow 1 = 2 by [].
      rewrite -/u; lra.
    rewrite /e; apply/error_le_half_ulp'.
    by rewrite bpow_plus -/u.
  rewrite /relative_errorDWFP F2Sum_correct_abs//;try apply/generic_format_round =>//.
    set zh := fst _.
    have->:  ((zh + (sh + v - zh) - (x + y))) = e.
      rewrite /e TwoSum_correct //; lra.
    rewrite Rabs_mult.
    have xyi : 0< /(x + y) by apply/Rinv_0_lt_compat; lra.
    rewrite (Rabs_pos_eq (/ (x + y))); try apply/Rinv_le; try lra.
    apply/(Rmult_le_reg_r (x + y)); try lra.
    rewrite Rmult_assoc Rinv_l; try lra.
    apply/Rmult_le_compat; try lra.
    by apply/Rabs_pos.
  rewrite (Rabs_pos_eq sh);  lra.
have xhyb: 2 -u < xh + y<= 4 - 4* u by  lra.
have : 2 <= sh .
  apply/round_N_ge_midp=>//.
  rewrite (pred_bpow _ _ 1) /fexp.  
  by rewrite bpow_plus;have ->: pow 1 = 2 by []; rewrite -/u; lra.
move=> sh2.
have sh4 : sh <= 4 -4*u.
  suff -> : 4 -4 *u = RN (4 - 4*u) by apply/round_le;lra.
rewrite round_generic //.
  suff->: 4 -4*u = pred two fexp 4.
    apply/generic_format_pred.
    apply/(generic_format_bpow _ _  2); rewrite /fexp ; lia.
  rewrite (pred_bpow _ _ 2).
  by rewrite bpow_plus -/u.
have pow2: pow 2 = 4 by [].
have slub: Rabs sl <= 2*u.
  have ->: 2*u= /2 * u * (pow 2) by rewrite pow2 ;lra.
  rewrite -Rabs_Ropp TwoSum_correct //  Ropp_minus_distr.
  apply/error_le_half_ulp'; try lia.
  by rewrite pow2 Rabs_pos_eq; lra.
have xlsl_ub: Rabs (xl +sl) <= 3 *u.
  by apply :(Rle_trans _ _  _  (Rabs_triang _ _)); lra.
have v_ub: Rabs v <= 3 *u.
  suff ->: 3 * u = RN (3 * u) by rewrite -RN_abs //; apply/round_le.
  rewrite round_generic //.
  have h: 3 * u =  F2R (Float two 3 (-p)) by rewrite /F2R.
  apply/(FLX_format_Rabs_Fnum _ h).
    rewrite /=.
    apply/(Rlt_le_trans _ (pow 2)).
      by  rewrite pow2 Rabs_pos_eq; lra.
    apply/bpow_le; lia.
have eub: Rabs e <= 2*u^2.
  have ->: 2 * u^2 = /2 * u * pow (2 -p).
    rewrite bpow_plus -/u pow2; field.
  rewrite /e; apply/error_le_half_ulp'.
  rewrite bpow_plus -/u pow2; lra.
rewrite /relative_errorDWFP F2Sum_correct_abs //;try apply/generic_format_round.
  set zh := fst _.
  have->:  ((zh + (sh + v - zh) - (x + y))) = e.
    rewrite /e TwoSum_correct //; lra.
  rewrite Rabs_mult.
  have xyi : 0< /(x + y) by apply/Rinv_0_lt_compat; lra.
  rewrite (Rabs_pos_eq (/ (x + y))); try apply/Rinv_le; last lra.
  apply/(Rmult_le_reg_r (x + y)); first  lra.
  rewrite Rmult_assoc Rinv_l; last lra.
  have ->: u^2 * (x +  y) = 2*u ^2 * (/2  * (x + y)) by lra.
  apply/Rmult_le_compat; try  lra.
  by apply/Rabs_pos.
rewrite (Rabs_pos_eq sh); lra.
Qed.




End DWPlus_pre.

(* Notation TwoSum_sum := (TwoSum_sum p emin choice). *)
(* Notation TwoSum_err := (TwoSum_err p emin choice). *)



(* Notation TwoSum_correct := (TwoSum_correct emax). *)

Fact DW_0 xh xl (DWx: double_word xh xl) : xh = 0 -> xl = 0.
Proof.
move=>xh0.
case:DWx =>[[Fxh Fxl] xhE].
rewrite -(Rplus_0_l xl) -{1}xh0.
have:  RN (xh + xl ) = 0 by rewrite -xh0.
by apply/(Plus_error.round_plus_eq_0 two fexp).
Qed.




Notation  sh xh y:= (TwoSum_sum xh y). 
Notation  sl xh y:=  (TwoSum_err xh y). 
Notation  v xh y xl:= (RN (xl + (sl xh y))). 

Notation  Hp1 := pGt1.
(* Notation choice_sym:= choice_sym. *)

(* Notation  emin_max_p:= emin_max_p. *)

Fact vC y xh xl (Fxh: format xh) (Fy: format y) : v  xh y xl = v  y xh xl.
Proof. 
have  Hp1 := pGt1.
by rewrite TwoSum_errC. Qed.

Fact  DWPlusFPC  xh xl y  (Fxh :format xh) (Fy: format y):
    (DWPlusFP  p choice   xh xl y) =(DWPlusFP p choice    y xl xh).
Proof.

   rewrite /DWPlusFP vC //.
have ->: (sh y xh) = sh xh y by rewrite TwoSum_sumE Rplus_comm.
by [].
Qed.


Fact relative_error_DWFPC  xh xl y  (Fxh : format xh)(Fy : format y):
  (relative_errorDWFP  p choice xh xl y) =  (relative_errorDWFP  p choice  y xl xh).
Proof.
have  Hp1 := pGt1.
rewrite /relative_errorDWFP.
have <-: (y + xl + xh) = (xh + xl + y) by ring.
by rewrite /TwoSum_sum TwoSumC // vC.
Qed.

Lemma round_plus_eps x y (Fx : format x) (xpos: 0 <= x ) ( ysmall : 0 < y < /2* ulp x):
  RN(x + y) = x.
Proof.
have ux0:  0<=  ulp x by apply ulp_ge_0.
have hZ: round two fexp Zfloor (x + y) = x.
  by rewrite round_DN_plus_eps_pos //; lra.
have hC: round two fexp Zceil(x + y) = x + ulp x.
  rewrite round_UP_plus_eps_pos //; lra.
by rewrite round_N_eq_DN //;lra.
Qed.


 Fact TwoSum_sumS  x y 
 (ypos : 0 <= y <= x) 
 (* (magxge: (emin + 1 <= mag two x)%Z) *) (* true as soon as (format (ulp x)*):
 TwoSum_sum (x * pow (1 -mag two x)) (y * pow (1 -mag two x))  = 
     (TwoSum_sum  x y   ) * pow (1 - mag two x).
Proof.
rewrite  !TwoSum_sumE -Rmult_plus_distr_r.
case: ypos=> yge0 ylex.
case: yge0=>[ypos| <-]; last first.
  rewrite Rplus_0_r round_bpow //.
rewrite round_bpow //.

Qed.


 Fact TwoSum_errS x y 
 (ypos : 0 <= y <= x) 
 (Fx : format x) (Fy : format y) 
:
 TwoSum_err (x * pow (1 -mag two x)) (y * pow (1 -mag two x))  = 
     (TwoSum_err  x y   ) * pow (1 - mag two x).
Proof.
have  Hp1 := pGt1.
rewrite !TwoSum_correct ?TwoSum_sumS //; first  ring.
  by apply/mult_bpow_exact_FLX.
by apply/mult_bpow_exact_FLX.
Qed.


 Fact TwoSum_errS' x y 
 (ypos : 0 <= y <= x) 

 (Fx : format x) (Fy : format y) 
 (Fxp: format (x * pow (1 - mag two x))) (Fyp: format (y * pow (1 -mag two x))):
 TwoSum_err (x * pow (1 -mag two x)) (y * pow (1 -mag two x))  = 
     (TwoSum_err  x y   ) * pow (1 - mag two x).
Proof.
have  Hp1 := pGt1.

by rewrite !TwoSum_correct ?TwoSum_sumS //;  ring.
Qed.


Definition  ex x := pow (1 - mag two x)%Z.



Lemma scale_generic_12'  x (Fx: format x) (xpos:  0 < x):
   (exists e, 1 <= x * pow e <= 2 - 2 * u /\ (format (x * pow e))).
Proof.
have xn0: x <> 0 by lra.  
case:(mag two x)=>ex /(_ xn0)=> hmag.
have hamg : (ex = mag two  x)%Z by rewrite (mag_unique two x ex).
exists (1 -ex)%Z.
split;[split|].
    apply/(Rmult_le_reg_r (pow(ex -1))); first by apply/bpow_gt_0.
    rewrite Rmult_assoc -bpow_plus; ring_simplify (1 - ex + (ex - 1))%Z.
    have ->: pow 0 =1 by []. rewrite -(Rabs_pos_eq x); lra.
  rewrite bpow_plus; have->: pow 1 = 2 by [].
  suff : x * pow (-ex) <= 1 -u by lra.
  have->: 1 - u = pred two fexp 1.
    have ->: 1 = pow 0 by [].
    rewrite pred_bpow /fexp.
      rewrite bpow_plus.
      have ->: pow 0 = 1 by [].
      by rewrite /u; ring.
      apply/pred_ge_gt.
  + apply/Mult_error.mult_bpow_exact_FLX=>//; lia.
  + change (format (pow 0)).
    by apply/format_bpow.  
   apply/(Rmult_lt_reg_r (pow (ex))); first by apply/bpow_gt_0.
   rewrite Rmult_assoc -bpow_plus; ring_simplify(-ex +ex)%Z.
   by change(x * 1 < 1 * pow ex); rewrite -(Rabs_pos_eq x);lra.
apply/Mult_error.mult_bpow_exact_FLX =>//; lia.
Qed.

Lemma scale_generic_12  x (Fx: format x) (xpos:  0 < x):
    1 <= x * pow (1 -mag two x) <= 2 - 2 * u .
Proof.
have xn0: x <> 0 by lra.  
set ex := mag two x.
case:(mag two x)=>ex0 /(_ xn0)=> hmag.
have magE : (ex0 = mag two  x)%Z by rewrite (mag_unique two x ex0).
rewrite /ex -magE.
split.
    apply/(Rmult_le_reg_r (pow(ex0 -1))); first by apply/bpow_gt_0.
    rewrite Rmult_assoc -bpow_plus; ring_simplify (1 - ex0 + (ex0 - 1))%Z.
    have ->: pow 0 =1 by []. rewrite -(Rabs_pos_eq x); lra.
  rewrite bpow_plus; have->: pow 1 = 2 by [].
  suff : x * pow (-ex0) <= 1 -u by lra.
  have->: 1 - u = pred two fexp 1.
    have ->: 1 = pow 0 by [].
    rewrite pred_bpow /fexp.
      rewrite bpow_plus.
      have ->: pow 0 = 1 by [].
      by rewrite /u; ring.
  apply/pred_ge_gt.
  + apply/Mult_error.mult_bpow_exact_FLX=>//; lia.
  + change (format (pow 0)).
    by apply/format_bpow; rewrite /fexp.
   apply/(Rmult_lt_reg_r (pow (ex0))); first by apply/bpow_gt_0.
   rewrite Rmult_assoc -bpow_plus; ring_simplify(-ex0 +ex0)%Z.
   by change(x * 1 < 1 * pow ex0); rewrite -(Rabs_pos_eq x);lra.
Qed.

Lemma Fsmag x (Fx : format x) : format (x * (pow (1 - mag two x))).
Proof.
apply/Mult_error.mult_bpow_exact_FLX =>//.
Qed.

Theorem DWPlusFP_ge_0 (xh xl y :R)
         (Fxh: format xh)
         (Fxl: format xl)
         (Fy : format y)
         (n_neg_xh: 0 <= xh) (n_neg_y : 0 <= y)
         (xl_ub: Rabs xl <= /2 * ulp (Rmax  xh y)):
         0 <= fst (DWPlusFP p choice xh xl y).
Proof.
wlog xhy: xh y Fxh Fy  n_neg_xh n_neg_y  xl_ub / y <= xh.
  move=> Hwlog.
  case:(Rle_lt_dec y xh ) => hyxh.
    by apply/Hwlog.
  rewrite DWPlusFPC //; apply/Hwlog =>//.
    by rewrite Rmax_comm.
  lra.
have->: 0 = RN 0 by rewrite round_0.
apply/round_le.

have : Rabs (v xh y xl )<= Rabs (sh xh y).
  apply/v_le_sh=>//.
  by move:   xl_ub; rewrite Rmax_left.
rewrite (Rabs_pos_eq  (sh xh y)).
move/Rabs_le_inv; lra.

have->: 0 = RN 0 by rewrite round_0.
apply/round_le; lra.
Qed.

(* Th 3.1 *)

Theorem  DWPlusFP_correct_aux ( xh xl y  : R)
         (Fxh: format xh)
         (Fxl: format xl)
         (Fy : format y)
         (n_neg_x: 0 <= xh) (n_neg_y : 0 <= y) 
         (xl_ub: Rabs xl <= /2 * ulp (Rmax  xh y)):
 relative_errorDWFP p choice xh xl y  <= u^2.
Proof.
have Hp1:= pGt1.
have RN_sym:= RN_sym.
have F0: format 0 by apply/generic_format_0.
have upos: 0 < u by apply/bpow_gt_0.
have u2pos: 0 < u^2 by nra.
case:n_neg_x => xh0; last first.
  rewrite -xh0  /relative_errorDWFP  TwoSum_correct  //  TwoSum_sumE Rplus_0_l.
  rewrite (round_generic _ _ _ y) // (Rminus_diag_eq y) // Rplus_0_r.
  rewrite (round_generic _ _ _ xl) //.
  case: n_neg_y =>y0 ; last first.
    rewrite -y0  Fast2Sum0f //= Rabs_mult Rabs_0; lra.
  rewrite  F2Sum_correct_abs //=.
    rewrite Rabs_mult Rabs_0; lra.
  apply/(Rle_trans _ (/ 2 * ulp y)).
    move:xl_ub; rewrite -xh0 Rmax_right;  lra.
  apply/(Rle_trans _  ( ulp y)).
    suff: 0 <= ulp y by lra.
    by apply/ulp_ge_0.
  apply/ulp_le_abs=>//;lra.
case: n_neg_y => y0; last first.
  rewrite -y0  /relative_errorDWFP  TwoSum_correct // TwoSum_sumE Rplus_0_r.
  rewrite (round_generic _ _ _ xh) // (Rminus_diag_eq xh) // Rplus_0_r .
  rewrite (round_generic _ _ _ xl) // F2Sum_correct_abs //= -?xhE.
    rewrite Rabs_mult Rabs_0;  last ring.
    by rewrite Rmult_0_l; lra.
  apply/(Rle_trans _  (/ 2 * ulp xh)).
    move: xl_ub  ; rewrite Rmax_left; lra.
  apply/(Rle_trans _ (ulp xh)).
    suff: 0 <= ulp xh by lra.
    by apply/ulp_ge_0.
  by apply/ulp_le_abs=>//; lra.
case:(Req_dec xl 0)=>xl0.
rewrite xl0 /relative_errorDWFP  Rplus_0_l Rplus_0_r  (round_generic _ _ _ (sl _ _ )); 
   last by apply/generic_format_round.
  rewrite F2Sum_correct_abs //=.
  + rewrite  Rabs_mult Rabs_0 ?Rmult_0L0; try lra.
    rewrite TwoSum_correct //; ring.
  +  by apply/generic_format_round.
  +  by apply/generic_format_round.
  rewrite TwoSum_correct //.
  have ->:  (xh + y - sh xh y) = - ((sh xh y ) - (xh + y)) by ring.
  rewrite Rabs_Ropp.
  apply/(Rle_trans _ (Rabs xh)).
    by apply/plus_error_le_l.
  rewrite !Rabs_pos_eq; try lra.
    have : xh <= xh + y by lra.
    move/(round_le two fexp (Znearest choice)).
    by rewrite (round_generic _ _ _ xh).
  have ->: 0 = RN 0 by rewrite round_0.
  by apply/round_le; lra.

(* xl <> 0 *)

wlog xhy: xh y Fxh Fy xl_ub xh0 y0 / y <= xh.
  move=>HWlog.
 case: (Rle_lt_dec y xh)=>yxh.
    by apply/HWlog.
  rewrite relative_error_DWFPC //.
  apply/(HWlog y xh) =>//.
    by rewrite Rmax_comm.
  by lra.
move:xl_ub; rewrite Rmax_left;last lra. 
  move=> xl_ub.
apply/DWPlusFP_rele_bound_W=>//; lra.
Qed.

Notation dw_ulp := (dw_ulp pGt1).

(* Th 3.1 general case *)

Theorem  DWPlusFP_correct ( xh xl y  : R) 
         (Fy : format y)
         (DWx: double_word xh xl) (n_neg_x: 0 <= xh) (n_neg_y : 0 <= y):
 relative_errorDWFP p choice xh xl y  <= u^2.
Proof.
case:(DWx); move  => [Fxh Fxl] xE.
apply/ DWPlusFP_correct_aux=>//.
case:(Rle_lt_dec xh y) =>xh_y.
rewrite Rmax_right //.
apply/(Rle_trans _ (/2 * ulp xh)).
    apply/(dw_ulp DWx)
.
  apply/Rmult_le_compat_l; first lra.
apply/ulp_le; rewrite !Rabs_pos_eq; lra.
rewrite Rmax_left; last lra.
by apply/(dw_ulp DWx).
Qed.

Lemma v_le_sh_W xh xl y (DWx: double_word xh xl) (Fy : format y)
             (xhpos : 0 <= xh) (ypos : 0 <= y):
   Rabs (v xh y xl)  <= Rabs (sh xh y).
Proof.
have  Hp1 := pGt1.
case:(DWx)=>[[Fxh Fxl] xE].
case:xhpos=>x0; last first.
  have xl0: xl = 0  by apply/(DW_0 DWx).
  rewrite  TwoSum_correct // -x0 xl0 TwoSum_sumE !Rplus_0_l
 
    (round_generic _ _ _ y) //.
  have->: y -y = 0 by ring.
  by rewrite round_0 Rabs_R0; apply/Rabs_pos.
rewrite TwoSum_correct //.
rewrite -!RN_abs //.
apply/round_le.
apply :(Rle_trans _ _  _  (Rabs_triang _ _)); try lra.
rewrite (Rabs_pos_eq (xh + y)); last lra.
apply/Rplus_le_compat.
  apply/(Rle_trans _ (/ 2 * ulp xh)).
    apply/(dw_ulp DWx) =>//.
  apply/(Rle_trans _ (ulp xh)).
    suff: 0 <= ulp xh by lra.
    by apply/ulp_ge_0.
  rewrite -[X in _ <= X]Rabs_pos_eq //; try lra.

  by apply/ulp_le_abs =>// ; lra.
rewrite  -Rabs_Ropp  Ropp_minus_distr -[X in _ <= X]Rabs_pos_eq; try lra.
by apply/plus_error_le_r.
Qed.



(* Prorp 3.2 *)



Theorem   DWPFP_bound_small_y ( xh xl y  : R) 
         (Fy : format y)
         (DWx: double_word xh xl) (n_neg_x: 0 <= xh):
          u <= /16 -> (- 2* u - u^2)* (xh + xl) <= y -> 
          relative_errorDWFP p choice xh xl y  <= u^2 + 3 * u^3.
Proof.
move=> u16 ysmall.
have upos: 0 < u by apply/bpow_gt_0.
have ult1: u < 1 by apply/bpow_lt_1; lia.
have hu:  0 < 1 - 2 * u - u ^ 2.
  suff: 4*u  <= 1 by nra.
  change (pow 2 * pow (-p) <= pow 0); rewrite -bpow_plus; apply/bpow_le; lia.
case:(Rle_lt_dec 0 y)=> [ypos |yneg].
  apply/(Rle_trans _ (u^2)).
    by apply/DWPlusFP_correct.
  nra.
case:  n_neg_x => [xpos|x0]; last first.
  rewrite rel_errorE null_xh // Rabs_R0 Rmult_0_l; nra.

wlog [xh1 xh2]: xl xh  y  DWx Fy  xpos  yneg ysmall  / 1 <= xh  <= 2-2*u.
  move=> Hwlog.
  case:(DWx)=>[[Fxh Fxl] xE].
  case: (scale_generic_12' Fxh xpos)=> exh Hxhs.
  have powgt0 := (bpow_gt_0 two exh). 
  case:(Req_dec (xh + xl + y) 0)=> xy0.
    have ye: y = -(xh + xl) by lra.
    case: (Rle_lt_dec (xh + xl ) 0) => xhxl0.
      have : RN (xh + xl) <=  RN 0 by apply/round_le.
      by case:DWx=> [_] <-; rewrite round_0; lra.
    have : (xh + xl) <= (2* u + u^2) * (xh + xl) by lra.
    rewrite -{1}(Rmult_1_l (xh + xl)).
    move/(Rmult_le_reg_r _ _ _ xhxl0); nra.
 
  rewrite -(relative_error_DWFPS _ _ exh Fy Fxh) //.
  case:(Hwlog (xl * pow exh) (xh * pow exh) (y * pow exh))=>//;
      try nra.
    exact:DW_S.
  exact: formatS.

pose v := v xh y xl.
pose sh := sh xh y.
pose sl := sl xh y.
pose e := v - (xl + sl).
pose x := xh + xl.
have dw_ulp:= (dw_ulp  DWx).
move: (dw_ulp); rewrite ulp_neq_0 /cexp /fexp ; last lra.
rewrite (mag_unique_pos two   _ 1) bpow_plus pow1 ?powm1; last lra.
have-> :  / 2 * (2 * pow (- p)) = u by rewrite /u; lra.
move/Rabs_le_inv => xlu.
have xub: x <= 2 - u by rewrite /x; lra.
have F1 : format 1.
   rewrite -pow0; apply/generic_format_bpow; rewrite /fexp; lia.
have pred1: pred two fexp 1 = 1 - u.
    rewrite -pow0 pred_bpow /fexp bpow_plus pow0 /u; lra.
have F1mu: format (1 - u) by rewrite -pred1; apply/generic_format_pred.
case:(DWx)=>[[Fxh Fxl] xE].
case : (Rle_lt_dec (1-u/2) x)=> xu; last first.
  have [xbl xbu]: pred two fexp 1 <= x < 1.
     by move: xu; rewrite pred1 /x; lra.
  case: xbl; last first.
    move=> h; suff : xh = 1 - u by lra.
    case: DWx => [_] ->.
    rewrite round_generic -/x; try  lra.
    by rewrite -h; apply/generic_format_pred.
  move=> xlb.
  have RNfx: round two fexp Zfloor x =  1 - u.
    apply/round_DN_eq.
      by rewrite -pred1; apply/generic_format_pred.
    rewrite -pred1 succ_pred //; lra.

  have RNcx:  round two fexp Zceil x =  1.
     apply/round_UP_eq =>//; lra.   
  suff : xh = 1 - u by lra.
  case: DWx => [_] ->.
 rewrite -/x round_N_eq_DN; lra.
have h12: (1 - 2*u - u^2) * (xh + xl) <= (xh + xl) + y  < 2 - u by lra.
have xypos : 0 < xh + xl + y .
suff: (0 < xh + xl) by nra.
case:(Rle_lt_dec (xh + xl) 0)=>// h.
suff : xh <= 0 by lra.
have ->: xh = RN (xh + xl ) by case: DWx.
have -> : 0 = RN 0 by rewrite round_0.
by apply/round_le.

case:(Rle_lt_dec 1 (x + y)); rewrite /x=> xy1.
  have xhyb: (1 - u) <=  xh + y < 2 - 2 * u by lra.
  have shb: (1 - u) <=  sh <= 2 - 2 * u.
    have -> : 1 - u = RN (1 - u) by rewrite round_generic.    
    have -> : (2 - 2 * u) = RN (2 - 2 * u).
      rewrite round_generic //; have ->:  (2 - 2 * u) =  ( 1 - u) * pow 1 by rewrite pow1; ring.
      by apply/formatS.
    split; apply/round_le; lra.
  have slub: Rabs sl <= u.
    rewrite /sl TwoSum_correct // -Ropp_minus_distr Rabs_Ropp.
    case: shb ; case; last first.
      move=> h _.
      have ush: ulp sh = u.
        rewrite ulp_neq_0 /cexp /fexp; last lra.
        rewrite (mag_unique_pos _  _ 0)  bpow_plus pow0 ?powm1 /u; lra.
      apply/(Rle_trans _ (/2 * u)); last lra.
      by rewrite -ush; apply/error_le_half_ulp_round.
    move=> shlb shub.
    have sh1: 1 <= sh .
      have -> : 1 = succ two fexp  (pred two fexp 1) by rewrite succ_pred .
      apply/succ_le_lt.
      + by apply/generic_format_pred.
      + by apply/generic_format_round.
      lra.
    have ush: ulp sh = 2 * u.
      rewrite ulp_neq_0 /cexp /fexp; last lra.
      by rewrite (mag_unique_pos _  _ 1)  bpow_plus pow1 ?powm1 /u; lra.
    have->: u = /2 * ulp sh by lra.
    by apply/error_le_half_ulp_round.
  have eub: Rabs e <= u^2.
    rewrite /e.
    have->: u^2 =  / 2 * u * pow (-p + 1) by rewrite bpow_plus -/u pow1; lra.
    apply/error_le_half_ulp'.
    apply :(Rle_trans _ _  _  (Rabs_triang _ _)); rewrite bpow_plus pow1 -/u -/sl.
    by move/Rabs_le: xlu; lra.
  rewrite rel_errorE errorDWFPe_abs=>//.
    rewrite  -Ropp_minus_distr Rabs_Ropp.
    have ->:  (DWPlus.v xh y xl - (DWPlus.sl xh y + xl)) = e by rewrite /e /v /sl; ring.
    apply/(Rle_trans _ ( u^2 * 1)); last nra.
    apply/Rmult_le_compat =>//; try apply/Rabs_pos.
    have h: 0 < /(xh + xl + y) by apply/Rinv_0_lt_compat; lra.
    rewrite Rabs_pos_eq; try lra.
    by rewrite -Rinv_1; apply/Rinv_le; lra.
(* v <= sh *)
  apply/(Rle_trans _ (1 - u)); last by move:shb; rewrite /sh => shb; rewrite  Rabs_pos_eq; try lra.
  rewrite -RN_abs.
  have ->: (1 -u ) = RN (1 -u) by rewrite round_generic.
  apply/round_le.
  apply/(Rle_trans _ (2 * u)).
    apply :(Rle_trans _ _  _  (Rabs_triang _ _)).
    move/Rabs_le: xlu; move: slub; rewrite /sl; lra.
  suff: 4* u <= 1 by lra.
  by change (pow 2 * pow (-p) <= pow 0); rewrite -bpow_plus; apply/bpow_le; lia.
(* x + y < 1 *)
case: (Rle_lt_dec 1 (xh + y))=> xhy1.
  have xl0: xl <0 by lra.
  have xy1mu: 1 - u <= xh + xl + y by lra.
  have xy1pu: xh + y < 1+u by lra.
  pose eps := (xh + y - 1).
  have sh1 : sh = 1.
    case: xhy1 => xhy1; last first.
    rewrite /sh TwoSum_sumE -xhy1 round_generic //.
    rewrite /sh TwoSum_sumE; have ->: xh + y = 1 + eps by rewrite /eps; ring.
    apply/round_plus_eps=>//; try lra.
    rewrite -u_ulp1 /eps -/u; lra.
  have sl0u: 0 <= sl < u by rewrite /sl TwoSum_correct // -/sh sh1; lra.
  have xlslub: Rabs (xl + sl) <= u by apply/Rabs_le; lra.
  have eub: Rabs e <= /2 * u^2.
    rewrite /e.
   have->: /2 * u^2 =  / 2 * u * pow (-p ) by rewrite -/u ; lra.
    apply/error_le_half_ulp'.
    by rewrite -/sl -/u.
  rewrite rel_errorE errorDWFPe_abs=>//.
    rewrite  -Ropp_minus_distr Rabs_Ropp.
    have ->:  (DWPlus.v xh y xl - (DWPlus.sl xh y + xl)) = e by rewrite /e /v /sl; ring.
    apply/(Rle_trans _ (/2 * u^2 /(1 -u))).

    apply/Rmult_le_compat =>//; try apply/Rabs_pos.
    have h: 0 < /(xh + xl + y) by apply/Rinv_0_lt_compat; lra.
    rewrite Rabs_pos_eq; try lra.
     by rewrite -Rinv_1; apply/Rinv_le; lra.
    apply/(Rmult_le_reg_r (1 -u)); first lra.
rewrite Rmult_assoc Rinv_l ?Rmult_1_r ; try lra.
field_simplify.
suff : 6 * u ^ 2 <=  1 by nra.
suff: 8 * u^2 <= 1 by lra.
rewrite /= Rmult_1_r.
have ->: 8 = pow 3 by [].
rewrite /u -pow0 -!bpow_plus; apply/bpow_le; lia.

apply/(Rle_trans _ u).
rewrite -RN_abs; have-> : u = RN u.
rewrite round_generic // /u ; apply/generic_format_bpow ;  rewrite /fexp; lia.
by apply/round_le; rewrite -/sl.
by rewrite -/sh sh1 Rabs_pos_eq ; lra.

(* xh + y < 1 *)
 have slu: Rabs sl <= /2* u.
  have->: /2 * u =  / 2 * u * pow 0 by rewrite pow0; ring.
rewrite /sl TwoSum_correct // -Ropp_minus_distr Rabs_Ropp; 
apply/error_le_half_ulp'.
  rewrite pow0 Rabs_pos_eq; try lra.
  suff :u <= xh + xl + y by lra.
  suff : u <= (1 - 2 * u - u ^ 2) * (xh + xl) by lra.
  apply/(Rle_trans _ ((1 - 2 * u - u ^ 2) * (1 - u / 2))); last first.
    apply/Rmult_le_compat_l=>//; lra.

    field_simplify.
    suff: 8*u <=  2 by nra.
    change (pow 3 * pow (-p) <= pow 1).
    rewrite -bpow_plus; apply/bpow_le; lia.

have xlsl: Rabs (xl + sl) <= 3/2*u.
    apply :(Rle_trans _ _  _  (Rabs_triang _ _)).

   move/Rabs_le: xlu; lra.
have eub: Rabs e <= u^2.
    rewrite /e.
   have->:  u^2 =  / 2 * u * pow (-p +1 ) by rewrite bpow_plus pow1 -/u; lra.
    apply/error_le_half_ulp'.
    rewrite -/sl -/u bpow_plus pow1 -/u; lra.

have : 1 - 5/2*u + /2*u^3 <= x + y.
  apply/(Rle_trans _ ((1 - 2*u -u^2)* (1 - u/2))).
    field_simplify; lra.
  rewrite /x; apply/(Rle_trans _  ((1 - 2 * u - u ^ 2) * (xh + xl))); try lra.
  apply/(Rmult_le_compat_l); try lra.
  by move: xu; rewrite /x.
rewrite /x=> xylb.

have den0: 0 < 1 - 5 / 2 * u + / 2 * u ^ 3 .
  suff : 4* u <= 1 by nra.
  change(pow 2 * pow (-p) <= pow 0); rewrite -bpow_plus ; apply/bpow_le; lia.
have den10: 0 < /(1 - 5 / 2 * u + / 2 * u ^ 3 ).
  by apply/Rinv_0_lt_compat.

rewrite rel_errorE errorDWFPe_abs=>//.
    rewrite  -Ropp_minus_distr Rabs_Ropp.
    have ->:  (DWPlus.v xh y xl - (DWPlus.sl xh y + xl)) = e by rewrite /e /v /sl; ring.
  apply/(Rle_trans _ (u^2 /(1 - 5 / 2 * u + / 2 * u ^ 3))).
    apply/Rmult_le_compat=>//; try apply/Rabs_pos.
    rewrite Rabs_pos_eq.
       by apply/Rinv_le.
    by apply/Rlt_le/Rinv_0_lt_compat.
apply/(Rmult_le_reg_r (1 - 5 / 2 * u + / 2 * u ^ 3))=>//.
rewrite Rmult_assoc Rinv_l ?Rmult_1_r; last lra.
field_simplify.

clear -upos u16 ;nra.

(* v <= sh *)
apply/(Rle_trans _ ( RN (2 * u))).
  rewrite -RN_abs -/sl.
  apply/round_le; lra.

  rewrite -RN_abs -/sh.

  apply/round_le.
suff  h : 2 * u <= xh + y .
 by rewrite Rabs_pos_eq; lra.


(* suff : 3*u <= xh + xl + y by lra. *)
(* suff: 3*u <= 1 - 5 / 2 * u + / 2 * u ^ 3 by lra. *)
suff: 6*u <= 2 - 5*u + u^3 by lra.
suff: u <= 2 by nra.
lra.
Qed.

(* Rem 3.3 *)

Lemma  errorDWFPE(xh xl y : R):
       format xh ->
       format y ->
       0 <= y ->
       Rabs xl <= / 2 * ulp xh ->
       0 <= xh ->
       Rabs (errorDWFP p choice xh xl y) =
       Rabs (v xh y xl - (xl + sl xh y)).
Proof.
wlog : y xh xl / y <= xh.
  move=> HWlog.
  case:(Rle_lt_dec y xh); first apply/HWlog.
  move=> *; rewrite error_DWFPC // vC //.
  have->: (sl xh y) = (sl y xh).
    by rewrite !TwoSum_correct// !TwoSum_sumE Rplus_comm. 
  apply/( HWlog xh y)=>//; try lra.
  apply/(Rle_trans _ (/ 2 * ulp xh))=>//.
  apply/Rmult_le_compat_l; first lra.
  apply/ulp_le; rewrite !Rabs_pos_eq; try lra.
by move=>*; apply/errorDWFPE_W.
Qed.

Section Lange_Rump.


Variable x  e: nat -> R.

Fixpoint  s (i:nat) := match i with 
| 0 => (x 0) * (1 + (e 0))
| S i => (x (S i) +  (s i))  * (1 + (e (S i)))
end.

Definition d i := match i with 
0 => (x 0 ) * (e 0)
| S i => (x (S i) +  (s i)) * (e (S i)) 
end.


Fact s0E : s 0 = x 0 + d 0 .
Proof. 
by rewrite /s /d ; ring.
Qed.

Fact siE i: s (S i) = x (S i) + s i + d (S i).
Proof.
have ->:  s (S i) =  (x (S i) +  (s i))  * (1 + (e (S i))) by [].
rewrite /d ; ring.
Qed.

Fact di0: forall i,  ((x (S i) ) +  (s i)) = 0 -> d (S i)  = 0.
Proof.
move=>i hi.
have ->: d (S i) =  (x (S i) +  (s i)) * (e (S i))  by [].
by rewrite hi; lra.
Qed.


Fact si0: forall i,  ((x (S i) ) +  (s i)) = 0 -> s (S i)  = 0.
Proof.
move=>i hi.
have ->:  s (S i) =  (x (S i) +  (s i))  * (1 + (e (S i))) by [].
by rewrite hi; lra.
Qed.

(* Definition e (i :nat) :=  match i with  *)
(* 0=> 0 *)
(* | (S j) => match (Req_EM_T ((x i ) +  (s j)) 0) with  *)
(*             left _ => 0  *)
(*             | _ => (d i)/((x i ) +  (s j)) end  *)
(* end. *)

Fact d0E : d 0 = (x 0 ) * (e 0).
by [].
Qed.


Fact  diE i : d (S i)  = (x ( S i)  + (s i))* (e (S i)).
Proof.
by [].
Qed.


Fact sE i:   s i = (sigma (fun i =>  (d i)) 0 i) +
                   (sigma (fun i =>  (x i)) 0 i).
Proof.
elim: i.
  rewrite /sigma  /sum_f_R0 /= ; ring.
move=> i Hi.
rewrite sigma_last /=; last lia.
rewrite (@sigma_last _ _ (S i)) /=;try lra; lia.
Qed.


Definition D i := (s i) - (sigma (fun i =>  (x i)) 0 i).


Fact DE i : D i =  (sigma (fun i =>  (d i)) 0 i).
Proof. by rewrite /D sE //; ring. Qed.

Notation Sx k := (sigma (fun i => Rabs (x i)) 0 k).
Notation Se k := (sigma (fun i => Rabs (e i)) 0 k).
Notation Sd k := (sigma (fun i => Rabs (d i)) 0 k).


Definition xi i := match i with
0 => Rabs (e 0)
|S i => (Rabs (d (S i))) / ( Sx (S i) + Sd i)
end.

Notation Sxi k := (sigma (fun i =>  (xi i)) 0 k).

(* Fact dxiE n: Rabs (d (S n)) = (xi (S n)) * (Sx (S n) +  ). *)



Lemma xi_den_ge0 i:  0 <= Sx (S i) + Sd i.
Proof.
by apply/Rplus_le_le_0_compat;
apply/cond_pos_sum=>j; apply/Rabs_pos.
Qed.

(* cond_pos_sum: *)
(*   forall (An : nat -> R) (N : nat), *)
(*   (forall n : nat, 0 <= An n) -> 0 <= sum_f_R0 An N *)

(* Rplus_le_lt_0_compat: forall r1 r2 : R, 0 <= r1 -> 0 < r2 -> 0 < r1 + r2 *)

(* Rplus_lt_le_0_compat: forall r1 r2 : R, 0 < r1 -> 0 <= r2 -> 0 < r1 + r2 *)

(* Rplus_le_le_0_compat: forall r1 r2 : R, 0 <= r1 -> 0 <= r2 -> 0 <= r1 + r2 *)

Fact sum_f_pos :  forall (An : nat -> R) (N : nat), (forall n : nat, (n <= N)%nat -> 0 <= An n) -> 0 <= sum_f_R0 An N.
Proof.
move=> An.
elim.
  by apply; lia.
move=> n Hn hSN.
rewrite sum_N_predN /=; last lia.
 apply/Rplus_le_le_0_compat; first 
apply/Hn=>i hi; apply/hSN; lia.
Qed.

Fact sum_f_pos_0 :  forall (An : nat -> R) (N : nat), (forall n : nat, (n <= N)%nat -> 0 <= An n) -> sum_f_R0 An N = 0 -> 
(forall n : nat, (n <= N)%nat ->  An n = 0).
Proof.
move=> An.
elim.
  by rewrite /=; move =>  _ A00;case=>//; lia.
move => N HN h.
rewrite /=.
case/Rplus_eq_R0.
+ apply/sum_f_pos.
   move=> n hn ; apply h; lia.
   apply/h; lia.
move=> h1 h2.
move=> n.
case/Nat.le_succ_r=> [hN|->] //.
apply/HN=>//.
by move=> i hi ; apply/h; lia.
Qed.

Lemma xi_den_0 i : Sx (S i) + Sd i = 0  -> 
    (forall j , (j <= i)%nat ->  d j  = 0 /\ (x j) = 0 ) /\ x (S i) = 0.
Proof.
case/Rplus_eq_R0;try (apply/cond_pos_sum=>j; apply/Rabs_pos).
move=> hx hd;split.
move=> j hj; split.
apply/Rabs_eq_R0.


apply/(@sum_f_pos_0  _ _ _ hd).
move=>*; apply/Rabs_pos.
lia.
apply/Rabs_eq_R0.
apply/(@sum_f_pos_0  _ _ _ hx).
move=>*; apply/Rabs_pos.
lia.

apply/Rabs_eq_R0.
apply/(@sum_f_pos_0  _ _ _ hx).
move=>*; apply/Rabs_pos.

lia.
Qed.







Fact xi_ge0 i : 0 <= xi i.
Proof.
case:i; first by  rewrite /=; apply/Rabs_pos.
move=> i.
rewrite /xi.

case:(Req_dec (Sx (S i) + Sd i) 0)=>[|den0]; last first.
apply/(Rmult_le_reg_r (Sx (S i) + Sd i)).
move:(xi_den_ge0 i); lra.
rewrite Rmult_0_l Rmult_assoc Rinv_l // Rmult_1_r; apply/Rabs_pos.


move/xi_den_0=>[h0 h1].
rewrite Rabs_0.
rewrite /Rdiv Rmult_0_l; lra.
apply/di0.
have ->: x (S i) = 0 by [].

rewrite Rplus_0_l.
rewrite sE.
rewrite /sigma !sum_eq_R0; first ring.
move=>j /= hj.
case:(h0 j)=>//.
lia.

move=>j /= hj.
case:(h0 j)=>//.
lia.
Qed.

Hypothesis LR_hyp : forall k , Rabs (d k) <= (1 + (sigma (fun i => (xi i)) 0 k)) * Rabs (x k).

Lemma xi_den_0' i : Sx (S i) + Sd i = 0 -> d (S i) = 0 /\ (x (S i)) = 0.
Proof.
move=> hi.
suff: x (S i) = 0.
move/Rabs_0 =>xi0.
move:(LR_hyp (S i)).
rewrite xi0 Rmult_0_r.
split; last by apply/Rabs_eq_R0.

case : LR_hyp0; last by apply/Rabs_eq_R0.
move:(Rabs_pos (d (S i))); lra.

case/Rplus_eq_R0: hi;
 try (apply/cond_pos_sum=>j; apply/Rabs_pos).
move=> hSi _; move: hSi;


rewrite sigma_last /=; last lia.
case/Rplus_eq_R0; try apply/Rabs_pos.
apply/cond_pos_sum=>?; apply/Rabs_pos.
by move/Rabs_eq_R0.
Qed.


Fact xi_ge0' i : 0 <= xi i.
Proof.
case:i; first by  rewrite /=;apply/Rabs_pos.
move=> i.
rewrite /xi.
case:(Req_dec (Sx (S i) + Sd i) 0)=>[|den0].


case/xi_den_0' => ->.

by rewrite Rabs_R0 /Rdiv Rmult_0_l; lra.
apply/(Rmult_le_reg_r (Sx (S i) + Sd i)).
move:(xi_den_ge0 i); lra.
rewrite Rmult_0_l Rmult_assoc Rinv_l // Rmult_1_r; apply/Rabs_pos.
Qed.

(* Lemma xi_den_ge0 i:  0 <= Sx (S i) + Sd i. *)
(* Lemma xi_den_0 i : Sx (S i) + Sd i = 0 -> d (S i) = 0 /\ (x (S i)) = 0. *)

(* Rplus_eq_0_l: forall r1 r2 : R, 0 <= r1 -> 0 <= r2 -> r1 + r2 = 0 -> r1 = 0 *)

(* Rplus_eq_R0: *)
(*   forall r1 r2 : R, 0 <= r1 -> 0 <= r2 -> r1 + r2 = 0 -> r1 = 0 /\ r2 = 0 *)

(* Lemma xi_ge0 i: 0 <= xi i. *)

(* Hypothesis d00: d 0 = 0. *)
(* Hypothesis e00: e 0 = 0. *)

(* Hypothesis de0: forall i , d i = 0 -> e i = 0. *)


Lemma xi_le_e i: xi i <= Rabs (e i).
Proof.
case:i.
 rewrite /=; by lra.

move=>j.
rewrite /xi.
case: (Req_dec (x (S j) + s j) 0)=> den0.
  have d0: d (S j)= 0.
    by rewrite di0.
  by rewrite /xi d0 Rabs_R0 /Rdiv Rmult_0_l; apply/Rabs_pos.
have SX_ge0: 0 <= Sx (S j) by apply/cond_pos_sum => i;  apply/Rabs_pos.
have Sd_ge0: 0 <= Sd j by  apply/cond_pos_sum => i;  apply/Rabs_pos.


case:(Req_dec  (Sx (S j) + Sd j) 0 )=> deneq0.






have hx : Sx (S j) = 0 .
by apply/(Rplus_eq_0_l (Sx (S j))  (Sd j)).
have hd: Sd j = 0 by lra.

have x0: forall n : nat, (n <= (S j))%nat -> Rabs (x n) = 0.

apply/sum_f_pos_0.
 by move=>*; apply/Rabs_pos.
 by rewrite -hx.

have d0: forall n : nat, (n <= j)%nat -> Rabs (d n) = 0.

apply/sum_f_pos_0.
 by move=>*; apply/Rabs_pos.
 rewrite -hd /sigma  -minus_n_O.

by apply/sum_eq; move=>*; rewrite plus_O_n.
suff : x (S j) + s j = 0 by lra.

rewrite sE Rplus_comm Rplus_assoc.
have-> : (sigma (fun i : nat => x i) 0 j + x (S j)) = sigma (fun i : nat => x i) 0 (S j).
rewrite [RHS]sigma_last /=; first by ring.
lia.
set s1 := sigma _ _ _.

suff ->: s1 = 0 .
set s2 := sigma _ _ _.

suff ->: s2 = 0  by ring.
rewrite /s2; apply/sum_eq_R0.
rewrite -minus_n_O; move=>*; rewrite plus_O_n.
by apply/Rabs_eq_R0/x0.
rewrite /s1; apply/sum_eq_R0.
rewrite -minus_n_O; move=>*; rewrite plus_O_n.
by apply/Rabs_eq_R0/d0.

rewrite /xi.


 have  xi0 : 0 < Sx (S j) + Sd j by lra.
apply/(Rmult_le_reg_r (Sx (S j) + Sd j)) => //.
rewrite Rmult_assoc Rinv_l // Rmult_1_r.
rewrite diE Rabs_mult Rmult_comm.
apply/Rmult_le_compat_l; first by apply/Rabs_pos.
apply :(Rle_trans _ _  _  (Rabs_triang _ _)).
rewrite sE.
have -> : Sx (S j) = Rabs (x (S j)) + Sx j.
  by rewrite sigma_last /=; last lia.
rewrite Rplus_assoc ; apply/Rplus_le_compat_l.
rewrite [X in _ <= X]Rplus_comm.
apply :(Rle_trans _ _  _  (Rabs_triang _ _)).

apply/Rplus_le_compat; apply/sum_f_R0_triangle.
Qed.



Lemma Lange_Rump n : Rabs (D n) <= (Se n) * (Sx n).
Proof.

rewrite DE.
apply/(Rle_trans  _ (sigma (fun i : nat => Rabs ( d i)) 0 n)).
  apply/Rsum_abs.
apply/(Rle_trans _ (Sxi n * Sx n )); last first.
  apply/Rmult_le_compat_r.
    by apply/cond_pos_sum=> i;  apply/Rabs_pos.
  apply/sum_Rle=> i.
  rewrite !plus_0_l -minus_n_O=>hi.
  by apply/xi_le_e.
elim : n.
rewrite sigma_eq_arg.




rewrite /sigma /= Rabs_mult; lra.
move=> n Hn.


 rewrite sigma_last /=; last lia.

have SxiE:  Sxi (S n) =  (xi (S n)) + Sxi n.
  by rewrite sigma_last //; last lia.
have SxE:  Sx (S n) =  (Rabs (x (S n))) + Sx  n.
  by rewrite sigma_last //; last lia.
apply/(Rle_trans _ (  Rabs (d (S n)) + Sxi n * Sx n)).
  by apply/Rplus_le_compat_l.

 case: ( Req_dec (Sx (S n) + Sd n) 0)=> den0.
case:( xi_den_0'  den0).
do 2! (rewrite (@sigma_last _ _ (S n))/=; last lia).

move/Rabs0 ->.
move/Rabs0 ->.
rewrite /Rdiv Rmult_0_l; lra.
case:(Rlt_le_dec (Rabs ( x (S n))) ((xi (S n)) * (Sx n)))=> h.
apply/(Rle_trans _ ((1 + sigma (fun i : nat => xi i) 0 (S n)) * Rabs (x (S n)
) +  Sxi n * Sx n)).

  by apply/Rplus_le_compat_r.
by rewrite SxiE SxE;  lra.


have->: Sxi (S n) * Sx (S n) = (xi (S n)) * (Sx (S n)) + (Sxi  n) *  Rabs(x (S n)) +  Sxi n * Sx n.
rewrite  SxiE.
suff: (Sxi n) * Sx (S n) = Sxi n * Rabs (x (S n)) + Sxi n * Sx n by lra.
rewrite SxE.
ring.
apply/Rplus_le_compat_r.
have : Rabs (d (S n)) = (xi (S n)) * ((Sx (S  n)) +  (Sd n)).

rewrite /xi /=.



rewrite Rmult_assoc Rinv_l //.

ring.

have xipos:= xi_ge0.
suff :   xi (S n) * ( Sd n) <= Sxi n * Rabs (x (S n)).
by lra.

try nra.
apply/(Rle_trans _ (xi (S n) * (Sxi n * Sx n))).
apply/Rmult_le_compat_l=>//.
have->:  xi (S n) * (Sxi n * Sx n) = Sxi n * (xi (S n) * Sx n) by ring.
apply/Rmult_le_compat_l=>//.
by apply/cond_pos_sum.
Qed.

End Lange_Rump.



Lemma Lange_Rump_DWFP_hyp ( xh xl y  : R) 
         (Fy : format y)
         (DWx: double_word xh xl) (n_neg_x: 0 <= xh) (n_neg_y : 0 <= y):
        Rabs (errorDWFP p choice xh xl y) <= Rmin (xh + xl) y.
Proof.
have Hp1 := pGt1.
have RN_sym:= RN_sym.
case:(DWx)=> [[Fxh Fxl] xE].
case : n_neg_x => x0; last first.
  have xl0: xl = 0 by apply /(DW_0 DWx).
  rewrite Rabs_0.
   apply/Rmin_glb; lra.
  rewrite /errorDWFP TwoSum_correct //  TwoSum_sumE  xl0 -x0 !Rplus_0_l.
  rewrite (round_generic _ _ _ y) //; ring_simplify (xl + (y -y)). 
  have ->: y - y = 0 by ring.
  rewrite round_0 Fast2Sumf0  =>//=.
  ring.
have  xl_ub' :=  (xl_ub' xh xl (dw_ulp DWx)).
have uxh: ulp xh <= xh.
  rewrite -[X in _ <= X](Rabs_pos_eq xh); try lra.
  by  apply/ulp_le_abs =>// ; lra.
have haux: (/ 2 * ulp xh) <=  xh .
  apply/(Rle_trans _ (ulp xh)) =>//.
  suff: 0 <= ulp xh by lra.
  by apply/ulp_ge_0.
have xhl0 : 0 <= xh + xl by  lra.
case:  n_neg_y => y0; last first.
  rewrite Rabs_0.
   apply/Rmin_glb;  lra.
  rewrite /errorDWFP TwoSum_correct //  TwoSum_sumE -y0 !Rplus_0_r round_generic //.
  ring_simplify (xl + (xh -xh)); rewrite  round_generic //.
  rewrite F2Sum_correct_abs //=; try lra.
  apply/(Rle_trans _ (/ 2 * ulp xh));first  by apply/(dw_ulp DWx).
  by rewrite Rabs_pos_eq ; lra.
rewrite /errorDWFP F2Sum_correct_abs //=; try apply/generic_format_round;
 last by apply v_le_sh_W; try lra.
set  f:= RN _.
ring_simplify  (xh + xl + y - (f + (sh xh y + v xh y xl - f))).
have ->: (xh + xl + y - sh xh y = xl + sl xh y ) by rewrite TwoSum_correct //; ring.
rewrite -Rabs_Ropp  Ropp_minus_distr.
apply/Rmin_glb.
  apply/(Rle_trans _ (Rabs xl)).
  apply/plus_error_le_l =>//; try apply/generic_format_round.
  apply/(Rle_trans _ (/ 2 * ulp xh)); first  by apply/(dw_ulp DWx).
  suff : ulp xh <= xh by lra.
  by [].
apply/(Rle_trans _ (Rabs (sl xh y))).
apply/plus_error_le_r =>//; try apply/generic_format_round.
rewrite TwoSum_correct // -Rabs_Ropp  Ropp_minus_distr.
rewrite -[X in _ <= X]Rabs_pos_eq; try lra.
by apply/plus_error_le_r.
Qed.



Section exampleDWFP.

Notation xh := (1 + 2*u).
Notation xl :=  (-u + u^2).
Notation y := u.

(* Hypothesis Fyl_emin:( emin <= 2 -p -p)%Z. *)
Hypothesis ZNE : choice = fun n => negb (Z.even n).

Definition DWPlusFP_res a b c:= (fst (DWPlusFP p choice a b c) + snd (DWPlusFP p choice a b c)).


Lemma exDWFP (Hp2:( 2 < p)%Z):
  DWPlusFP_res (1 + 2*u) (-u + u^2) u =
  1 + 2*u.
Proof.
have  Hp1 := pGt1.
have RN_sym:= RN_sym.
have ulp_1 : u = /2* ulp 1 by apply/u_ulp1.
have upos: 0 < u by apply/bpow_gt_0.
have u2pos: 0 < u^2 by nra.
have u_lt1: u < /4  by rewrite /u ; change (pow (-p) < pow (-2) ); apply/bpow_lt; lia.
have uiE: u = /(pow p).
        have h:= (bpow_gt_0 two p).
       apply/(Rmult_eq_reg_l (pow p)); try lra.
       rewrite  /u -?bpow_plus; ring_simplify (p + -p)%Z; rewrite pow0; field; lra.
have xhpos: 0 < xh by lra.
have ypos: 0 < y by lra.
have F1: format 1.
  rewrite -pow0; apply/generic_format_bpow; rewrite /fexp; lia.
pose x := xh + xl.

rewrite /DWPlusFP_res /DWPlusFP.
have s1E: 1+2*u = succ two fexp 1.
  by rewrite succ_eq_pos; move: u_ulp1; lra.
have s1p2uE: 1+4*u = succ two fexp (1 + 2*u).
   rewrite succ_eq_pos ?ulp_neq_0 /cexp ; try  lra.
   rewrite (mag_unique_pos _ _ 1) ?pow0 ?pow1; last  lra.
   rewrite /fexp .
   by rewrite bpow_plus pow1 -/u; ring.
have Fxh: format xh by  rewrite s1E; apply/generic_format_succ.
have Fy: format y.
 rewrite /u; apply/generic_format_bpow; rewrite /fexp; lia.
have F1p4u: format (1+4*u) by rewrite s1p2uE; apply/generic_format_succ.
have shE: sh xh y = 1 + 4*u.
  have shlub: 1+2*u < xh + y < 1+4*u by lra.
  have rc: round two  fexp Zceil (xh + y) =  1 + 4*u.
    apply/round_UP_eq  =>//; split.
      by rewrite   s1p2uE pred_succ //; lra.
    by lra.
  have rf: round two  fexp Zfloor (xh +y )=  1+2*u.
  apply/round_DN_eq=>//; split; lra.
  rewrite TwoSum_sumE round_N_middle rc rf.
    have smE: scaled_mantissa two fexp (round two  fexp Zfloor (xh + y)) = pow (p-1) +1.
       rewrite /scaled_mantissa  rf /cexp /fexp (mag_unique_pos _ _ 1) ?pow0 ?pow1 ; 
       last  lra.
      have->: (-(1 -p) = p -1)%Z  by ring.
      rewrite Rmult_plus_distr_r Rmult_1_l.
      rewrite -pow1 /u -!bpow_plus.
      by ring_simplify  (1 + - p + (p - 1))%Z.

    have ->: (Zfloor (scaled_mantissa two fexp (xh + y)) = 2^(p-1) +1)%Z.
    apply/eq_IZR.
     rewrite -scaled_mantissa_DN; last lra.

      by rewrite smE plus_IZR (IZR_Zpower two) //;lia.
    rewrite ZNE.
    set t:= negb _; suff ->  : t  = true by [].
    rewrite /t -Zodd_even_bool Zodd_bool_iff.
    suff->: (2 ^ (p - 1) = 2* (2^(p-2)))%Z  by apply/Zodd_2p_plus_1.
    rewrite -Z.pow_succ_r; last lia.
    rewrite -Z.add_1_r; congr (_^_)%Z; ring.
  ring.
have slE: sl xh y  = - u .
  rewrite TwoSum_correct //; lra.
have ulp_2u: ulp (2 *u) = 4*u^2.
  rewrite ulp_neq_0; last lra.
  rewrite /cexp  (mag_unique_pos _ _ (2-p)).
    rewrite /fexp .
    rewrite !bpow_plus pow2 -/u;ring.
  by ring_simplify(2-p-1)%Z; rewrite !bpow_plus -/u pow1 pow2; lra.
(* have vE' : sl xh y  + xl  = -(2*u -u^2) by lra. *)
have upowp:= (bpow_gt_0 two p).
have vE: v xh  y xl = -(2*u).
  rewrite slE;ring_simplify(xl + -y).
  have->: u^2 -2*u = - (2*u - u^2) by ring.
  rewrite RN_sym //;congr Ropp.
  have F2u: format (2*u).
    rewrite -pow1 /u -bpow_plus; apply/generic_format_bpow; 
      rewrite  /fexp; lia.
  have p2u: pred two fexp (2*u)= 2*u -2*u^2.
    rewrite -pow1 /u -bpow_plus pred_bpow/fexp.
    rewrite !bpow_plus; ring.
  have Fp2u:  format (2*u -2*u^2).
    by rewrite -p2u ; apply/generic_format_pred.
  have rc: round two  fexp Zceil  (2*u -u^2) =  2*u.
    by apply/round_UP_eq=>//; split; lra.
  have rf: round two  fexp Zfloor (2*u - u^2) =  2*u -2*u^2.
    by apply/round_DN_eq; rewrite // -p2u succ_pred //; split; lra.
  rewrite round_N_middle rc rf.
    have smE: scaled_mantissa two fexp (round two  fexp Zfloor (2*u -u^2)) = pow p -1.
       rewrite /scaled_mantissa  rf /cexp /fexp (mag_unique_pos _ _ (1-p)); last first.
         ring_simplify(1-p-1)%Z; rewrite bpow_plus pow1 -/u; split; try lra.
         suff: 2*u < 1 by nra.
         lra.
      have ->: (- (1 - p - p) = p + p -1)%Z by ring.
      rewrite Rmult_plus_distr_r !bpow_plus powm1.
      have->:  2 * y * (pow p * pow p * / 2) = u* (pow p) * (pow p) by field.
      rewrite -/u -(bpow_plus _ (-p)); ring_simplify(-p +p)%Z; rewrite pow0 Rmult_1_l.
rewrite uiE; field ; lra.
    have ->: (Zfloor (scaled_mantissa two fexp  (2 * y - y ^ 2)) = 2^p -1)%Z.
      apply/eq_IZR.
      rewrite -scaled_mantissa_DN; last nra.
      by rewrite smE plus_IZR (IZR_Zpower two) //;lia.
    rewrite ZNE.
    set t:= negb _; suff ->  : t  = true by [].
    rewrite /t -Zodd_even_bool Zodd_bool_iff.
    suff ->: (2 ^ p - 1 = 2 * (2^(p-1)  -1) +1 )%Z by apply/Zodd_2p_plus_1.
    have-> :  (2 * (2^(p-1)  -1) =  2 * 2^(p-1)  -2)%Z by ring.
    rewrite -Z.pow_succ_r; last lia.
    rewrite -Z.add_1_r; ring_simplify(p-1+1)%Z;  ring.
  ring.
rewrite F2Sum_correct_abs =>//; try lia; try apply/generic_format_round.
  lra.
rewrite vE shE Rabs_Ropp !Rabs_pos_eq;  lra.
Qed.

Variable eps: R.
Variable K: R.
Hypothesis epsE: eps <= 4* u^4.

Notation den := ( u^2 + 2*u + 1).

Lemma error_exDWFP (Hp2:( 2 < p)%Z):
   relative_errorDWFP p choice (1 + 2*u) (-u + u^2) u <=  u^2 -2*u^3 + 4*u^4.
Proof.
have uiE: u = /(pow p).
have h:= (bpow_gt_0 two p).
  apply/(Rmult_eq_reg_l (pow p)); try lra.
  rewrite  /u -?bpow_plus; ring_simplify (p + -p)%Z; rewrite pow0; field; lra.
have upos: 0 < u by apply/bpow_gt_0.
have denE: den = (1 + u)^2 by  ring.
have denpos: 0 < den by nra.
have denipos: 0 </den by apply/Rinv_0_lt_compat.
(* rewrite rel_errorDWFPE. *)
rewrite /relative_errorDWFP.
set g := fst _ + snd _.
have -> : g =  DWPlusFP_res xh xl y by [].
rewrite   exDWFP //; ring_simplify(xh- (xh + xl +y)); ring_simplify (xh + xl + y).
rewrite Rabs_mult Rabs_Ropp !Rabs_pos_eq; last  nra; last  lra.
apply/(Rmult_le_reg_r den); try lra.
rewrite Rmult_assoc Rinv_l ?Rmult_1_r; try lra.
rewrite Rmult_plus_distr_r Rmult_plus_distr_r.
ring_simplify( y ^ 2 * den + - (2 * y ^ 2) * den ).
suff : y^2 - y^2*den + 2* y^3 *den <= 4 * y ^ 4 * den by lra.
ring_simplify( y^2 -y^2* den + 2* y^3 *den).
ring_simplify(4 * y ^ 4 * den).
suff: 0<= 4 * y ^ 6 + 8 * y ^ 5 + 4 * y ^ 4 - ( 2 * y ^ 5 + 3 * y ^ 4 ) by lra.
ring_simplify (4 * y ^ 6 + 8 * y ^ 5 + 4 * y ^ 4 - ( 2 * y ^ 5 + 3 * y ^ 4 )).
have->: 4 * y ^ 6 + 6 * y ^ 5 + y ^ 4 = y^4 *(1 + 6*y + 4*y^2) by ring.
nra.
Qed.

End exampleDWFP.

End DWPlus.


Section DWPlusDW.

(* Notation TwoSum_sum := (TwoSum_sum p emin choice). *)
(* Notation TwoSum_err := (TwoSum_err p emin choice). *)


(* Notation TwoSum_correct := (TwoSum_correct emax). *)


Hypothesis pgt2: (2 < p)%Z.

Fact Dwpos xh xl (DWx: double_word xh xl): 0 < xh -> 0 < xh + xl.
Proof.
move=> xhpos.
case:(DWx)=>[[Fxh Fxl]xE].
suff : (0 <= xh + xl ) /\ (xh + xl <> 0) by lra.
split.
have dw_ulp:= (Rabs_le_inv _ _ (dw_ulp pGt1 DWx)).
apply/(Rle_trans _ ( xh - (/2* ulp xh))); try lra.
suff: /2*ulp xh <= xh by lra.
apply/(Rle_trans _ (ulp xh)).
suff : 0 <= ulp xh by lra.
by apply/ulp_ge_0.
rewrite -[X in _ <= X]Rabs_pos_eq; try lra.
apply/ulp_le_abs =>//; lra.
move=> x0.
have: xh = 0 by rewrite xE x0 round_0.
lra.
Qed.


Fact DW_ge_0 xh xl (DWx: double_word xh xl): 0 <= xh -> 0 <= xh + xl.
Proof.
move=> xhpos.
case:(DWx)=>[[Fxh Fxl]xE].
case: xhpos => xh0.


case: (Rle_lt_dec 0 (xh + xl))=> //=>h.
suff : RN (xh + xl)  <= 0 by lra.
  have ->: 0 = RN 0 by rewrite round_0.
    by apply /round_le; lra.
by rewrite -xh0 (DW_0  DWx) //; lra.
Qed.

Section DWPlusDW_pre.

Variables xh xl yh yl: R.
Hypothesis DWx: (double_word xh xl).
Hypothesis DWy: (double_word yh yl).

Notation  x := (xh + xl).
Notation y := (yh + yl).

Notation  sh := (TwoSum_sum xh yh).
Notation  sl :=  (TwoSum_err xh yh).
Notation  v := (RN (xl + yl)).
Notation w := (RN (sl + v)).

(* Hypothesis xpos: 0 <= xh. *)
(* Hypothesis ypos: 0 <= yh. *)

Fact x0 (xE0: x = 0) : (xh = 0 ) /\ (xl = 0).
Proof.
case: DWx => [[Fxh Fxl] xE].
have : xh = 0 by rewrite xE xE0 round_0.
lra.
Qed.

Fact y0 (yE0: y = 0) : (yh = 0 ) /\ (yl = 0).
Proof.
case: DWy => [[Fyh Fyl] yE].
have : yh = 0 by rewrite yE yE0 round_0.
lra.
Qed.


Fact Dw0  th tl (DWt : double_word th tl) (thE0: th = 0) :  tl = 0.
Proof.
case: DWt => [[Fth Ftl] tE].

have RN0: RN (th + tl) = 0  by lra.
have : th + tl = 0.
 apply/ (round_plus_eq_0  _  _  _  _ _   _ _  RN0) =>//.
lra.

Qed.



Notation  F2Sum a b := (Fast2Sum  p  choice a b).

Notation  F2Sum_sum a b  := (fst (F2Sum a b)).
Notation  F2Sum_err a b  := (snd (F2Sum a b)).

Notation zh :=   (F2Sum_sum sh w).
Notation zl := (F2Sum_err sh w).

Definition DWPlusDW := (F2Sum sh w).

Definition  relative_errorDW2 := (Rabs (((zh + zl) - (x  + y))/ (x  + y))).
Definition errorDW2:= (x + y) - (zh + zl).
Fact rel_errorDW2E: relative_errorDW2 = Rabs errorDW2 * (Rabs (/( x + y))).
Proof.
rewrite /relative_errorDW2 /errorDW2 /Rdiv !Rabs_mult -Ropp_minus_distr Rabs_Ropp //.
Qed.

Notation ev := (v - (xl + yl)).
Notation ew := (w - (sl + v)).

Lemma errorDW2e_abs :(Rabs w <= Rabs sh) 
                ->errorDW2  =  -(ev + ew).
Proof.

move=> h.
have Hp1:= pGt1.
have RN_sym:= RN_sym.


have Fsh: format sh by apply: generic_format_round.
have Fw: format w by apply: generic_format_round.
have Fxh : format xh by case:DWx;try case.
have Fyh : format yh by case:DWy;try case.
rewrite /errorDW2 F2Sum_correct_abs //.
rewrite TwoSum_correct //=;  ring.
Qed.

Lemma rel_errorDW2e_abs :  (Rabs w <= Rabs sh) 
                ->relative_errorDW2  =  Rabs( (ev + ew) /(x + y)).
Proof.
move=> *.
by rewrite  rel_errorDW2E  errorDW2e_abs // Rabs_Ropp Rabs_mult.
Qed.


 Fact xl_ub : Rabs xl <= / 2 * ulp xh. by apply/(dw_ulp _ DWx). Qed.

Fact rel_errorDW2_0: xh = 0 ->  relative_errorDW2 = 0.
Proof.
move=> xh0.
have Hp1:= pGt1.
have RN_sym:= RN_sym.


have F0: format 0 by apply/generic_format_0.
have upos: 0 < u by apply/bpow_gt_0.
have u2pos: 0 < u^2 by nra.
case:(DWy)=>[[Fyh Fyl] yE].
have xl0: xl =  0 by apply/(DW_0 DWx); lra.
rewrite /relative_errorDW2 xl0 xh0  TwoSum_correct //. 
rewrite  Rplus_0_r  !Rplus_0_l TwoSum_sumE Rplus_0_l.
rewrite round_generic // (round_generic _ _ _ yl) //.
  have->: yh -yh = 0 by ring.
  rewrite Rplus_0_l round_generic //.
rewrite F2Sum_correct_abs //=.
rewrite Rabs_mult Rabs_0 ?Rmult_0_l; lra.
case:(Req_dec yh 0)=>yh0.
  rewrite yh0; have ->  : yl=0 by rewrite (DW_0 DWy).
  lra.
apply/(Rle_trans _ (/ 2 * ulp yh )).
  apply/(dw_ulp _ DWy)=>//.
apply/(Rle_trans _ (ulp yh)).
  suff: 0 <= ulp yh by lra.
  by apply/ulp_ge_0.
by apply/ulp_le_abs.
Qed.


Hypothesis xhpos : 0 < xh.
Hypothesis yhpos: 0 < yh.
Hypothesis yh_le_xh: yh <= xh.


Fact w_le_sh_aux  : Rabs w <= Rabs sh.
Proof.
have Hp1:= pGt1.

rewrite -!RN_abs //.
apply/round_le.
case:(DWx)=>[[Fxh Fxl] xE].
case:(DWy)=>[[Fyh Fyl] yE].

apply :(Rle_trans _ _  _  (Rabs_triang _ _)); try lra.
rewrite (Rabs_pos_eq (xh + yh)); last lra.
rewrite (Rplus_comm xh).
apply/Rplus_le_compat.
  rewrite  TwoSum_correct // -Rabs_Ropp  Ropp_minus_distr -[X in _ <= X]Rabs_pos_eq; try lra.
  by apply/plus_error_le_r.
  apply/(Rle_trans _ ( ulp xh)).
have ->: ulp xh = RN (ulp xh).
   by rewrite round_generic //; apply/generic_format_ulp.
rewrite -RN_abs //; apply/round_le.
apply :(Rle_trans _ _  _  (Rabs_triang _ _)).
have ->: ulp xh = /2* (ulp xh) + /2* (ulp xh) by lra.
apply/Rplus_le_compat; first by apply/(dw_ulp _ DWx).
apply/(Rle_trans _ (/2* ulp yh)); first by apply/(dw_ulp _ DWy).
apply/Rmult_le_compat_l/ulp_le; try lra; by rewrite !Rabs_pos_eq; lra.
rewrite -[X in _ <= X]Rabs_pos_eq; try lra.
apply/ulp_le_abs =>//; lra.
Qed.






(* Fact uxh_lbDW2: xl <> 0 -> (pow (emin -p +2) <= ulp xh). *)
(* Proof. *)
(* move=> xl0. *)
(* case:(DWx)=>[[Fxh Fxl] xE]. *)
(* have ->: (emin -p +2 = emin -p +1 +1)%Z  by ring. *)
(* rewrite bpow_plus pow1. *)
(* suff : pow (emin - p + 1)  <= /2 * ulp xh by lra. *)
(* apply/(Rle_trans _ (Rabs xl))=>//. *)
(* apply/alpha_minFP=>//; try lra. *)
(* apply/(dw_ulp DWx). *)
(* Qed. *)

(* Fact uyh_lbDW2: yl <> 0 -> (pow (emin -p +2) <= ulp yh). *)
(* Proof. *)
(* move=> yl0. *)
(* case:(DWy)=>[[Fyh Fyl] yE]. *)
(* have ->: (emin -p +2 = emin -p +1 +1)%Z  by ring. *)
(* rewrite bpow_plus pow1. *)
(* suff : pow (emin - p + 1)  <= /2 * ulp yh by lra. *)
(* apply/(Rle_trans _ (Rabs yl))=>//. *)
(* apply/alpha_minFP=>//; try lra. *)
(* apply/(dw_ulp DWy). *)
(* Qed. *)
(* Fact hmagxhDW2: (xl <> 0 \/ yl <> 0) ->(emin + 2 <= mag two xh)%Z. *)
(* Proof. *)
(* case=> [xl0 | yl0]. *)
(*   move: (uxh_lbDW2 xl0); rewrite ulp_neq_0; last lra. *)
(*   move/le_bpow; rewrite /cexp /fexp. *)
(*   by case:(Zmax_spec (mag two xh - p) (emin - p + 1));  lia. *)
(* apply/(Z.le_trans _ (mag two yh)). *)
(*   move: (uyh_lbDW2 yl0); rewrite ulp_neq_0; last lra. *)
(*   move/le_bpow; rewrite /cexp /fexp. *)
(*   by case:(Zmax_spec (mag two xh - p) (emin - p + 1));  lia. *)
(* by apply/mag_le . *)
(* Qed. *)

Fact DWPlusDW_FP : yl = 0 -> DWPlusDW  = DWPlusFP p choice xh xl yh.
Proof.
move=> yl0.
case:DWx=>[[Fxh Fxl]_].
rewrite /DWPlusDW /DWPlusFP yl0 Rplus_0_r (round_generic _ _ _ xl) //.
by rewrite (Rplus_comm xl).
Qed.

Fact DWPlusDW_FP_x : xl = 0 -> DWPlusDW  = DWPlusFP p choice yh yl xh.
Proof.
have Hp1:= pGt1.

move=> xl0.
case:DWy=>[[Fyh Fyl]_].
case:DWx=>[[Fxh Fxl]_].
rewrite /DWPlusDW /DWPlusFP xl0  Rplus_0_l (round_generic _ _ _ yl) //.
rewrite !TwoSum_correct // !TwoSum_sumE.

rewrite (Rplus_comm _ xh).
congr Fast2Sum.
congr round; ring.
Qed.


Fact rel_errorDW2E': relative_errorDW2 = Rabs ((fst DWPlusDW + snd DWPlusDW - (x + y))/(x + y)).
Proof.
by [].
Qed.



Lemma DWPlusDW_rele_bound_W:   relative_errorDW2  <= 3* u ^ 2.
Proof.

have Hp1 := pGt1.
have RN_sym:= RN_sym.

have upos : 0 < u by rewrite /u; apply/bpow_gt_0.
have u2pos : 0 < u^2 by nra.
have ult1: u < 1  by change (pow (-p) < pow 0); apply/bpow_lt ;lia.

(* have emin_neg : (emin < 0)%Z by  lia. *)

case:(DWx)=>[[Fxh Fxl]xE].
case:(DWy)=>[[Fyh Fyl]yE].
have xhn0: xh <>  0 by lra.
case:(mag two xh ) => magxh /(_ xhn0).
set k := (magxh -1)%Z.
rewrite Rabs_pos_eq; last lra.
have magxhE: (magxh = k +1)%Z by lia.
move=> xh_b.
  have xhne0: 0 <> xh by lra.
have kE: ( magxh= mag two xh)%Z by rewrite (mag_unique_pos _ _  magxh) // -/k. 
rewrite magxhE in xh_b.
case:(Req_dec xl 0)=> xl0.
  rewrite rel_errorDW2E'  DWPlusDW_FP_x  //.
  apply/(Rle_trans _ (u^2)); last lra.
  rewrite xl0 Rplus_0_r (Rplus_comm xh).  (* -rel_errorDWFPE. *)
  by apply/DWPlusFP_correct ; try lra.
have DWxpos:= (Dwpos DWx xhpos).
have DWypos:= (Dwpos DWy yhpos).
have xypos: 0 < x + y by lra.
have xyipos: 0 < / (x + y) by apply/Rinv_0_lt_compat.
  rewrite rel_errorDW2e_abs;  last by apply w_le_sh_aux.
rewrite Rabs_mult (Rabs_pos_eq (/_)); last lra.
apply/(Rmult_le_reg_r (x + y)); first lra.
rewrite Rmult_assoc  Rinv_l ?Rmult_1_r; last lra.
(* have kub: (emin +1 <= k)%Z. *)
 (*  move:(alpha_minFP Fxh xhne0 ); rewrite Rabs_pos_eq ; last lra. *)
 (*  move/(mag_le two);move /(_ (bpow_gt_0 two Emin)). *)
 (*  rewrite mag_bpow. *)
 (* have :(emin + 2 <= mag two xh)%Z. *)
 (*    apply/(@hmagxh xh xl yh)=>//. *)
 (*    by apply/dw_ulp. *)
 (*  lia. *)


have uxh:  ulp xh = 2* u * pow k.
  rewrite ulp_neq_0; last lra.
  rewrite /cexp /fexp (mag_unique_pos _ _  (k+1)).
      rewrite 2!bpow_plus -/u  pow1; ring.
  by ring_simplify  (k + 1 - 1)%Z.

have xh_ub : xh <= 2* (pow k) * (1 - u).
  have ->: 2 * pow k *( 1 - u) =  pred two fexp (pow (k + 1)).
    rewrite pred_bpow /fexp.
    rewrite (bpow_plus _ _ (-p)) -/u (bpow_plus _ k) pow1; ring.
  apply/pred_ge_gt; rewrite ?pow1 //; try lra.
  apply/generic_format_bpow; rewrite /fexp ; lia.
move/Rabs_le_inv :(xl_ub); rewrite uxh -!Rmult_assoc Rinv_l ?Rmult_1_l; try lra;
    move=> xl_u.
have  yl_ub' :=  (xl_ub' yh yl (dw_ulp pGt1 DWy)).
have uyh: ulp yh <=  2 * u * pow k.
  rewrite -uxh ; apply/ulp_le; rewrite !Rabs_pos_eq;  lra.


have y_lbub :  y <= (2 - u)* pow k by lra.
have yh_ub:  yh <= 2* (pow k) * (1 - u) by lra.

have xy_lbub: (1 -u)* pow k < x + y <= (4 - 2*u) * pow k by lra.

have Fpk: format (pow k)  by apply/format_bpow; lia.
have succkpE:  succ two fexp (pow k)  = (1 + 2*u)* pow k.
  rewrite succ_eq_pos ?ulp_neq_0; try lra.
  rewrite /cexp /fexp (mag_unique_pos _  _ (k + 1)); last first.
    have -> : (k+1 -1 =k)%Z by ring.
    rewrite bpow_plus pow1 ; lra.
  rewrite (bpow_plus _ _ (-p)) -/u (bpow_plus _ k) pow1;ring.


have xl_u':= (dw_ulp pGt1 DWx).
have yl_u':= (dw_ulp pGt1 DWy).
have xlub: Rabs xl <= u*pow k by lra.
have ylub: Rabs yl <= u*pow k.
  apply/(Rle_trans _ (/ 2 * ulp yh))=>//.
  apply/(Rle_trans _ (/ 2 * ulp xh))=>//; try lra.

have xlyl_ub : Rabs (xl + yl) <= 2 * u *pow k.
  apply :(Rle_trans _ _  _  (Rabs_triang _ _)).
  lra.
have xhyh_ub: Rabs (xh + yh) <= pow (2 + k).
  apply :(Rle_trans _ _  _  (Rabs_triang _ _)).
  rewrite bpow_plus pow2 !Rabs_pos_eq; lra.
have slub: Rabs sl <= 2 *u* pow k.
  have<-: /2*u*(pow (2+ k)) = 2 * u  * pow k.
    rewrite bpow_plus pow2 ; field.
  rewrite TwoSum_correct // -Rabs_Ropp.
    have->:  (- (xh + yh - sh)) = sh - (xh + yh) by ring.
  apply/error_le_half_ulp'=>//.
(* case: (Z_lt_le_dec (k - p -p) Emin)=> hEmin; first last. *)
have evub: Rabs ev <= u^2* pow k.
  have <- : /2 * u * pow ( k-p +1) = u^2* pow k.
    rewrite 2!bpow_plus pow1 -/u; field.
  apply/error_le_half_ulp'.

  by rewrite 2!bpow_plus -/u pow1 ; lra.

have vub: Rabs v <= 2 * u * pow k.
  have ->: 2 * u * pow k  = RN (2 * u * pow k).
    rewrite round_generic //.
    have <-:   pow (1 -p +k)= (2 * u * pow k).
      by rewrite 2!bpow_plus pow1 -/u.
    by apply/format_bpow.
  by rewrite -RN_abs //; apply/round_le.

have ewub: Rabs ew <= 2*u^2*pow k.
  have <-: /2*u*(pow (2+ k -p)) = 2 * u ^ 2 * pow k.
    rewrite 2!bpow_plus pow2 -/u; field.
  apply/error_le_half_ulp'. 

  rewrite 2!bpow_plus pow2 -/u.
  apply :(Rle_trans _ _  _  (Rabs_triang _ _)); lra.
rewrite Rmult_1_r.

case :(Rlt_le_dec ( x + y) (pow k)) => hxy1;last first.
  apply/(Rle_trans _ (3 * u * u * (pow k))).
    apply :(Rle_trans _ _  _  (Rabs_triang _ _)); lra.
  apply/Rmult_le_compat_l; lra.
have: pow k <= xh by lra.
case=> xhp.
  have xh_lb: (1 +2* u) * pow k <= xh.
    rewrite -succkpE.
    by apply/succ_le_lt.
  have : (1 + u)* pow k <= xh + xl by lra.
  lra.
have: -u*pow k <xl + y < 0 by lra.
have : xl <0 by lra.

move=>xln xly_lb.


      have ukyh: ulp (pow k +yh) = 2 * u * (pow k).
      rewrite ulp_neq_0; last lra.
      rewrite /cexp /fexp (mag_unique_pos  _ _ (k + 1)).
        by rewrite 2!bpow_plus -/u pow1; lra.
      have ->: (k +1-1 = k )%Z by ring.
      by rewrite bpow_plus pow1; lra.


  have rnd_floor_1y: round two fexp Zfloor (pow k + y) = pow k.
    apply/round_DN_eq=>//; rewrite succkpE;lra.

  have rnd_floor_1yh: round two fexp Zfloor (pow k + yh) = pow k.
    apply/round_DN_eq=>//; rewrite succkpE;lra.
have hxl: - u * pow k <= xl < 0 by lra.
have hyhxl : yh <= - xl.
rewrite yE.
have ->: -xl = RN (-xl) .
rewrite RN_sym // round_generic //.
apply/round_le=> //.
lra.
suff h:   Rabs (ev + ew) <= 5*/2*u*u*pow k.
apply/(Rle_trans _ ( 5 * / 2 * u * u * pow k)) =>//.
apply/(Rle_trans _ (3 * u * u *(1 - u) * pow k)).
apply/Rmult_le_compat_r.
apply/bpow_ge_0.
suff :  0<= u*u -6*u^3 by lra.
suff: 6*u <= 1 by nra.
apply/(Rle_trans _ (8 * u)); try lra.
change(pow 3 *u <= pow 0).
rewrite /u -bpow_plus.
apply/bpow_le; lia.
have->:   3 * u * u * (1 - u) * pow k  =   3 * u * u * ((1 - u) * pow k) by lra.
apply/Rmult_le_compat_l; try lra.


    apply :(Rle_trans _ _  _  (Rabs_triang _ _)); try lra.
suff : Rabs ev <= /2 *u *u *pow k by lra.

case:(Rle_lt_dec (-/2 * u * pow k )  xl)=>hxl'.


apply/(Rle_trans _ (Rabs yl)).
apply/

plus_error_le_r =>//.

apply/(Rle_trans _ ( / 2 * ulp yh));try lra.
suff: ulp yh <= u*u*pow k by lra.
apply/(Rle_trans _ (ulp (-xl))).
apply/ulp_le.
rewrite !Rabs_pos_eq;  lra.
rewrite ulp_opp.
apply/(Rle_trans _ (ulp (/2*  u * pow k))).
apply/ulp_le.
rewrite (Rabs_pos_eq (/2*u * pow k)); try lra.
rewrite -Rabs_Ropp Rabs_pos_eq; lra.
have->: /2 = pow (-1) by [].
rewrite /u -!bpow_plus ulp_bpow.
apply/bpow_le.
rewrite /fexp; lia.

have pxhe: pred two fexp xh = (1 - u) * pow k.
  rewrite -xhp pred_bpow/fexp.
  rewrite (bpow_plus _ k)  -/u;  ring.
have xhE: succ two fexp ((1-u)*pow k) = xh.
rewrite - pxhe succ_pred_pos //.
have rnd_floor_x: round two fexp Zfloor (xh+xl) =(1 -u )*pow k.
  apply/round_DN_eq.
    by rewrite -pxhe; apply/generic_format_pred.
  rewrite -xhE; lra.

(* rewrite round_N_eq_DN //.  *)
have rnd_ceil_x: round two fexp Zceil (xh+xl) =pow k.
  apply/round_UP_eq.
    by rewrite xhp.
  rewrite -xhp in pxhe.
  rewrite pxhe.
  have:  (1-u)*pow k <= xh + xl by lra.
  case =>h //.
    lra.
  move:xE ; rewrite -h round_generic; try lra.

  by rewrite -pxhe; apply/generic_format_pred.
move : xE; rewrite round_N_eq_DN   rnd_floor_x ; lra.
Qed.


(* have  ev0:  ev = 0. *)

(*   rewrite round_generic; first ring. *)
(*   apply/Hauser =>//. *)
(*   apply/(Rle_trans _ (pow ( k -p + 1))). *)
(*     rewrite 2!bpow_plus -/u pow1; lra. *)
(*   by apply/bpow_le; lia. *)
(* rewrite ev0 Rplus_0_l Rmult_1_r. *)


(* have ewub: Rabs ew <= 2*u^2*pow k. *)
(*   case: (Z_lt_le_dec  emin (1 + k -p))=> h. *)
(*     have <-: /2*u*(pow (2+ k -p)) = 2 * u ^ 2 * pow k. *)
(*       rewrite 2!bpow_plus pow2 -/u; field. *)
(*     apply/error_le_half_ulp'; first lia. *)
(* have vE: v = xl + yl by lra. *)
(* have vub: Rabs v <=  2 * u * pow k by  rewrite vE;lra. *)
(*    apply :(Rle_trans _ _  _  (Rabs_triang _ _)); rewrite 2!bpow_plus pow2 -/u; try lra. *)
(* have->: ew=0 . *)
(*   rewrite round_generic; first ring. *)
(*   apply/Hauser=>//; try apply/generic_format_round. *)
(*    apply :(Rle_trans _ _  _  (Rabs_triang _ _)). *)

(*   apply/(Rle_trans _ (pow ( k -p + 2))). *)
(*     rewrite 2!bpow_plus -/u pow2; try lra. *)
(* have vE: v = xl + yl by lra. *)
(* have vub: Rabs v <=  2 * u * pow k by  rewrite vE;lra. *)
(* lra. *)

(* apply/bpow_le. *)
(* try lia. *)
(* rewrite Rabs_R0; try lra. *)
(* have<-: pow ( 1 -p -p +k)=  2 * u ^ 2 * pow k. *)
(* rewrite 3!bpow_plus pow1 -/u; ring. *)
(* apply /bpow_ge_0. *)
(* apply/(Rle_trans _ (2 * u ^ 2 * pow k))=>//. *)
(* apply/(Rle_trans _ ( 3 * u * u * ((1 -u)*pow k))); last first. *)
(* apply/Rmult_le_compat_l; try lra. *)
(* rewrite -(Rmult_assoc _ _ (pow k)). *)
(* apply/Rmult_le_compat_r. *)
(* apply/bpow_ge_0. *)
(* suff : 3* u^3 <= u^2 by lra. *)
(* suff : 3*u <= 1 by nra. *)
(* apply/(Rle_trans _ (4*u)); try lra. *)
(* rewrite -pow2 /u -bpow_plus -pow0; apply/bpow_le. *)
(* lia. *)
(* Qed *)

Definition DWPlusDW_res := (fst DWPlusDW + snd DWPlusDW).

End DWPlusDW_pre.

Notation  sh xh yh := (TwoSum_sum xh yh).
Notation  sl xh yh  :=  (TwoSum_err xh yh).
(* Notation  v xl yl := (RN (xl + yl)). *)
Notation w xh xl yh yl := (RN ((sl xh yh)  + (RN( xl + yl )))).

Fact wC xh xl yh yl (DWx : double_word xh xl) 
                    (DWy: double_word  yh yl):
   w xh xl yh yl = w yh yl xh xl.
Proof.
have Hp1:= pGt1.

case:DWx=>[[Fxh Fxl] xE].
case:DWy => [[Fyh Fyl] yE].
rewrite !TwoSum_correct //.
by rewrite TwoSum_sumE !(Rplus_comm xh) (Rplus_comm xl).
Qed.




Fact w_le_sh xh xl yh yl (DWx : double_word xh xl) (xhpos: 0 <= xh)
                           (DWy: double_word  yh yl) (yhpos: 0 <= yh):
   Rabs (w xh xl yh yl) <= Rabs (sh xh yh) .
Proof.
have Hp1:= pGt1.
 
case:(DWx)=>[[Fxh Fxl] xE].
case:(DWy) => [[Fyh Fyl] yE].
case: xhpos => xhpos; last first.
  have xl0: xl= 0. apply/(Dw0 DWx); lra.
  rewrite  !TwoSum_correct // TwoSum_sumE  -xhpos xl0 !Rplus_0_l.
  have->: RN yh = yh by apply/round_generic.
  have->: RN yl = yl by apply/round_generic.
  have->: yh -yh = 0 by ring.
  rewrite Rplus_0_l round_generic //.
  case: yhpos=>yhpos; first last.
    have yl0: yl= 0 by  apply/(Dw0 DWy); lra.
    rewrite yl0 -yhpos ; lra.
  apply/(Rle_trans _ (/2 * ulp yh)).
    by apply/(dw_ulp _ DWy).
  apply/(Rle_trans _ (ulp yh)).
    suff: 0<= ulp yh by lra.
    by apply/ulp_ge_0.
  by apply/ulp_le_abs; try lra.
case: yhpos => yhpos; last first.
  have yl0: yl= 0 by apply/(Dw0 DWy); lra.
  rewrite  !TwoSum_correct // TwoSum_sumE  -yhpos yl0 !Rplus_0_r.
  have->: RN xh = xh by apply/round_generic.
  have->: RN xl = xl by apply/round_generic.
  have->: xh -xh = 0 by ring.
  rewrite Rplus_0_l round_generic //.
  apply/(Rle_trans _ (/2 * ulp xh)).
    by apply/(dw_ulp _ DWx).
  apply/(Rle_trans _ (ulp xh)).
    suff: 0<= ulp xh by lra.
    by apply/ulp_ge_0.
  by apply/ulp_le_abs; try lra.
clear Fxh Fxl xE Fyh Fyl yE.
wlog yhlexh: xh xl  yh yl DWx DWy xhpos yhpos  /  yh <= xh.
  move=>HWlog.
  case:(Rle_lt_dec yh xh)=> yhcxh.
    by apply/HWlog.
  rewrite wC // TwoSum_sumE (Rplus_comm xh).
  apply/(HWlog)=>//;  lra.
by apply/ w_le_sh_aux.
Qed.




Lemma DWle th tl (DWt: double_word th tl) : Rabs tl <= Rabs th.
Proof.
case:(DWt)=>[[Fth Ftl]tE].
case: (Req_dec th 0 )=> th0.
  by rewrite th0 (Dw0 DWt); lra.

apply/(Rle_trans _ (/2 * ulp th)).
  by apply/(dw_ulp _ DWt).
apply/(Rle_trans _ (ulp th)).
suff : 0 <= ulp th by lra.
  by apply/ulp_ge_0.

rewrite -ulp_abs;  apply/ulp_le_id; try by apply /generic_format_abs.
by apply/Rabs_pos_lt.
Qed.



Lemma relative_error_DW2C  xh xl yh yl (DWx: double_word xh xl) 
     (DWy: double_word yh yl):
  (relative_errorDW2 xh xl yh yl) = (relative_errorDW2 yh yl  xh xl).
Proof.
have Hp1:= pGt1.

case:(DWx)=>[[Fxh Fxl] xE].
case:(DWy)=> [[Fyh Fyl] yE].
rewrite /relative_errorDW2.
have <-: (yh +yl  + (xh + xl)) = (xh + xl + (yh +yl)) by ring.
by rewrite /TwoSum_sum  /TwoSum_err  TwoSumC // (Rplus_comm xl).
Qed.

Lemma errorDW2C  xh xl yh yl (DWx: double_word xh xl) 
     (DWy: double_word yh yl):
  (errorDW2 xh xl yh yl) = (errorDW2 yh yl  xh xl).
Proof.
have Hp1:= pGt1.

case:(DWx)=>[[Fxh Fxl] xE].
case:(DWy)=> [[Fyh Fyl] yE].
rewrite /errorDW2.
have <-: (yh +yl  + (xh + xl)) = (xh + xl + (yh +yl)) by ring.
by rewrite /TwoSum_sum  /TwoSum_err  TwoSumC  // (Rplus_comm xl).
Qed.


Theorem  DWPlusDW_correct ( xh xl yh yl  : R)
(DWx: double_word xh xl)
(DWy: double_word yh yl)
         (n_neg_x: 0 <= xh) (n_neg_y : 0 <= yh) :

 relative_errorDW2 xh xl yh yl <= 3* u^2.
Proof.
have Hp1:= pGt1.
 have RN_sym := RN_sym.
have F0: format 0 by apply/generic_format_0.
have upos: 0 < u by apply/bpow_gt_0.
have u2pos: 0 < u^2 by nra.
case:(DWy)=>[[Fyh Fyl] yE].
case:(DWx)=>[[Fxh Fxl] xE].
case:n_neg_x => xh0; last first.
  rewrite -xh0  /relative_errorDW2 TwoSum_correct // TwoSum_sumE 
   !Rplus_0_l.
  rewrite (round_generic _ _ _ yh) // (Rminus_diag_eq yh) // Rplus_0_l.
  rewrite round_generic; last apply/generic_format_round.
  rewrite (Dw0 DWx) ?Rplus_0_l; last lra.
  rewrite round_generic //.
  rewrite F2Sum_correct_abs //=; try apply/generic_format_round.
    rewrite Rabs_mult Rabs_0; last ring.
    rewrite Rmult_0_l; lra. 
  by apply/DWle.
case:n_neg_y => yh0; last first.
  rewrite -yh0  /relative_errorDW2 TwoSum_correct //  TwoSum_sumE 
  Rplus_0_l Rplus_0_r round_generic //.
  rewrite (Dw0 DWy) ?Rplus_0_r; try lra.
  have-> : (RN (xh - xh + RN xl)) = xl.
    ring_simplify  (xh - xh + RN xl).
    by rewrite !round_generic.
  rewrite  F2Sum_correct_abs //=.
    by rewrite  Rabs_mult Rabs_0 ?Rmult_0_l; lra.
  by apply/DWle.
clear Fyh Fyl yE Fxh Fxl xE.
wlog xhyh: xh yh xl yl DWx DWy xh0 yh0 / yh <= xh.
  move=>HWlog.
  case: (Rle_lt_dec yh xh)=>yhxh.
    by apply/HWlog.
  rewrite relative_error_DW2C //.
  apply/HWlog=>//; try lra.
by apply/DWPlusDW_rele_bound_W.
Qed.


Section exampleDW2.

Notation xh := (1 + 2*u).
Notation xl :=  (-u + u^2).
Notation yh := (9*u).
Notation yl := (-6* u^2 -8* u^3).


(* Fact Fyl_p  (Fyl: format yl): (Emin <= 3 -p -p )%Z. *)
(* Proof. *)
(* have upos: 0 < u by apply/bpow_gt_0. *)
(* have u2pos: 0 < u^2 by nra. *)
(* have  ult1: u < 1 by change (pow (-p) < pow 0); apply bpow_lt; lia. *)
(* case:(Z_le_dec Emin (3 -p -p))=>//. *)
(* move/Znot_le_gt/Z.gt_lt=> hpemin. *)
(* have yl0: 0<>  yl  by nra. *)
(* move:(alpha_minFP Fyl yl0). *)
(* have ->: yl = -(6*u^2 + 8*u^3) by ring. *)
(* rewrite Rabs_Ropp Rabs_pos_eq; last nra. *)

(* move=>h. *)
(* have: pow Emin <= pow ( 3 -p -p). *)
(* rewrite !(bpow_plus _  _ (-p)) pow3 -/u. *)
(* apply/(Rle_trans _  (6 * u ^ 2 + 8 * u ^ 3))=>//. *)
(* suff : 8*u^3 <= 2 * u^2 by lra. *)
(* suff : 4 *u <= 1 by nra. *)
(* rewrite -pow2 -pow0 /u -bpow_plus; apply/bpow_le;lia. *)
(* by move/le_bpow. *)
(* Qed. *)


(* Hypothesis Fyl_emin:( emin <= 2 -p -p)%Z. *)
Hypothesis ZNE : choice = fun n => negb (Z.even n).



Lemma exDW2 (Hp4: (4 <= p)%Z):
  DWPlusDW_res (1 + 2*u) (-u + u^2) (9*u) (-6* u^2 -8* u^3) =
  1 + 10*u -8* u^2.
Proof.
have Hp1:= pGt1.
have RN_sym := RN_sym.
have u_ulp1 : pow (-p) = /2 * ulp 1 
  by apply/u_ulp1; move: emin_max_p;lia.
(* have emin_neg : (emin < 0)%Z by  lia. *)
have upos: 0 < u by apply/bpow_gt_0.
have u2pos: 0 < u^2 by nra.
have xhpos: 0 < xh by lra.
have yhpos: 0 < yh by lra.

set x := xh + xl.
set y := yh + yl.

have den: x+ y = 1+ 10*u -5* u^2 -8 * u^3 by rewrite /x /y; ring.
 rewrite /DWPlusDW_res /DWPlusDW.
have F1 : format 1.
  change (format (pow  0)).
  apply/generic_format_bpow; rewrite /fexp; lia.
set sh:=  (TwoSum_sum xh yh).
set  sl :=  (TwoSum_err xh yh).
(* set  v := (RN (xl + yl)). *)
set w := (RN (sl + RN ( xl + yl))).
have xhE: xh = succ two fexp 1.
 rewrite succ_eq_pos /u ?u_ulp1;  lra.
have Fxh: format xh.
rewrite xhE;apply/generic_format_succ=>//.
have Fxl: format xl.
have ->: -u + u^2 = - (u - u^2) by ring.
apply/generic_format_opp.
have mxlE: u -u^2 = pred two fexp u.
rewrite /u pred_bpow/fexp.
rewrite bpow_plus; ring.
rewrite mxlE ; apply/generic_format_pred.
  by rewrite /u; apply/format_bpow.
have Fyh: format yh.
 pose f9:= Float two 9 0.
 have fE: 9 = F2R f9.
  by rewrite /F2R /=; ring.
  have F9: format 9.
    apply/(FLX_format_Rabs_Fnum p  fE) =>/=. 
    apply/(Rlt_le_trans _ (pow 4)).
      rewrite Rabs_pos_eq ; try lra.
      change (9 < 16); lra.
    by apply/bpow_le; lia.
  rewrite /u ; apply/mult_bpow_exact_FLX =>//.
  (* rewrite (mag_unique_pos _ _ 4). *)
  (*   lia. *)
  (* by ring_simplify(4-1)%Z; change (8 <= 9 < 16) ; lra. *)

have Fyl:format yl.
  have ->: -6*u^2 -8* u^3 = - ((6  + 8*u)* u^2)by ring.
  apply/generic_format_opp.
  have->: u^2 = pow (-p -p) by rewrite bpow_plus /u; ring.
  have mE: ( 3 = (mag two (6 )))%Z.
    rewrite (mag_unique_pos _ _ 3) //.
    ring_simplify (3-1)%Z; rewrite pow2 ; split; try lra.
    have->: pow 3 = 8 by [].
    suff: 4*u < 1 by lra.
    change (pow 2 * pow (-p) < pow 0).
    by rewrite -bpow_plus; apply/bpow_lt; lia.
  apply/mult_bpow_exact_FLX.

      suff->: 6+ 8*u = succ two fexp 6.
        apply/generic_format_succ.
        pose f6:= Float two 6 0.
        have f6E: 6 = F2R f6.
          by rewrite /F2R /=; ring.
        apply/(FLX_format_Rabs_Fnum p f6E) =>/=.
        rewrite Rabs_pos_eq; last lra.
        apply/(Rlt_le_trans _ 8); try lra.
        change(pow 3 <= pow p); apply/bpow_le; lia.
      rewrite succ_eq_pos; try lra.
      rewrite ulp_neq_0; try lra.
      rewrite /cexp -mE /fexp.
    by rewrite bpow_plus /u; have->: pow 3 = 8 by [].

have  ult1: u < 1 by change (pow (-p) < pow 0); apply bpow_lt; lia.
have DWx: double_word xh xl.
  split=>//.
  have->: xh + xl = 1+ u + u^2 by ring.
  have rdn1: (round two fexp Zfloor (1 + u + u ^ 2) )=1.
    rewrite Rplus_assoc round_DN_plus_eps_pos //; try lra.
    have ->: ulp 1 = 2*u by rewrite /u u_ulp1; field.
    nra.
  have rup: round two fexp Zceil (1 + u + u ^ 2) = 1+2*u.
    have ulpxE: 2*u = ulp  (xh+xl).
      rewrite ulp_neq_0; try lra.
      rewrite /cexp (mag_unique_pos _ _ 1) /fexp.
        by rewrite bpow_plus. 
      rewrite pow0 pow1; split; try lra.
      suff: u + u^2 <1 by lra.
      apply/(Rlt_le_trans _ (2*u)); try lra.
        have-> : u+u^2 = (1 + u) *u by ring.
        apply/Rmult_lt_compat_r; lra.
      rewrite -pow1 /u -bpow_plus -pow0; apply/bpow_le; lia.
    rewrite ulpxE.
    rewrite round_UP_DN_ulp  ?rdn1 //.
      congr Rplus ; congr ulp; ring.
    apply/(@generic_format_discrete _ _ _ (2^(p -1))).
    set ce:= cexp _ _ _.
    have->:( ce =  1-p)%Z.
      rewrite /ce /cexp (mag_unique_pos _ _ 1).
        by rewrite /fexp.
      rewrite pow0 pow1; split; try lra.
(* factoriser *)
      suff: u + u^2 <1 by lra.
      apply/(Rlt_le_trans _ (2*u)); try lra.
        have-> : u+u^2 = (1 + u) *u by ring.
        apply/Rmult_lt_compat_r; lra.
      rewrite -pow1 /u -bpow_plus -pow0; apply/bpow_le; lia.
    set e := (1-p)%Z.
    rewrite /F2R /= /e bpow_plus pow1 -/u.
    rewrite plus_IZR (IZR_Zpower two); last lia.
    rewrite Rmult_plus_distr_r.
    have ->:   pow (p - 1) * (2 * u) = 1 .
      rewrite -pow1 /u -!bpow_plus.
      by ring_simplify (p - 1 + (1 + - p))%Z.
    split; try lra.
    by suff:  u < 1 by nra.
  rewrite round_N_eq_UP //.
  rewrite rdn1 rup.
  lra.
    
have magyh: (mag two yh = (-p +4)%Z :>Z).
  rewrite (mag_unique_pos _ _ (-p+4))=>//.
  ring_simplify(- p + 4 - 1)%Z.
  rewrite 2!bpow_plus -/u pow3.
  have->: pow 4 = 16 by [].
  lra.
have DWxpos:= (Dwpos DWx xhpos).

have ypos: 0 < y.
  suff : 6*u+8*u^2 < 9  by  rewrite /y ; nra.
  apply/(Rlt_trans _ (7*u)); last lra.
  suff : 8*u < 1 by nra.
  change (pow 3 * pow (-p) < pow 0).
  rewrite -bpow_plus; apply/bpow_lt; lia.

have magy: (mag two y = (-p +4)%Z :>Z).
  rewrite (mag_unique_pos _ _ (-p+4))=>//.
        rewrite  3!bpow_plus -/u powm1.
        have->: pow 4 = 16 by [].
  split.
    suff : 6*u + 8*u^2 <= 1 by rewrite /y;  nra.
    apply/(Rle_trans _ (8*u)); try lra.
      suff: 4*u <= 1 by nra.
      rewrite -pow2 /u -pow0 -bpow_plus ; apply/bpow_le; lia.
    rewrite -pow3 /u -pow0 -bpow_plus ; apply/bpow_le; lia.
  by rewrite /y ; nra.

have DWy: double_word yh  yl.
  split=>//.
  have yE: yh + yl = 9*u -(6*u^2 +8 * u^3) by ring.
  have ulpyhE: ulp yh = 16* u^2.
    rewrite ulp_neq_0; last lra.
    rewrite /cexp /fexp.
      rewrite magyh 2!bpow_plus -/u.
      have->: pow 4 = 16 by [].
      ring.
  have pyh: pred two fexp yh = yh - 16*u^2.
  rewrite pred_eq_pos; try lra.
  rewrite /pred_pos.
  rewrite magyh.
  ring_simplify (- p + 4 - 1)%Z.
  rewrite bpow_plus pow3 -/u  Req_bool_false; lra.
   have ulpyE: ulp y = 16* u^2.   
    rewrite ulp_neq_0; last lra.
    rewrite /cexp /fexp.
    rewrite magy; try lia.
    rewrite  2!bpow_plus -/u.
    have->: pow 4 = 16 by [].
    ring.
  have rc: round two  fexp Zceil y =  yh.
    apply/round_UP_eq=>//; split.
      rewrite pyh.
      apply/Rplus_lt_compat_l.
      suff: 6* u^2  + 8*u^3 < 16 *u^2 by lra.
      apply/(Rlt_trans _ (8*u^2)); try lra.
      suff: 4*u<1 by nra.
      rewrite -pow2 -/u -pow0 -bpow_plus ; apply/bpow_lt; lia.
    rewrite /y;nra.
  have rf: round two  fexp Zfloor y =  yh -16*u^2.
    have ->: y = yh - (6*u^2 + 8*u^3) by rewrite /y; ring.
    rewrite -pyh round_DN_minus_eps_pos// pyh ulp_neq_0.
      split; try nra.
      rewrite /cexp/fexp    (mag_unique_pos _ _ (-p+4))=>//; try nra.
        rewrite !bpow_plus -/u ; have->: pow 4 = 16 by [].
        suff: u <= 1 by nra.
        lra.
      ring_simplify (-p +4 -1)%Z ; rewrite !bpow_plus pow3 -/u.
      split;try nra.
        suff: 16*u <= 1  by nra.
        change (pow 4 * pow (-p) <= pow 0); rewrite -bpow_plus; apply/bpow_le.
        lia.
      have->: pow 4 = 16 by [];  nra.
    suff: 2 * u < 1  by nra.
    rewrite -pow1 -pow0 -bpow_plus; apply/bpow_lt; lia.
  rewrite round_N_eq_UP //.
  rewrite rc rf ; try lra.
  suff : -8 * u^2 < yl by lra.
  suff:  4*u < 1 by nra.
  by rewrite -pow2 -pow0 /u -bpow_plus; apply/bpow_lt; lia.
rewrite  F2Sum_correct_abs //; try  lia; try apply/generic_format_round; last first.
  by  apply/w_le_sh =>//; lra.
set f := fst _.
ring_simplify( f + (sh + w - f)).
clear f.
pose f := Float two (2^(p-1)  + 5) 1.
have h:pow p + 10 = F2R f.
  rewrite /f /F2R /=  plus_IZR (IZR_Zpower two); try lia.
  have ->: ( IZR (Z.pow_pos 2 1) = 2 ) by [].
  rewrite bpow_plus powm1; field.
have h12: 12 < pow p.
  apply/(Rlt_le_trans _ (pow 4)); rewrite pow4; try lra.
  by rewrite -pow4; apply/bpow_le; lia.
have Fp10: format (pow p + 10).
  apply/(FLX_format_Rabs_Fnum p h); rewrite /f/=.
  rewrite plus_IZR (IZR_Zpower two); last lia.
  rewrite bpow_plus powm1 Rabs_pos_eq; last first.
    move:(bpow_ge_0 two p); lra.
  suff : 10 < pow p by lra.
  by lra.
have mp10:   mag two (pow p + 10) = (p + 1 )%Z :>Z.
  rewrite (mag_unique_pos _  _ (p+1)) //; 
   ring_simplify (p +1 -1)%Z; rewrite bpow_plus pow1; split;lra.
have mp11:   mag two (pow p + 11) = (p + 1 )%Z :>Z.
  rewrite (mag_unique_pos _  _ (p+1)) //; 
   ring_simplify (p +1 -1)%Z; rewrite bpow_plus pow1; split;lra.
have up10:  ulp (pow p + 10) = 2.
  rewrite ulp_neq_0; try lra.
  rewrite /cexp /fexp .
  by rewrite mp10; ring_simplify(p+1-p)%Z.

have rf: round two  fexp Zfloor (pow p + 11) =  pow p + 10.
  have->: pow p +  11 = pow p + 10 + 1 by ring.
  rewrite round_DN_plus_eps_pos //; try  lra.
have nFp11: ~ format (pow p + 11).
apply/(generic_format_discrete _ _ _ (2^(p-1)+5)).
rewrite  /cexp /fexp mp11.
ring_simplify(p + 1 -p)%Z.
rewrite /F2R /= IZR_Zpower_pos /= Rmult_1_r !plus_IZR (IZR_Zpower two); last lia.

by rewrite bpow_plus powm1; lra.
have rc : round two fexp Zceil (pow p + 11) = pow p + 10 + 2.
  rewrite -rf round_UP_DN_ulp // ulp_neq_0 /cexp  /=; try lra.
  rewrite mp11 /fexp.
  by ring_simplify(p+1-p)%Z; rewrite pow1.


have shE : sh = 1 + 12 *u.
  rewrite /sh TwoSum_sumE .
  have ->: xh + yh = 1 + 11 *u by lra.
  have ->: (1 + 11*u) = (pow p  + 11) *u .
    rewrite /u Rmult_plus_distr_r -bpow_plus.
    ring_simplify (p + -p)%Z; rewrite pow0; ring.
  rewrite round_bpow .
  apply/(Rmult_eq_reg_r (pow p)); last by  move: (bpow_gt_0 two p); lra.
  rewrite Rmult_assoc  -bpow_plus; ring_simplify (-p+p)%Z; rewrite Rmult_1_r.


  rewrite round_N_middle rc rf.
  rewrite ZNE /=.
set test:= negb _.
case e: test.

rewrite Rmult_plus_distr_r /u Rmult_assoc -bpow_plus; ring_simplify(-p +p)%Z.
rewrite pow0 ;ring.



move:e  ; rewrite /test.
move/Bool.negb_false_iff.
rewrite /scaled_mantissa /cexp mp11/fexp.
ring_simplify(-(p +1 -p))%Z.
rewrite Rmult_plus_distr_r -bpow_plus powm1 /Zfloor.
have ->: (p + (-1) = p-1)%Z by ring.
have->: up (pow (p - 1) + 11 * / 2) = (2^ (p - 1) + 6)%Z.
apply/eq_sym/tech_up;rewrite plus_IZR (IZR_Zpower two); try lra; lia.


move/Zeven_bool_iff.
suff: Zodd ( (2 ^ (p - 1) + 6 - 1)).

by move/Zodd_not_Zeven.
have->: ((2 ^ (p - 1) + 6 - 1) = (2 * (2 ^ (p-2) + 2) +1))%Z.
suff <-: (2 * (2 ^ (p - 2)) = 2^ (p-1))%Z by lia.
have ->:   (2 * 2 ^ (p - 2) = 2^1 * 2 ^( p -2))%Z by ring.
rewrite -Zpower_plus; try lia.
by ring_simplify (1 + (p - 2))%Z.

apply/Zodd_2p_plus_1.
ring.
  
have slE: sl = -u.
rewrite /sl TwoSum_correct //.
move: shE; rewrite /sh => ->; ring.

have magu : mag two  u = (1-p)%Z :>Z.
rewrite (mag_unique_pos _ _ (1 -p)) //.
ring_simplify(1-p-1)%Z; rewrite -/u ;split; try lra.
rewrite bpow_plus pow1 -/u; lra.
have ulpu: ulp u = 2*u^2.
rewrite ulp_neq_0; try lra.
rewrite /cexp /fexp .
rewrite magu !bpow_plus pow1 -/u; ring.

have vE : RN (xl + yl) = -u  - 6*u^2.
  have ->: xl + yl = -(u + 5*u^2 + 8 * u^3) by ring.
  rewrite RN_sym //.
  suff :  RN (u + 5 * u ^ 2 + 8 * u ^ 3) = (1 + 6*u) * u by lra.
  have-> : u + 5 * u ^ 2 + 8 * u ^ 3= (1 + 5*u + 8 * u^2)*u by ring.
  rewrite  round_bpow.
(* rewrite (mag_unique_pos _ _ 1). *)
(* rewrite pow0 pow1; split; try lra. *)
(* suff:  5 * u + 8 * u ^ 2 < 1 by lra. *)
(* apply/(Rlt_le_trans _ (8*u)); try lra. *)
(* suff: 4*u < 1 by nra. *)
(* change(pow 2 * pow (-p) < pow 0); rewrite -bpow_plus ; apply/bpow_lt; lia. *)
(* change (pow 3 * pow (-p) <= pow 0); rewrite -bpow_plus ; apply/bpow_le; lia *)
apply/Rmult_eq_compat_r.


have F1p2uE: 1+ 2*u = succ two fexp 1.
  rewrite succ_eq_pos /u ?u_ulp1; lra.
have xhp2u : 1+ 4*u = succ two fexp xh.
  rewrite succ_eq_pos; try lra.
  rewrite ulp_neq_0; last lra.
  rewrite /cexp /fexp (mag_unique_pos _ _ 1).
  rewrite bpow_plus pow1 -/u; ring.
  rewrite pow0 pow1;split; try lra.
  suff: 2*u < 1 by lra.
  rewrite -pow1 -pow0 /u -bpow_plus; apply/bpow_lt; lia.

have ->:  (1 + 5 * u + 8 * u ^ 2)= 1+4*u +( u + 8*u^2) by ring.
have s1p6u : 1+ 6*u = succ two fexp (1+ 4*u).
   rewrite succ_eq_pos; last lra.
rewrite ulp_neq_0; last lra.
rewrite /cexp  (mag_unique_pos _ _ 1).
rewrite bpow_plus pow1 /u; ring.
rewrite pow0 pow1; split; try lra.
suff: 4*u < 1 by lra.
rewrite -pow2 -pow0 /u -bpow_plus; apply/bpow_lt; lia.
clear rc rf.
have rc:   round two fexp Zceil (1 + 4 * u + (u + 8 * u ^ 2)) = 1 + 6 * u.
    apply/round_UP_eq.
        rewrite s1p6u;
   apply/generic_format_succ; rewrite xhp2u ; apply/generic_format_succ=>//.

rewrite s1p6u pred_succ.
split; try lra.
rewrite -s1p6u.
suff:  8 * u  <= 1 by nra.
change (pow 3 * pow (-p) <= pow 0); rewrite -bpow_plus ; apply/bpow_le; lia.
rewrite   xhp2u; apply/generic_format_succ=>//.

have rf : round two fexp Zfloor (1 + 4 * u + (u + 8 * u ^ 2)) = 1 + 4 * u.
apply/round_DN_eq.
rewrite  xhp2u; apply/generic_format_succ=>//.
split; try lra.
rewrite -s1p6u.
suff:  8 * u  < 1 by nra.
change (pow 3 * pow (-p) < pow 0); rewrite -bpow_plus ; apply/bpow_lt; lia.
rewrite round_N_eq_UP rc // rf.
lra.

rewrite shE.
suff ->: w = -2*u -8 * u^2 by ring.
have -> : w = RN (-u -u -6*u^2 ).
rewrite /w slE vE; congr round; ring.
have->: (- u - u - 6 * u ^ 2)  = -( 2*u + 6*u^2) by ring.
rewrite RN_sym //.
suff: RN (2 * u + 6 * u ^ 2)= 2*u + 8*u^2 by lra.
have-> : (2 * u + 6 * u ^ 2) = (1 + 3*u) * (2*u) by ring.
rewrite -pow1 /u -bpow_plus.
rewrite  round_bpow.
(*   rewrite (mag_unique_pos _ _ 1). *)
(*   rewrite pow0 pow1 -/u; split; try lra. *)
(*    suff: 3*u < 1 by lra. *)
(* apply/(Rlt_le_trans _ (4*u)); try lra. *)
(* rewrite -pow2 -pow0 /u -bpow_plus; apply/bpow_le; lia. *)

have ->:  pow (1 + - p) + 8 * pow (- p) ^ 2 = (1 + 4*u) * (2*u).
rewrite bpow_plus pow1 -/u ; ring.
rewrite bpow_plus pow1 -/u.
apply/(Rmult_eq_compat_r).

have F1p2uE: 1+ 2*u = succ two fexp 1.
rewrite succ_eq_pos /u ?u_ulp1; lra.
have xhp2u : 1+ 4*u = succ two fexp xh.
  rewrite succ_eq_pos; try lra.
rewrite ulp_neq_0; last lra.
rewrite /cexp /fexp (mag_unique_pos _ _ 1).
rewrite bpow_plus pow1 -/u; ring.
rewrite pow0 pow1;split; try lra.
suff: 2*u < 1 by lra.
rewrite -pow1 -pow0 /u -bpow_plus; apply/bpow_lt; lia.
clear rf rc.
have rf  : round two  fexp Zfloor (1+ 3*u ) = 1+ 2*u.
have ->: 1+ 3*u  = (1 + 2*u) + u by ring.
rewrite round_DN_plus_eps_pos //; try lra.
  rewrite ulp_neq_0; last lra.
  rewrite /cexp/fexp (mag_unique_pos _ _ 1); try lra.
    rewrite bpow_plus pow1 -/u; lra.
rewrite pow0 pow1; split; try lra.
suff: 2*u <1 by lra.
rewrite -pow0 -pow1 -bpow_plus; apply/bpow_lt; lia.
have rc:  round two  fexp Zceil (1+ 3*u ) = 1+ 4*u.
apply/round_UP_eq.
    by rewrite   xhp2u; apply/generic_format_succ.
  by rewrite xhp2u pred_succ //;  lra.

rewrite round_N_middle; last lra.
rewrite ZNE -Zodd_even_bool.
set z := Z.odd _.

suff ->: z = true; first lra.
apply/Zodd_bool_iff.
rewrite /scaled_mantissa /cexp  (mag_unique_pos _ _ 1); last first.
rewrite pow0 pow1; split; try lra.
suff : 3*u < 1 by lra.
apply/(Rlt_le_trans _ (4*u)); try lra.
rewrite -pow2 -pow0 /u -bpow_plus; apply/bpow_le; lia.
rewrite /Zfloor.
have->: (-(1-p) = p-1 )%Z by ring.
have ->:  ((1 + 3 * u) * pow (p - 1)) = pow (p-1) + 3 * (/2).
rewrite Rmult_plus_distr_r /u Rmult_assoc -bpow_plus.
ring_simplify(- p + (p - 1))%Z; rewrite powm1; ring.
have->: up (pow (p - 1) + 3 * / 2) = (2^ (p - 1) + 2)%Z.
apply/eq_sym/tech_up;rewrite plus_IZR (IZR_Zpower two); try lra; lia.
suff ->:  ((2 ^ (p - 1) + 2 - 1) = 2 * (2^(p-2)) + 1 )%Z by apply/Zodd_2p_plus_1.
suff ->: (2* 2^(p-2) = 2^(p-1))%Z by ring.
rewrite -Z.pow_succ_r -?Z.add_1_r; try lia.
by ring_simplify(p-2+1)%Z.

Qed.

Notation den :=  (-8 * u ^ 3 - 5 * u ^ 2 + 10 * u + 1).

Lemma error_exDW2 (Hp3:( 3 < p)%Z) K (Ke : K = 395):
   relative_errorDW2 (1+2*u) (-u+u^2) (9*u) (-6*u^2 -8*u^3)  <=  3*u^2 -38*u^3 + 395*u^4.
Proof.
have uiE: u = /(pow p).
  have h:= (bpow_gt_0 two p).
  apply/(Rmult_eq_reg_l (pow p)); try lra.
  by rewrite  /u -?bpow_plus; ring_simplify (p + -p)%Z; rewrite pow0; field; lra.
have upos: 0 < u by apply/bpow_gt_0.
have denpos: 0 < den.
  apply/(Rlt_le_trans _ (1+8*u)); first lra.
  suff: 5*u +8*u^2 <= 2 by nra.
  apply/(Rle_trans _ (8*u)); try lra; last first.
    rewrite -pow3 -pow1 /u -bpow_plus; apply/bpow_le; lia.
  suff: 4*u <= 1 by nra.
  rewrite -pow2 -pow0 /u -bpow_plus; apply/bpow_le; lia.
have denipos: 0 </den by apply/Rinv_0_lt_compat.
rewrite rel_errorDW2E /errorDW2.
have u16: u <= /16.
  change (pow (-p) <= pow (-4)); apply/bpow_le;lia.
set g := fst _ + snd _.
have -> : g =  DWPlusDW_res xh xl yh yl by [].
rewrite   exDW2;  last lia.

 ring_simplify (xh + xl + (yh + yl) - (1 + 10 * u - 8 * u ^ 2)).
  ring_simplify (xh + xl + (yh+yl)).
rewrite !Rabs_pos_eq; try nra.
apply/(Rmult_le_reg_r den); try lra.
rewrite Rmult_assoc Rinv_l ?Rmult_1_r; nra.
Qed.

End  exampleDW2.



Fact DWlp_ub' xh xl (DWx: double_word xh xl) :  Rabs xl <= u* Rabs(xh+xl).
Proof.
have upos: 0 < u by apply/bpow_gt_0.
case:(DWx)=> [[Fxh Fxl] xE].
pose x := xh + xl.
case:(Req_dec xh 0)=> xh0.
  rewrite( Dw0 DWx) xh0 // Rplus_0_r Rabs_0  ; lra.
case:(Req_dec 0 xl)=>  xl0.
rewrite -xl0 Rabs_R0 Rplus_0_r; move:(Rabs_pos xh);   nra.

(*  have : pow (emin - p +2) <= ulp (xh+xl). *)
(* apply/(Rle_trans _ (2*Rabs xl)). *)
(* have->:(emin - p +2 = (emin -p +1) +1)%Z by ring. *)
(* rewrite bpow_plus pow1; suff: pow Emin <= Rabs xl by lra. *)
(* by apply/alpha_minFP. *)
(* move: (dw_ulpx  DWx); lra. *)
have xn0: xh + xl <> 0.
  case: (Req_dec (xh+xl) 0) =>// h.
  have : RN (xh + xl) = 0 by rewrite h round_0.
  by rewrite -xE.
(* rewrite ulp_neq_0; try   lra. *)

(* rewrite /cexp /fexp. *)
(* move/le_bpow. *)

(* case/Z.max_le; try lia. *)
(* move=>hh. *)

  apply/(Rle_trans _ ( /2*ulp (xh + xl))).
    have h : ((RN (xh + xl ))  - (xh + xl))  = -xl
       by rewrite -xE; ring.
    by rewrite -Rabs_Ropp -h; apply/error_le_half_ulp.

  suff:  ulp (xh + xl) <= 2*  u * Rabs(xh + xl) by lra.
  rewrite -pow1 /u -bpow_plus.
  have->: pow (1 + - p) * Rabs (xh + xl) =  Rabs x * pow (1 - p).
  rewrite Rmult_comm; congr Rmult.
(* rewrite Rabs_pos_eq; last by  apply/(DW_ge_0 DWx); lra. *)
rewrite ulp_neq_0  /cexp /fexp ; last lra.
rewrite  !bpow_plus  -Rmult_assoc; apply/Rmult_le_compat_r.
apply/bpow_ge_0.
suff:   pow (mag two (xh + xl)) * /2 <= Rabs x by rewrite pow1; lra.
rewrite -powm1 -bpow_plus.
(* -[X in _<= X] Rabs_pos_eq; rewrite /x; try lra. *)
apply/bpow_mag_le; try lra.
Qed.


Lemma error_plus_le a b (Fa : format a ) (Fb: format b):
     Rabs (RN (a+b) - (a + b) ) <= u * (Rabs (a +b)).
Proof.
have upos: 0 < u by apply/bpow_gt_0.
have ult1: u < 1 by rewrite /u -pow0; apply/bpow_lt; lia.

 have -> :u =  / 2 * pow (- p + 1).
rewrite !bpow_plus pow1 /u; lra.
by apply/relative_error_N_FLX.
Qed.

Fact DWlp_ub xh xl (DWx: double_word xh xl) (xh_ge0: 0 <= xh):  Rabs xl <= u* (xh+xl).
Proof.
have xge0:= (DW_ge_0 DWx xh_ge0).
case:(DWx)=> [[Fxh Fxl] xE].
have {1}->: xl = -( RN (xh + xl) - (xh + xl)) by rewrite -xE; ring.
rewrite Rabs_Ropp -[X in _ <= _ * X](Rabs_pos_eq _ xge0).
by apply/error_plus_le.
Qed.




Lemma Rmin_glb_inv : forall x y z : R,  z <= Rmin x y ->  (z <= x /\ z <= y ).
Proof.
move=> x y z.
rewrite /Rmin; case:(Rle_dec x  y);  lra.
Qed.
Variable K:R.

End  DWPlusDW.

End Main.


