(* Copyright (c) ENS de Lyon and Inria. All rights reserved. *)

Require Import Reals  Psatz.
From Flocq Require Import Core Relative Sterbenz Operations Mult_error  Plus_error.
From mathcomp Require Import ssreflect.all_ssreflect.


Require Import Bayleyaux.
Require Import F2Sum DWPlus.
Require Import EN_aux_flx square_root_flx.

From Flocq Require Import Div_sqrt_error Sqrt.

From Interval Require Import Tactic.

Open Scope R_scope.

Set Implicit Arguments.

Section Main.

Notation two  := radix2.
Notation pow e := (bpow two e).


Variable p: Z.
Context { prec_gt_0_ : Prec_gt_0 p }.
Variable choice : Z -> bool.
Hypothesis choice_sym: forall x, choice x  = negb (choice (- (x + 1))).


Hypothesis pGt1: (1 < p)%Z.

Let u := pow (-p).


(* no underflow, no overflow *)

(***********************   FLX ************************************)
Notation fexp := (FLX_exp   p).
Notation format := (generic_format two fexp ).
Notation RN :=(round two  fexp (Znearest choice) ).


Local Notation ulp := (ulp two fexp).

Variable A : nat -> R.
Hypothesis FA : forall i,  format (A i).


Definition Fast2Mult (a b : R) :=
  let pi := RN (a * b) in
  let rho := RN (a * b - pi) in
  (pi, rho).

Notation F2P_prod a b  :=  (fst (Fast2Mult a b)).
Notation F2P_error a b  :=  (snd (Fast2Mult a b)).

Hypothesis F2Mult_correct:
  forall a b, a * b =  F2P_prod a b +  F2P_error a b.

Fact  F2P_errorE : forall a b, F2P_error a b =  a * b - RN(a * b).
Proof. move=> ? ?; rewrite {1}F2Mult_correct /=; ring. Qed.

Fact DW_F2P  a b : double_word p choice (F2P_prod a b ) ( F2P_error a b).
Proof.
split; [split; apply/generic_format_round|].
by rewrite -F2Mult_correct.
Qed.

Section Algo_10_11.

Section Algo_10_pre.


Variable n: nat.
Hypothesis n1: (1 <= n)%N.

Variable y : nat -> R * R.

Notation y_h i := (fst (y i)).
Notation y_l i := (snd (y i)).

Notation double_word := (double_word p choice).


Hypothesis DWyi: forall i, double_word (y_h i) (y_l i).

Hypothesis  yh_pos: forall i,  0 <= (y_h i).

Fact yluyh i: Rabs (y_l i) <= u * y_h i.
Proof.
case:(DWyi i)=>[[Fyh Fyl]] yE.
rewrite -(Rabs_pos_eq (y_h _)) //.
have->: (y_l i) =-( y_h i - (y_h i + y_l i)) by ring.
rewrite {1 3}yE Rabs_Ropp.
have<-: / 2 * pow (- p + 1)= u.
  by rewrite bpow_plus -/u pow1;  lra.
by apply/ relative_error_N_FLX_round.
Qed.

Fact ypos i: 0 <= y_h i + y_l i.
Proof.
move/Rabs_le_inv:  (yluyh i)=> yluyh.
apply/(Rle_trans _ (( 1 - u) * (y_h i)));last  lra.
have ult1: u < 1 by rewrite -pow0 /u; apply/bpow_lt; lia.
move:(yh_pos i); nra.
Qed.

(* Algo10.1  *)

Notation TwoSum := (TwoSum p choice).
Notation DWPlusFP := (DWPlusFP p choice).


(* Algo10.2*)
Definition x1 := (TwoSum  (y_h 0) (y_h 1)).


Fixpoint x (i : nat) : R *  R :=
match i with 
|0 => (0,0)
|1 => x1
|S m =>   DWPlusFP  (fst (x m)) (snd  (x  m)) (y_h i)
end.

Notation x_h i := (fst (x i)).
Notation x_l i := (snd (x i)).

(* Algo10.3 *)

Fixpoint sigma_i (i : nat) : R:=
match i with 
| 0 => (y_l 0)
|S m => RN ((sigma_i m) + (y_l i))
end.

(* Algo10.4*)

Definition res10_S := match n with 
0 => (0 ,0)
|1 => y 0
|_ => DWPlusFP (x_h (n-1)) (x_l (n-1)) (sigma_i (n-1))
end.


Notation S_h := (fst res10_S).
Notation S_l := (snd res10_S).

Fact Fsigma_i i: format (sigma_i i).
Proof. 
case:(DWyi i)=>[[Fy_h Fy_l]] yE.
case:i Fy_h Fy_l yE => //==>i Fy_h Fy_l yE; apply/generic_format_round.
Qed.

(* to verify... *)
Hypothesis Hp3: (3 <= p)%Z.

Fact Fy_h i: format (y_h i).
Proof.
by case:(DWyi i); case.
Qed.


Fact Fy_l i: format (y_l i).
Proof.
by case:(DWyi i); case.
Qed.



Fact DWx_i i: (double_word (x_h i) (x_l i)).
Proof.
have F0: format 0 by apply /generic_format_0.
case:i.
  rewrite /=; split=>//.
  by rewrite Rplus_0_r round_0.
elim.
  apply/DW_TwoSum =>//; apply/Fy_h.
move=> i hi.
  case:( @DWPlus.DWPlusFP_correct  _  _  _ choice_sym   Hp3 (x_h i.+1)(x_l i.+1) (y_h (i.+2)))=>//.
by apply/Fy_h.
Qed.

Fact Fx_h i: format (x_h i).
Proof.
case:i=>[| i] /=.
  by apply/generic_format_0.
by case:i => [| i] /=; apply/generic_format_round.
Qed.

Fact Fx_l i: format (x_l i).
Proof.
case:i=>[| i] /=.
  by apply/generic_format_0.
by case:i => [| i] /=; apply/generic_format_round.
Qed.

Fact DW0: double_word  0 0 .
Proof.
have F0 : format 0 by apply/generic_format_0.
by split=>//; rewrite Rplus_0_r round_0.
Qed.

Lemma DW_DWPlusFP xh xl z (DWx: double_word  xh xl) (Fz: format z) :
 double_word (DWPlusFP xh xl z).1  (DWPlusFP xh xl z).2 .
Proof.
 by case:(@DWPlus.DWPlusFP_correct p _ choice _   _ xh xl z).
Qed.


Fact DWres10_S: double_word S_h S_l.
Proof.
rewrite /res10_S.
case:n; first by apply/DW0.
case; first by apply/DWyi.
move=> i; apply/DW_DWPlusFP.
  by apply/DWx_i.
by apply/Fsigma_i.
Qed.

Fact Fres10_S1: format S_h.
Proof. by case:DWres10_S=>[[]]. Qed.
Fact Fres10_S2: format S_l.
Proof. by case:DWres10_S=>[[]]. Qed.


Fact xh_pos i : 0 <= (x_h i).
Proof.
case:i=>[|i].
  by  rewrite /= ;lra.
elim:i=>[| i hi].
  have ->: 0 = RN 0 by rewrite round_0.
  apply/round_le.
  by apply/Rplus_le_le_0_compat; apply/yh_pos.
have ->: (x i.+2) =  DWPlusFP (x i.+1).1 (x i.+1).2 (y_h i.+2).
  clear hi; by case:i.
apply/DWPlusFP_ge_0 =>//; try apply/Fx_h; try apply/Fx_l.
  by apply/Fy_h.
apply/(Rle_trans _  (/ 2 * ulp  (x_h i.+1))).
  by apply/(dw_ulp _ (DWx_i i.+1)).
apply/Rmult_le_compat_l; first lra.
apply/ulp_le.
rewrite !Rabs_pos_eq //.
  by apply/Rmax_l.
by apply/(Rle_trans _ (x_h i.+1))=>//; apply/Rmax_l.
Qed.

Definition res_E:= (sigma (fun i => y_h i + y_l i )  0 (n-1) ).
Notation  Syh:= (sigma (fun i => y_h i)  0 (n-1) ).
Notation  Syl:= (sigma (fun i => y_l i)  0 (n-1) ).

Fact Syh_pos: 0 <= Syh.
Proof. by apply/cond_pos_sum/yh_pos. Qed.

Fact yh0: Syh = 0 -> forall i,  (i <= n-1)%coq_nat ->  y_h i = 0.
Proof.
move=> Syh0; apply/sum_f_pos_0.
  by move=>*; apply/yh_pos.
rewrite -Syh0 /Syh /sigma -minus_n_O.
by apply/PartSum.sum_eq.
Qed.

Fact yl0: Syh = 0 -> forall i,  (i <= n-1)%coq_nat ->  y_l i = 0.
Proof.
move=> SyhE0 i hi.
have yh0: y_h i = 0 by rewrite yh0.
case:(DWyi i)=>[[Fyh Fyl]] yE.
suff: (y_h i + y_l i) = 0 by lra.
suff : RN (y_h i + y_l i ) = 0 by apply/FLX_round_N_eq0.
lra.
Qed.

Fact Syl0: Syh = 0 ->  Syl = 0.
Proof.
move=>*; apply/sum_eq_R0=> i hi ; apply/yl0 =>//; lia.
Qed.

Fact res_E_ge_0 : 0 <= res_E.
Proof. by apply/cond_pos_sum => i; apply/ypos. Qed.

Fact DW_nneg xh xl (DWx: double_word xh xl): 0 <= xh -> 0 <= xh + xl.
Proof.
case:(DWx)=> [[Fxh Fxl] xE].
move=> [xpos|x0]; last first.
  suff->: xl = 0 by lra.
  have := (Rabs_le_inv _ _ (dw_ulp  pGt1 DWx)).
  by rewrite -x0 ulp_FLX_0; lra.
case:(Rle_lt_dec (xh + xl) 0)=> h; last by lra.
have: RN (xh + xl) <= RN 0 by apply/round_le.
by rewrite round_0 -xE; lra.
Qed.

(* Lemma 4.1 *)

Definition d10_2 i := match i with 
| 0 => 0
| 1 => 0
|S m => (x i).1 + (x i).2 - ((x m).1 +  (x  m).2 +  (y_h i)) end.



Fact dE i : d10_2 i.+2 =  
 -(errorDWFP  p choice (x i.+1).1  (x i.+1).2  (y_h i.+2)).
Proof.
have->:  d10_2 i.+2 =  (x i.+2).1 + (x i.+2).2 - ((x i.+1).1 +  (x  i.+1).2 +  (y_h i.+2)) by [].
have->: x i.+2 = DWPlusFP  (x i.+1).1   (x  i.+1).2 (y_h i.+2).
  by [].
by rewrite /errorDWFP /DWPlusFP; ring.
Qed.


Fact dleyh i : Rabs (d10_2 i ) <= Rabs (y_h i).
Proof.
case:i.
rewrite /=  Rabs_R0 ; apply/Rabs_pos.
case.
  by rewrite /=  Rabs_R0 ; apply/Rabs_pos.
move=>i; rewrite dE Rabs_Ropp.
apply/(Rle_trans _ (Rmin _ ( Rabs (y_h i.+2)))); last by apply/Rmin_r.
have yhpos:= yh_pos.
rewrite (Rabs_pos_eq (y_h _))//.
apply:Lange_Rump_DWFP_hyp=>//.
+ by apply/Fy_h.
+ by apply/DWx_i.
by apply/xh_pos.
Qed.


Definition  e (i : nat):= match i with 
|0=>0
| S m => d10_2 i/((x m).1 +  (x  m).2 +  (y_h i)) end.



Lemma deE i : d10_2 i.+1 =  ((x i).1 +  (x  i).2 +  (y_h i.+1)) * (e i.+1).
Proof.
case: i.
  by rewrite /= /Rdiv Rmult_0_l Rmult_0_r.
move=> i.
case:(Req_dec (x_h i.+1 + x_l i.+1 + y_h i.+2) 0)=> den0; first last.
  rewrite /e.
  by field.
have Fx_h:= Fx_h.
have Fy_h:= Fy_h.
have xh_pos := xh_pos.
rewrite den0 Rmult_0_l dE.
apply/Rabs_eq_R0; rewrite Rabs_Ropp errorDWFPE//.
  rewrite TwoSum_correct //  TwoSum_sumE.
  have->: x_l i.+1 + (x_h i.+1 + y_h i.+2 - RN (x_h i.+1 + y_h i.+2)) = -
    RN (x_h i.+1 + y_h i.+2) by lra.
  rewrite Rabs0 // RN_sym // round_round //.
  ring.
by apply/(dw_ulp _ (DWx_i i.+1)).
Qed.


Fact h45_aux : forall i, (0 < i)%N->  x_h i + x_l i =  s (fun i0 : nat => y_h i0) e i .
Proof.
case=>// i _.
have Fx_h:= Fx_h.
have Fy_h:= Fy_h.
elim:i.
  have -> : x 1 = x1 by [].
  rewrite /s.
  have ->: e 0 = 0 by [].
  have->: x1.1 = TwoSum_sum p choice  (y_h 0) (y_h 1) by [].
  have->: x1.2= TwoSum_err p choice  (y_h 0) (y_h 1) by [].
  rewrite TwoSum_correct // TwoSum_sumE.
  have ->: e 1 = 0 by rewrite /= /Rdiv Rmult_0_l.
  ring.
move=> i hi.
have ->:  s (fun i0 : nat => y_h i0) e i.+2 = (y_h i.+2 + 
          s (fun i0 : nat => y_h i0) e i.+1) * (1 + (e i.+2)) by [].
rewrite -hi.
have : d10_2 i.+2 = (x_h i.+2 + x_l i.+2) - 
      ((x i.+1).1 +  (x  i.+1).2 +  (y_h i.+2)) by [].
by rewrite deE;lra.
Qed.


Fact d10_2Ed i: d10_2 i =  d (fun i0 : nat => y_h i0) e i.
Proof.
case:i.
  by rewrite /=; ring.
case.
  rewrite /= /Rdiv Rmult_0_l; ring.
move=> i; rewrite /d -h45_aux // deE; ring.
Qed.

Fact  LR_hyp_yh :
  forall k : nat,
  Rabs (d (fun i : nat => y_h i) e k) <=
  (1 + sigma [eta xi (fun i0 : nat => y_h i0) e] 0 k) * Rabs (y_h k).
Proof.
move=>i;apply/(Rle_trans _ (1*  Rabs (y_h i))); last first.
  apply/Rmult_le_compat_r; first by apply/Rabs_pos.
  suff: 0 <= sigma [eta xi (fun i0 : nat => y_h i0) e] 0 i by lra.
  by apply/cond_pos_sum =>j;apply/xi_ge0.
rewrite Rmult_1_l.
suff-> :  (d (fun i0 : nat => y_h i0) e i) = d10_2 i.
  by apply/dleyh.
by rewrite d10_2Ed.
Qed.

 Fact  b1_lt_b2:  (INR n <= /u) -> (2*(INR n)-1) * u^2 + (INR n + 2) * u^3 + 
                                   (2*INR n - 2) * u^4 + (7*(INR n) - 7) * u^5
                                   + (3*(INR n) -3 )* u^6
                                  < (2 * (INR n) - 1)* u^2 + (INR n + 5)* u^3.
Proof.
set N := INR n.
move=>  hnu.
rewrite !Rplus_assoc;  apply/Rplus_lt_compat_l.
suff: ((2 * N - 2) * u ^ 4 + ((7 * N - 7) * u ^ 5 + (3 * N - 3) * u ^ 6)) < 
         3 * u^3 by lra.
ring_simplify ((2 * N - 2) * u ^ 4 + ((7 * N - 7) * u ^ 5 + (3 * N - 3) * u ^ 6)).
have upos: 0 < u by apply/bpow_gt_0.
have u2pos: 0 < u^2 by nra.
suff: 3 * N * u ^ 6 + 7 * N * u ^ 5 + 2 * N * u ^ 4 <=  3 * u ^ 3 by nra.
have uipos: 0 < /u by apply/Rinv_0_lt_compat.
have nu1: N * u <= 1 by apply/(Rmult_le_reg_r (/u))=>//; field_simplify; lra.
have->:  3 * N * u ^ 6 + 7 * N * u ^ 5 + 2 * N * u ^ 4  = 
       ( 3 * N * u ^ 3 + 7 * N * u ^ 2 + 2 * N * u ) * u^3 by ring.
apply/Rmult_le_compat_r; first  nra.
suff: 3 * u ^ 2 + 7  * u <= 1 by nra.
have u8: u<= /8 by change ( pow (-p)<= pow (-3)); apply/bpow_le; lia.
nra.
Qed.

Fact DWPlusFP0 xh xl z (DWx: double_word xh xl) (Fy : format z) : 
    xh + xl + z =  0 -> DWPlusFP xh xl z = (0,0).
Proof.
move=> xy0.
case: (DWx)=> [[Fxh Fyh]] xE.
rewrite /DWPlusFP TwoSum_correct // TwoSum_sumE.
have ->: xh + z = -xl by lra.
rewrite RN_sym //.
have RNxl: RN xl = xl by rewrite  round_generic.
rewrite /Fast2Sum /F2SumFLX.Fast2Sum /= RNxl .
ring_simplify(xl + (- xl - - xl)); rewrite RNxl.
ring_simplify (-xl + xl).
rewrite round_0; ring_simplify (0 -- xl).
rewrite RNxl; ring_simplify (xl - xl); rewrite round_0 //.
Qed.

Hypothesis Niu: INR n <= /u.
Hypothesis u16: u <= /16.

Definition beta t := (2 * t - 1) * u^2 + (t + 2) * u^3 + (2 * t - 2) * u^4 + ( 7 * t - 7) * u^5 + 
                     (3 * t - 3) * u^6.

Fact beta_gt0 : forall t, 1 <= t -> 0 < beta t.
Proof.
move=>t t1.
rewrite /beta.
have ->:   (2 * t - 1) * u ^ 2 + (t + 2) * u ^ 3 + (2 * t - 2) * u ^ 4 + (7 * t - 7) * u ^ 5 +
  (3 * t - 3) * u ^ 6 =   ((2 * t - 1) + (t + 2) * u + (2 * t - 2) * u ^ 2 + (7 * t - 7) * u ^ 3+
  (3 * t - 3) * u ^ 4)* u^2 by ring.
have upos: 0 < u by apply/bpow_gt_0.
have uipos: 0 < /u  by apply/Rinv_0_lt_compat.
have ult1: u < 1 by rewrite -pow0 /u; apply/bpow_lt; lia.
apply/Rmult_lt_0_compat; last nra.
have ui: forall i , 0 < u^i by move=> i; apply/pow_lt.
do !(apply/Rplus_lt_le_0_compat; try nra).
apply/Rmult_le_pos; nra.
Qed.

(* Lemma 4.1 *) 
Lemma algo_10_res_approx_1_pre: Rabs ((S_h + S_l) - res_E) <=
       ((2*(INR n)-1) * u^2 + (INR n + 2) * u^3 + (2*INR n - 2) * u^4 + 
       (7*(INR n) - 7) * u^5 + (3*(INR n) -3 )* u^6) * res_E.
Proof.
move:n1; case_eq n=>//.
move=> n2.
case_eq n2.
  move=> n2e0 ne1 _; rewrite  /res10_S ne1 {1}/res_E ne1 subnn sigma_eq_arg Rabs0; last ring.
  rewrite -/(beta (INR 1)).
  apply/Rmult_le_pos.
    apply/Rlt_le/beta_gt0; rewrite /=; lra.
  by apply/res_E_ge_0.
move=>n3 n2e ne _.
set N := n3.+2.
have upos: 0 < u by apply/bpow_gt_0.
have uipos: 0 < /u  by apply/Rinv_0_lt_compat.
have N2: 2 <= INR N by change (INR 2 <= INR N); apply/le_INR/leP.
have Fx_h:= Fx_h.
have Fy_h:= Fy_h.
have xh_pos := xh_pos.
have yE: forall i, y_h i = RN (y_h i + y_l i).
  move=> i; by case: (DWyi i).
case: Syh_pos=> Syh0; last first.
  have yl0 := (yl0 (Logic.eq_sym   Syh0)).
  have yh0 := (yh0 (Logic.eq_sym   Syh0)).
  rewrite Rabs0.
    apply/Rmult_le_pos ; last by apply/res_E_ge_0.
    have u2pos: 0 < u^2 by nra.
    have u4pos: 0 < u^4 by nra.
    have u6pos: 0 < u^6 by nra.
    by do ! (apply/Rplus_le_le_0_compat; try nra).
  have ->: res_E = 0.
    apply/sum_eq_R0 =>  i hi .
    by rewrite yh0 ?yl0; try lra; lia.
  have F0: format 0 by apply/generic_format_0.
  have DW0: double_word  0 0 by split;[split=>//| rewrite Rplus_0_r round_0].
  rewrite /res10_S ne.
  have x0 : forall i, (i <= n - 1)%coq_nat ->  x i = ( 0 ,0).
    case=>//; elim.
      move=> hi; rewrite /= /x1.
      rewrite !yh0  /TwoSum //; try lia.
      by rewrite !(Rplus_0_r , round_0 , Rminus_0_r).
    move=> i Hi hi.
    have ->: x i.+2 = (DWPlusFP (x i.+1).1 (x i.+1).2 (y_h i.+2))      by [].
    rewrite Hi /=; last lia.
    rewrite DWPlusFP0f ?yh0 //; lra.
  rewrite -ne x0; last lia.
  have sig0 : forall i,   (i <= n - 1)%coq_nat-> (sigma_i i) = 0.
    elim.
      by rewrite /=; apply/yl0.
    move=> i Hi hi.
    have -> :sigma_i i.+1 =  RN (sigma_i i + y_l i.+1) by [].
    rewrite Hi // ?yl0 // ?(Rplus_0_r, round_0)//; lia.
  rewrite sig0; last lia.
  rewrite   DWPlusFP0f //=; ring.
have h45_1:  Rabs (x_h (n - 1) + x_l (n - 1) - Syh) <= (INR n - 1) * u ^ 2 * Syh.
(* 45 part 1 *)
  have : Rabs (D (fun i : nat => y_h i) e (n -1)) <=
   sigma (fun i : nat => Rabs (e i) ) 0 (n -1) * 
     sigma (fun    i : nat => Rabs (y_h i)) 0 (n -1).
    apply/(Lange_Rump (fun i =>  y_h i) e).
    by apply/ LR_hyp_yh.
  suff ->:   Rabs (D (fun i : nat => y_h i) e (n - 1)) =
    Rabs (x_h (n - 1) + x_l (n - 1) - Syh).
    move=>h.
    apply/(Rle_trans _ (sigma (fun i : nat => Rabs (e i)) 0 (n - 1) *
      sigma (fun i : nat => Rabs (y_h i)) 0 (n - 1)))=>//.
    have->:  sigma (fun i : nat => Rabs (y_h i)) 0 (n - 1) = Syh.
      by apply/PartSum.sum_eq=> i hi; rewrite Rabs_pos_eq // ; apply/yh_pos.
     apply/Rmult_le_compat_r.
       by apply/cond_pos_sum=> i; apply/yh_pos.
    rewrite sigma_first; last by rewrite ne; apply/leP.
    rewrite {1}/e  Rabs_R0 Rplus_0_l.
    apply/(Rle_trans _ (sigma (fun i : nat => u^2) 1 (n-1))).
      apply/sum_Rle=> i _.
      rewrite plusE.
      case:i.
        by rewrite addn0  /= Rabs_pos_eq; nra.
      move=>i.
      rewrite add1n /e dE -/u /errorDWFP Ropp_minus_distr;apply/DWPlusFP_correct=>//.
      by apply/DWx_i.
    rewrite /sigma sum_cte.
    rewrite Rmult_comm; apply/(Rmult_le_compat_r); first nra.
    rewrite minusE S_INR.  
    suff: INR (n - 1 - 1) + 2 <= INR n by lra.
    have ->: 2 = INR 2 by [].
    rewrite -plus_INR.
    apply/le_INR.
    rewrite plusE; apply/leP.
    suff -> : (n - 1 - 1 + 2 = n)%N by [].
    by rewrite ne subSS subn0 subSS subn0 !addnS addn0.
  rewrite /D h45_aux /Syh //.
  by rewrite ne.
have h45: Rabs ((x_h (n-1)) +(x_l (n-1)) - Syh ) <= 
                (INR n -1)* u^2 * ((1 + u) * res_E).
  apply/(Rle_trans _ ((INR n - 1) * u^2 * Syh ))=>//.
  apply/Rmult_le_compat_l.
    suff : 0 <= (INR n -1) by nra.
    have ->: 1 = INR 1 by [].
    rewrite -minus_INR; last by move:(leP n1); lia.
    have -> : 0 = INR 0  by [].
    apply/le_INR; rewrite minusE; lia.
  rewrite /res_E scal_sum.
  apply/sum_Rle =>i hi /=.
  have: Rabs (RN (y_h i + y_l i)  - (y_h i + y_l i)) <= 
          /2 * pow (-p + 1)* Rabs (y_h i + y_l i).
    by apply/relative_error_N_FLX.
  move/Rabs_le_inv=>[ _  ]; rewrite -yE bpow_plus pow1 -/u  
     Rabs_pos_eq; try lra.
  by apply ypos.
(* 46 part 1*)
have hr46_1:   Rabs (sigma_i (n - 1) - Syl) <= 
                  (INR n - 1) * u * sigma (fun i : nat => Rabs (y_l i)) 0 (n - 1).
  pose dyl := fun (i : nat)=>  match i with 
  0 => 0 
  | S m =>  RN (sigma_i m + y_l i) - (sigma_i m + y_l i) end.
  pose e (i : nat):=  dyl i/(sigma_i (i-1) + y_l i).
  have sigmaEs :forall i, sigma_i i =  s (fun i0 : nat => y_l i0) e i .
    elim.
      by rewrite /e  /= /Rdiv Rmult_0_l  Rplus_0_r Rmult_1_r.
    move=>i hi.
    have ->: sigma_i i.+1 = RN (sigma_i i + y_l i.+1)  by [].
    rewrite siE -hi /d.
    suff:   RN (sigma_i i + y_l i.+1) -(y_l i.+1 + sigma_i i )=
        (y_l i.+1 + s (fun i0 : nat => y_l i0) e i) * e i.+1 by lra.
    rewrite -hi Rplus_comm; 
      case:(Req_dec   (y_l i.+1 + sigma_i i)  0).
      move-> ; rewrite Rmult_0_l round_0; ring.
    move=>hn0.
    rewrite /e Rplus_comm.
    suff->: (i.+1 - 1 = i)%nat.
      rewrite /dyl ;field.
      lra.
    by rewrite subSS subn0.
  have LR_hyp : 
   forall k : nat,
   Rabs (d (fun i : nat => y_l i) e k) <=
   (1 + sigma [eta xi (fun i0 : nat => y_l i0) e] 0 k) * Rabs (y_l k).
    case.
      rewrite d0E sigma_eq_arg Rabs0; last by rewrite /e /= /Rdiv Rmult_0_l; ring.
      apply/Rmult_le_pos; last by apply/Rabs_pos.
      apply/Rplus_le_le_0_compat; first lra.
      by apply/xi_ge0.
    move=> i.
    rewrite diE -sigmaEs.
    have ->:  e i.+1 = dyl i.+1 /(sigma_i i + y_l i.+1).
      by rewrite /e subSS subn0.
    apply/(Rle_trans _ ( Rabs (y_l i.+1))).
      case:(Req_dec (y_l i.+1 + sigma_i i) 0)=> [->| hi].
        by rewrite Rmult_0_l Rabs_R0; apply/Rabs_pos.
      have->:  ((y_l i.+1 + sigma_i i) * (dyl i.+1 / (sigma_i i + y_l i.+1))) =
                (dyl i.+1).
        by field; lra.
      rewrite /dyl.
      by apply/plus_error_le_r; try apply/Fsigma_i; apply/Fy_l.
    set sgm := sigma _ _ _.
    suff : 0 <= sgm by move:(Rabs_pos  (y_l i.+1)); nra.
    apply/cond_pos_sum=>j.
    by apply/xi_ge0.
  have : Rabs (D (fun i : nat => y_l i) e (n -1)) <=
   sigma (fun i : nat => Rabs ( e i)) 0 (n -1) * 
     sigma (fun    i : nat => Rabs (y_l i)) 0 (n -1).
    by apply/(Lange_Rump (fun i =>  y_l i) e)=>//.
  rewrite /D -sigmaEs /Syl.
  move=> h.
  apply/(Rle_trans _       (sigma (fun i : nat => Rabs (e i)) 0 (n - 1) *
      sigma (fun i : nat => Rabs (y_l i)) 0 (n - 1)))=>//.
  apply/Rmult_le_compat_r.
    by     apply/cond_pos_sum=>j; apply/Rabs_pos.
  rewrite sigma_first; last first.
    by apply/lt_minus_O_lt; move/leP: n1; lia.
  have ->: ( e 0%N) = 0 by rewrite /e /dyl /Rdiv  Rmult_0_l.
  rewrite Rabs_R0 Rplus_0_l.
  have eu: forall i , Rabs (e i) <= u.
    case.
      by rewrite   /e /dyl /Rdiv Rmult_0_l Rabs_0; lra.
    move=> i .
    rewrite  /e.
    have ->: dyl i.+1 = RN (sigma_i i + y_l i.+1) - (sigma_i i + y_l i.+1) by [].
    rewrite subSS subn0.
    case:(Req_dec  (sigma_i (i) + y_l i.+1) 0)=> d0.
      rewrite d0 round_0 Rminus_0_r /Rdiv Rmult_0_l Rabs_R0 ; lra.
    apply/(Rmult_le_reg_r (Rabs (sigma_i i + y_l i.+1))).
      by apply/Rabs_pos_lt.
    rewrite -Rabs_mult Rmult_assoc Rinv_l // Rmult_1_r.

    by apply/error_plus_le=>//; try apply/Fy_l; apply/Fsigma_i.
  apply/(Rle_trans _ ( sigma (fun i : nat => u) 1 (n - 1))).
    apply/sum_Rle.
    by move=> *; apply/eu.
  rewrite /sigma sum_cte.
  rewrite Rmult_comm; apply/(Rmult_le_compat_r); first lra.

  have->:  (n - 1 -  1= n-2)%coq_nat.
    by rewrite ne.
  rewrite S_INR /=.
  suff:  INR (n - 2)%coq_nat  <= INR n  -2 by lra.
  rewrite minus_INR /=; first lra.
  by apply/leP ; rewrite ne //=.
have h46: Rabs ( sigma_i (n -1) - Syl) <= (INR n -1) * u^2 * res_E.
  apply/(Rle_trans _
  ((INR n -1)*u* (sigma (fun i : nat => Rabs (y_l i)) 0 (n - 1))))=>//.
  have->:(INR n - 1) * u ^ 2 * res_E = (INR n - 1) * u  * ( u * res_E) by ring.
  apply/Rmult_le_compat_l.
    suff : 0 <= (INR n -1) by nra.
    have ->: 1 = INR 1 by [].
    rewrite -minus_INR =>//; last by move:(leP n1); lia.
    have -> : 0 = INR 0  by [].
    apply/le_INR.
    move:(leP n1);lia.
  rewrite /res_E scal_sum.
  apply/sum_Rle =>i hi.
  rewrite plus_0_l.  
  have {1}->: y_l i = - (RN (y_h i + y_l i) -  (y_h i + y_l i)) 
    by rewrite -yE; ring.
  rewrite  Rabs_Ropp.
  have-> :  (y_h i + y_l i) * u = 
       /2 * pow (- p + 1)*  Rabs (y_h i + y_l i). 
    rewrite bpow_plus pow1 -/u Rabs_pos_eq; try lra.
    by apply/ypos.
  by apply/relative_error_N_FLX.
have h47: Rabs ( ((x_h (n-1)) +(x_l (n-1))+   (sigma_i (n -1))) - res_E ) <= 
                               (2 * (INR n  -1)* u^2 + (INR n - 1)* u^3)* res_E.
  have->: (x_h (n - 1) + x_l (n - 1) + sigma_i (n - 1) - res_E) = 
    ((x_h (n-1)) +(x_l (n-1)) - Syh ) +  ( sigma_i (n -1) - Syl) .
    suff ->: res_E = Syh + Syl by ring.
    by rewrite /res_E /Syh /Syl -sum_plus; apply/PartSum.sum_eq.
  apply/(Rle_trans _ (Rabs (x_h (n - 1) + x_l (n - 1) - Syh)  + 
       Rabs (sigma_i (n - 1) - Syl))).
    by apply/Rabs_triang.
  by lra.
have h48_0: Rabs (sigma_i (n - 1)) <= (u + (INR n  - 1) * u^2) * Syh.
  apply/(Rle_trans _  ((1 + (INR n - 1)* u)* 
                      (sigma (fun i : nat => Rabs (y_l i)) 0 (n - 1)))).
    move: (Rabs_triang_inv (sigma_i (n - 1)) Syl).
    move/(Rplus_le_compat_r (Rabs Syl)).
    rewrite Rplus_comm Rplus_minus=>h.
    apply/(Rle_trans _ _ _  h).
    rewrite (Rplus_comm 1) Rmult_plus_distr_r; apply/Rplus_le_compat=>//.
    by rewrite Rmult_1_l; apply/Rsum_abs.
  have->: (u + (INR n - 1) * u ^ 2)  = (1 + (INR n - 1) * u)* u by ring.
  rewrite Rmult_assoc ; apply/Rmult_le_compat_l.
     suff:  1 <= (INR n ) by nra.
     by change (INR 1 <= INR n); apply/le_INR; move/leP:n1; lia.
  rewrite scal_sum;apply/sum_Rle=> i _.
  by rewrite Rmult_comm; apply/yluyh.
have denpos: 0 < (1 - (INR n - 1) * u ^ 2).
  suff Nu: (INR n) * u <= 1 by nra.
  apply/(Rmult_le_reg_r (/u))=>//.
  by rewrite Rmult_assoc Rinv_r; lra.
have h1 :0 < (1 - (INR n - 1) * u ^ 2) * Syh.
  by apply/Rmult_lt_0_compat.
have : (1 - (INR n - 1) * u ^ 2) *Syh <=  x_h (n - 1) + x_l (n - 1).
  by move/Rabs_le_inv:h45_1; lra.
move=>h2.
have xpos: 0 < x_h (n - 1) + x_l (n - 1) by lra.
have xposi : 0 < /(x_h (n - 1) + x_l (n - 1)) by apply/Rinv_0_lt_compat.
move/(Rinv_le _ _ h1) :(h2) => h3.
have :  Rabs (sigma_i (n - 1)) /(x_h (n - 1) + x_l (n - 1))<= 
   (u + (INR n - 1) * u ^ 2) * Syh / ((1 - (INR n - 1) * u ^ 2) * Syh).
  apply/Rmult_le_compat=>//.
    by apply/Rabs_pos.
  by apply/Rlt_le/Rinv_0_lt_compat; lra.
have->:  (u + (INR n - 1) * u ^ 2) * Syh / ((1 - (INR n - 1) * u ^ 2) * Syh) =
 (u + (INR n - 1) * u ^ 2) / (1 - (INR n - 1) * u ^ 2)  by field; lra.
move=> h48_1.
pose f:= fun b => (u + (b - 1) * u ^ 2) / (1 - (b - 1) * u ^ 2).
have hf: forall b c,  1<= b <= c  -> c <=  /u -> f b <= f c.
  move=> b c bc cu.
  rewrite /f.
  have dc:  0 < (1 - (c  - 1) * u ^ 2).
    ring_simplify (1 - ( c- 1) * u ^ 2).
    suff:  c * u ^ 2 <  u ^ 2 + 1 by lra.
    have cu1: c * u <= 1.
      apply/(Rmult_le_reg_r (/u)); rewrite // Rmult_assoc Rinv_r; lra.
    nra.
  have db: 0 < 1 - (b - 1) * u ^ 2 by nra.
  set b1:= 1 - _.
  set c1 := 1 - _.
  apply/(Rmult_le_reg_r (b1 * c1 )); first  by rewrite /c1 /b1 ; nra.
  rewrite !Rmult_assoc.
  have ->: (/ b1 * (b1 * c1)) = c1 by field; rewrite /b1; lra.
  have ->:  (/ c1 * (b1 * c1)) = b1  by field; rewrite /c1; lra.
  by apply/(Rmult_le_compat); rewrite /c1 /b1; nra.
have: (u + (INR n - 1) * u ^ 2) / (1 - (INR n - 1) * u ^ 2)<=
          (u + (/u - 1) * u ^ 2) / (1 - (/u  - 1) * u ^ 2).
  move: (hf (INR n)); rewrite /f; apply; try lra.
  split; try lra.
  by change (INR 1 <= INR n); apply/le_INR; apply/leP; rewrite ne.
have hu : 0 < 1 - u + u ^ 2  by nra.
have -> : (u + (/ u - 1) * u ^ 2) / (1 - (/ u - 1) * u ^ 2)= (2*u - u^2)/(1 - u + u^2).
  by field; lra.
have :  (2 * u - u ^ 2) / (1 - u + u ^ 2) <= 2*u + u^2.
  apply/(Rmult_le_reg_r (1 - u + u ^ 2))=>//.
  rewrite Rmult_assoc Rinv_l ?Rmult_1_r; nra.
move=>*.
set d:= (x_h (n - 1) + x_l (n - 1)+sigma_i (n - 1) ).
have haux:  (-2 * u - u ^ 2) * (x_h (n - 1) + x_l (n - 1)) <= sigma_i (n - 1).
  suff:  Rabs (sigma_i (n - 1)) <=  (2*u + u^2) * (x_h (n - 1) + x_l (n - 1)).
    move/Rabs_le_inv; lra.
  apply/(Rmult_le_reg_r (/(x_h (n - 1) + x_l (n - 1))))=>//.
  rewrite Rmult_assoc Rinv_r ?Rmult_1_r; last lra.
  lra.
have: Rabs (sigma_i (n - 1)) / (x_h (n - 1) + x_l (n - 1)) <= 1.
  apply/(Rle_trans _  ((u + (INR n - 1) * u ^ 2) / (1 - (INR n - 1) * u ^ 2)))=>//.
  apply/(Rle_trans _    ((2 * u - u ^ 2) / (1 - u + u ^ 2)))=>//.
  apply/(Rmult_le_reg_r (1 -u + u^2))=>//.
  rewrite Rmult_assoc Rinv_l ?Rmult_1_r ?Rmult_1_l; last lra.
  suff: 4* u <= 1 by nra.
  lra.
move=> h.
have:  Rabs (sigma_i (n - 1)) <= (x_h (n - 1) + x_l (n - 1)).
  apply/(Rmult_le_reg_r (/(x_h (n - 1) + x_l (n - 1))))=>//.
  rewrite  Rinv_r ?Rmult_1_r //.
  lra.
move/Rabs_le_inv=> h_aux.
have dpos: 0 <= d.
  rewrite /d; lra.
have prop3_2:  Rabs (S_h + S_l - (x_h (n - 1) + x_l (n - 1) + sigma_i (n - 1)  )) 
               <= (u^2 + 3 * u^3 ) *  (x_h (n - 1) + x_l (n - 1)+sigma_i (n - 1) ).
  case : dpos =>[ dpos| d0]; last first.
    rewrite /res10_S DWPlusFP0 /= ?Rabs0 -/d -?d0 //; first lra.
    + rewrite ne /=; ring.
    + by apply/DWx_i.
    by apply/Fsigma_i.
  rewrite -/d.
  have->:   Rabs (S_h + S_l - d) =  (relative_errorDWFP p choice 
                       (x_h (n - 1))  (x_l (n - 1))  (sigma_i (n - 1))) * Rabs  d.
    rewrite rel_errorE Rmult_assoc -Rabs_mult Rinv_l; last by rewrite -/d; lra.
    rewrite (Rabs_pos_eq 1) ?Rmult_1_r; last lra.
    rewrite -Ropp_minus_distr Rabs_Ropp.
    congr Rabs.
    rewrite /d /errorDWFP; congr Rminus.
    by  rewrite  /res10_S ne.
  rewrite Rabs_pos_eq; last lra.
  apply/Rmult_le_compat_r; first lra.
  apply/DWPFP_bound_small_y=>//.
    by apply/Fsigma_i.
  by apply/DWx_i.
rewrite -/u //.
have h48_2:  x_h (n - 1) + x_l (n - 1) + sigma_i (n - 1)  <=
  (1 + 2 * (INR n - 1) * u ^ 2 + (INR n - 1) * u ^ 3) * res_E.
  by move/Rabs_le_inv:  h47; lra.
have h48:  Rabs (S_h + S_l - (x_h (n - 1) + x_l (n - 1) + (sigma_i (n - 1)))) <=
     (u ^ 2 + 3 * u ^ 3) *  (1 + 2 * (INR n - 1) * u ^ 2 + (INR n - 1) * u ^ 3) * res_E.
  apply/(Rle_trans _  
        ((u ^ 2 + 3 * u ^ 3) * (x_h (n - 1) + x_l (n - 1) + sigma_i (n - 1))))=>//.
  rewrite Rmult_assoc; apply/Rmult_le_compat_l=>//.
  nra.
have->:  (S_h + S_l - res_E)  = S_h + S_l - (x_h (n - 1) + x_l (n - 1) + 
           sigma_i (n - 1))+
           (x_h (n - 1) + x_l (n - 1) + sigma_i (n - 1) -res_E) by ring.
apply/(Rle_trans _ _  _  (Rabs_triang _ _)).
rewrite /N -ne; lra.
Qed.

Lemma DW_DWPlusFP_E xh xl z (DWx: double_word  xh xl) (Fz: format z) :
      (DWPlusFP xh xl z).1 = RN ( (DWPlusFP xh xl z).1 + (DWPlusFP xh xl z).2 ).
Proof.  by case: (@DW_DWPlusFP  xh xl z). Qed.

Fact res10_S_pos: 0 <= S_h.
Proof.
move: n1; case_eq n =>//.
move=> n2; case_eq n2.
  by rewrite /res10_S => n20 -> _;apply/yh_pos.
move=>n3 n2e ne _.
move/Rabs_le_inv: algo_10_res_approx_1_pre.
set beta := (2 * (INR n) - 1) * u^2 + ((INR n) + 2) * u^3 + (2 * (INR n) - 2) * u^4 + ( 7 * (INR n) - 7) * u^5 + (3 * (INR n) - 3) * u^6.
move=> h.
have->: 0 = RN 0 by rewrite round_0.
have ->: S_h = RN (S_h + S_l).
  rewrite /res10_S ne.
  apply/DW_DWPlusFP_E.
    by apply/DWx_i.
  by apply/Fsigma_i.
apply/round_le.
apply/(Rle_trans _ ((1 -beta)*res_E)); try lra.
move:res_E_ge_0.
suff L: beta <= 1 by nra.
rewrite /beta.
have upos: 0 < u by apply/bpow_gt_0.
have u2pos: 0 < u^2 by nra.
have: (1 < INR n) .
  by change (INR 1 < INR n); rewrite ne;  apply/lt_INR/leP.
have : 0 <= u^4 by nra.
have : 0 <= u^3 by nra.
move=>*.
have: 0 <= u^5.
  have-> : u^5 = u^4 * u by ring.
  nra.
move=>*.
field_simplify.
suff :  3 * INR n * u ^ 6 + 7 * INR n * u ^ 5 + 2 * INR n * u ^ 4 + INR n * u ^ 3 +
  2 * INR n * u ^ 2  + 2 * u ^ 3<= 1 by nra.
have->:   3 * INR n * u ^ 6 + 7 * INR n * u ^ 5 + 2 * INR n * u ^ 4 + INR n * u ^ 3 +
  2 * INR n * u ^ 2 + 2 * u ^ 3 =   (3  * u ^ 5 + 7 *  u ^ 4 + 2 * u ^ 3 + u ^ 2 +
  2  * u ) *(INR n * u) +   2 * u ^ 3 by ring.
have nu: INR n * u <= 1.
  apply/(Rmult_le_reg_r (/u)).
    lra.
  rewrite Rmult_assoc Rinv_r; lra.
suff:  (3 * u ^ 5 + 7 * u ^ 4 + 4 * u ^ 3 + u ^ 2 + 2 * u)   <= 1 by nra.
have->: 3 * u ^ 5 + 7 * u ^ 4 + 4 * u ^ 3 + u ^ 2 + 2 * u = 
        (3 * u ^ 4 + 7 * u ^ 3+ 4 * u ^ 2+ u  + 2 )* u by ring.
suff: 3 * u ^ 4 + 7 * u ^ 3 + 4 * u ^ 2 + u  <= 14 by nra.
have->: 3 * u ^ 4 + 7 * u ^ 3 + 4 * u ^ 2 + u = (3 * u ^ 3 + 7 * u ^ 2 + 4 * u + 1)* u.
  ring.
apply/(Rle_trans _ (15/2) ); try  lra.
apply/Rmult_le_compat; nra.
Qed.

End Algo_10_pre.

Section Algo_10.

Variable n: nat.
Hypothesis n1: (2 <= n)%N. 

Variable a : nat -> R.
Hypothesis Fa : forall i,  format (a i).

Notation y := (fun i => (Fast2Mult (a i) ( a i))).

Notation y_h i := (fst (y i)).
Notation y_l i := (snd (y i)).

Fact Fyh i : format (y_h i).
Proof. by apply/generic_format_round. Qed.

Fact Fyl i : format (y_l i).
Proof. by apply/generic_format_round. Qed.

Fact DWyi i: double_word p choice (y_h i) (y_l i).
Proof.
split; [split|]; try apply/Fyh; try apply/Fyl.
by rewrite   -F2Mult_correct /Fast2Mult /=.
Qed.

Fact yh_pos i: 0 <= (y_h i).
Proof.
have ->: 0 = RN 0 by rewrite round_0.
apply/round_le.
move:(Rle_0_sqr (a i)).
by rewrite /Rsqr.
Qed.

Notation  S_h := (fst (res10_S n y )).
Notation S_l := (snd (res10_S n y)).
Notation res_E := (sigma (fun i => Rsqr (a i)) 0 (n-1)).


(* to verify... *)
Hypothesis Hp3: (3 <= p)%Z. 

(* Lemma 4.1 *) 
Lemma algo_10_res_approx_1 (Niu: INR n <= /u) (u16: u <= /16):  Rabs ((S_h + S_l) - res_E) 
                     <=
       ((2*(INR n)-1) * u^2 + (INR n + 2) * u^3 + (2*INR n - 2) * u^4 + 
       (7*(INR n) - 7) * u^5 + (3*(INR n) -3 )* u^6) * res_E.
Proof.
have npos : (0 < n)%N by apply/leP; move/leP:n1; lia.
move:(algo_10_res_approx_1_pre n npos y).
have->: Algo_10_11.res_E n y = res_E.
  apply/PartSum.sum_eq=>i hi.
  rewrite -F2Mult_correct/Rsqr ; ring.
apply=>//.
  by apply/DWyi.
by apply/yh_pos.
Qed.

Lemma res_approx_2 : (res_E <> 0) -> (INR n <= /u) -> (u <= /16) -> 
       Rabs ((S_h + S_l) - res_E) <
       ((2*(INR n)-1) * u^2 + (INR n + 5) * u^3 )* res_E.
Proof.
move=> res_pos inu u16.
apply/(Rle_lt_trans _ (((2 * INR n - 1) * u ^ 2 + (INR n + 2) * u ^ 3 +
        (2 * INR n - 2) * u ^ 4 + (7 * INR n - 7) * u ^ 5 +
        (3 * INR n - 3) * u ^ 6) * res_E)).
  by apply/algo_10_res_approx_1.
apply/Rmult_lt_compat_r.
  suff : 0 <= res_E by lra.
  by apply/sum_f_pos=>i hi; apply/Rle_0_sqr.
by apply:b1_lt_b2.
Qed.
End Algo_10.

Section Algo_11.

Lemma sum_sigma: forall f n , sum_f_R0 f n = sigma f 0 n.
Proof.
move=> f n ; rewrite  /sigma -minus_n_O.
apply/PartSum.sum_eq => i _.
by rewrite plus_O_n.
Qed.

Lemma sum_decomp:  forall k m f, (1 <= k)%nat ->   (1 <= m)%nat -> 
  sum_f_R0 (fun l : nat => sigma f  (m * l) (m * (l + 1) - 1)) (k - 1) =
  sum_f_R0  f (k*m - 1).
Proof.
move=>k m f k1 m1.
case: k k1=>//= k _.
rewrite !sum_sigma.
elim: k.
  by rewrite subnn sigma_eq_arg muln0 mul1n add0n muln1.
move=> k.
rewrite subSS subn0 => Hk.
rewrite subSS subn0 sigma_last; last lia.
rewrite -pred_Sn Hk Rplus_comm.
have km0:   (0 < k.+1 * m)%coq_nat.
  apply/leP.
  clear Hk.
  by move: m1; case:m=>//=.
have->: ((m * k.+1) = (k.+1 * m - 1).+1)%nat.
  by rewrite mulnC subn1  -(S_pred _ 0).
rewrite -sigma_split.
  + by rewrite mulnC addn1.
  lia.
rewrite addn1.
apply/lt_S_n.
 have mk2: (k.+1 * m < m * k.+2)%coq_nat.
  rewrite (mulnC m).
  by apply/leP; rewrite ltn_mul2r m1 /=.
rewrite !subn1 -!S_pred_pos //;lia.
Qed.


Lemma L4_2 : forall n k m , (2 <= INR k) -> (2 <= INR m) -> INR k* INR m = INR n -> 
                            sqrt (INR n) + INR n/(sqrt (INR n)) <= INR k + INR m <= INR n/2 +  2.
Proof.
move=> n0 k m k2 m2 kmn.
set N := INR n0; set K := INR k; set M := INR m.
set rN := sqrt N.
have N4: 4 <= N by rewrite /N; nra.
have krnpos: 0 < K * rN.
  apply/Rmult_lt_0_compat; try (rewrite /K ; lra).
  by apply/sqrt_lt_R0; lra.
have krnipos: 0 < /(K * rN) by apply/Rinv_0_lt_compat.
set f := fun a =>  a + N/a.
have hf:  forall b, (2 <= b)  -> f K - f b = /(K*b) * (N - K*b) * (b - K).
  move=> b b2; rewrite/f /K; field;  lra.
have ->: M = N/K.
  apply/(Rmult_eq_reg_r K); last by rewrite /K; lra.
  rewrite Rmult_assoc Rinv_l /K /N /M;   lra.
have->:  K + N / K = f K by [].
have ->: rN + N / rN = f rN by [].
have ->: N / 2 + 2 = f 2 by rewrite Rplus_comm.
split.
  suff: 0 <= f K -f rN by lra.
  rewrite hf.
    apply/(Rmult_le_reg_l (K * rN))=>//.
    have->: K * rN * (/ (K * rN) * (N - K * rN) * (rN - K)) = (N - K * rN) * (rN - K).
      field;  nra.
    rewrite !Rmult_0_r.
    case: (Rle_lt_dec rN K)=> KrN.
      suff: (N <=  K * rN)  by nra.
      have -> : N = sqrt N * sqrt N.
        by rewrite sqrt_sqrt; lra.
      rewrite -/rN; apply/Rmult_le_compat_r=>//.
      by apply/sqrt_ge_0.
    suff: (K * rN <= N)  by nra.
    have -> : N = sqrt N * sqrt N.
        by rewrite sqrt_sqrt; lra.
    rewrite -/rN; apply/Rmult_le_compat_r; last lra.
    by apply/sqrt_ge_0. 
  have ->: 2 = sqrt (2 * 2).
    rewrite sqrt_square //; lra.
  apply/sqrt_le_1_alt; rewrite /N; nra.
suff: f K - f 2 <= 0 by lra.
rewrite hf; last lra.
apply/(Rmult_le_reg_l (K * 2)); try (rewrite/K; lra).
have->: K * 2 * (/ (K * 2) * (N - K * 2) * (2- K)) = (N - K * 2) * (2 - K).
  field; nra.
rewrite !Rmult_0_r.
suff: ( K * 2 <= N ) by (rewrite /K ;  nra).
rewrite /N /K -kmn;  nra.
Qed.

Variable n: nat.
Hypothesis n1: (2 <= n)%N. 
Variables  k m: nat.
Hypothesis nkm: (k * m = n)%nat.
Variable a : nat -> R.
Hypothesis Fa : forall i,  format (a i).
Definition res_aux  j :=  sigma (fun i : nat => (a i)²) (m * j)%nat (m * (j + 1) - 1)%nat.

Notation TwoSum := (TwoSum p choice).
Notation DWPlusFP := (DWPlusFP p choice).

Fact  res_aux_E  j   : res_aux  j   =  sigma (fun i : nat => (a (m*j + i)%nat )²) 0 (m   - 1)%nat.
Proof.
rewrite /res_aux {1}/sigma.
congr sum_f_R0.
rewrite mulnDr muln1 minusE subn0.
case:m =>//= m0; rewrite addnS !subn1 -!pred_Sn.
by rewrite -{2}(addn0 (m0.+1 * j)%N) subnDl subn0.
Qed.


(* Notation y := (y a). *)
Notation y := (fun i => (Fast2Mult (a i) ( a i))).

(*Algo 11.1 *)
Notation Z j := (res10_S  m %nat 
           ( fun i => y (m * j + i)%nat)).

Notation Z_h j := (fst (Z j)).
Notation Z_l j := (snd (Z j)).

(* Algo 11.2 *)
Definition Sig1 := (TwoSum (Z_h 0) (Z_h 1)).

Fixpoint Sig (i : nat) : R *  R :=
match i with 
|0 => (0,0)
|1 => Sig1
|S m =>   DWPlusFP  (fst (Sig  m)) (snd  (Sig   m)) (Z_h i)
end.
Notation Sig_h i := (fst (Sig i)).
Notation Sig_l i := (snd (Sig i)).

Hypothesis Hp3: (3 <= p)%Z.

Fact DWSig:forall i, double_word p choice (Sig_h i) (Sig_l i).
Proof.
case; first by apply/DW0.
elim.   
  apply/DW_TwoSum=>//.
    apply/Fres10_S1=>//.
    by move=>i; apply/DW_F2P.
  apply/Fres10_S1=>//.
  move=>i ; by apply/DW_F2P.
move=>i hi.
apply/DW_DWPlusFP=>//.
apply/Fres10_S1=>//.
by  move=>j; apply/DW_F2P.
Qed.

(* Algo 11.3 *)


Fixpoint tau_i (i : nat) : R:=
match i with 
| 0 => (Z_l 0)
|S m => RN ((tau_i m) + (Z_l i))
end.


Fact Ftau_i: forall i, format (tau_i i).
Proof.
case.
  rewrite /=.
  case:(DWres10_S m (fun i : nat => Fast2Mult (a (m * 0 + i)) (a (m * 0 + i)))).
  + by move=>i; apply/DW_F2P.
  + lia.
  by case.
by move  => i; apply/generic_format_round.
Qed.

(* Algo11.4*)

Hypothesis hp3 : (3 <= p)%Z.

Definition res11_S := match k with  
0 => (0 ,0)
|1 => Z 0
 |_ => DWPlusFP (Sig_h (k-1)) (Sig_l (k-1)) (tau_i (k-1))
end.

Fact res11_S1: (k = 1)%nat -> res11_S = Z 0.
Proof.
by move=> k1; rewrite /res11_S k1 .
Qed.

Fact res11_Sn: (1 < k )%nat -> res11_S = 
        DWPlusFP (Sig_h (k-1)) (Sig_l (k-1)) (tau_i (k-1)) .
Proof.
by rewrite /res11_S; case:k=>//; case.
Qed.



Notation S_h := (fst res11_S).
Notation S_l := (snd res11_S).


Fact DW_res11_S : double_word p choice S_h S_l.
Proof.
rewrite /res11_S.
case:k; first by  apply/DW0.
case.
  rewrite /res10_S.
  case:m;  first by  apply/DW0.
  case; first by apply/DW_F2P.
  move=>n0; apply/DW_DWPlusFP=>//.
    apply/DWx_i=>//.
    by move=>i; apply/DW_F2P.
  apply/Fsigma_i.
  by move=>i; apply/DW_F2P.
move=>n0; apply/DW_DWPlusFP=>//.
  by apply/DWSig.
by apply/generic_format_round.
Qed.


Hypothesis k1: (1 <= k)%nat.
Hypothesis m1: (1 <= m)%nat.






Definition res11_E:= sigma (fun i =>  Rsqr ( a i)) 0 ( n -1).


Fact DWPlusFPD0 xh xl (DWx: double_word p choice xh xl):  DWPlusFP xh xl 0 = (xh, xl).
Proof.
case:(DWx)=>[[Fxh Fxl]] xE.
have F0: format 0 by apply/generic_format_0.
rewrite /DWPlusFP  /=.
rewrite TwoSum_correct=>//.
rewrite TwoSum_sumE Rplus_0_r round_generic //.
have->:(xl + (xh - xh)) = xl by ring.
rewrite round_generic //.
rewrite /Fast2Sum/F2SumFLX.Fast2Sum.
rewrite -xE.
have->: xh -xh = 0 by ring.
by rewrite round_0 Rminus_0_r round_generic.
Qed.


Fact res10_S1  f : res10_S 1 f = f 0%nat.
Proof. by rewrite /res10_S. Qed.

Fact res10_Sn i f : ( 1 < i )%nat -> res10_S i f = 
     DWPlusFP (x f (i - 1)).1 (x f (i - 1)).2 (sigma_i f (i  - 1)).
Proof.
by rewrite /res10_S;case:i=>//;case. 
Qed.

Fact Prop4_4_m1: (m = 1)%nat  -> (res10_S  n y ) =  res11_S /\ (res_E n y) = res11_E.
Proof.
move=> me1.
have ZE: forall i,  Z i  =  y i.
  by move=>i; rewrite me1 res10_S1 addn0 mul1n.
split; last by apply/PartSum.sum_eq=> i hi; rewrite -F2Mult_correct /Rsqr.
have SigE: forall i , Sig i = x y i.
  elim=>//.
  case=>//.
    by move=> _; rewrite /Sig /Sig1  ZE me1 res10_S1 /x /x1.
  move=> i hi.
  have ->: Sig i.+2 = DWPlusFP (Sig i.+1).1 (Sig i.+1).2 (Z_h i.+2) by [].
  by rewrite me1 res10_S1 addn0 mul1n hi.
have k2: (1 < k)%nat.
  move: nkm k1  n1; rewrite me1 muln1.
  case: k=>//.
  case=>//.
  by move <-.
rewrite res10_Sn //  res11_Sn //  /res_E /res11_E.
have ->: (k = n)%nat by rewrite -nkm me1 muln1.
rewrite SigE.
have sto: forall i, sigma_i y i = tau_i i.
  elim=>//.
    by rewrite /=  me1 res10_S1 mul1n addn0//=.
  case=>//.
    by move=> _;  rewrite /sigma_i /tau_i me1 !res10_S1 !addn0 !mul1n.
  move=> i hi.
  have-> : tau_i i.+2 = RN (tau_i i.+1 + Z_l i.+2) by [].
  rewrite me1 res10_S1 addn0 mul1n.
  have ->:  sigma_i y i.+2 = RN (sigma_i y i.+1 + (y i.+2).2) by [].
  by rewrite hi.
by rewrite sto.
Qed.

Fact Prop4_4_k1: (k = 1)%nat  -> (res10_S  n y ) =  res11_S /\ 
                      (res_E n y) = res11_E.
Proof.
move=> ke1.
rewrite res11_S1 //  res10_Sn // muln0.
have ->: (m = n)%nat by rewrite -nkm ke1 mul1n.
rewrite res10_Sn  //  /res_E /res11_E.
split=>//.
apply/PartSum.sum_eq => i hi.
by rewrite plus_O_n /Rsqr  F2Mult_correct.
Qed.

Fact  Prop4_4_k1_S: (k = 1)%nat  -> res11_S = (res10_S  n y ).
Proof.
by move=> ke1; case:Prop4_4_k1. 
Qed.

Fact  Prop4_4_k1_E: (k = 1)%nat  -> res11_E = (res_E n y).
   Proof.
by move=> ke1; case:Prop4_4_k1. 
Qed.


Fact Prop4_4_m1_S: (m = 1)%nat  -> res11_S = (res10_S  n y ).
Proof.
by move=> me1; case:Prop4_4_m1. 
Qed.


Fact Prop4_4_m1_E: (m = 1)%nat  ->  res11_E = (res_E n y).
Proof.
by move=> me1; case:Prop4_4_m1. 
Qed.

(* Lemma 4.3 *)
Lemma algo_11_res_approx_1 (kiu: INR k <= /u) (miu: INR m <= /u)  (u16: u <= /16):  
               Rabs ((S_h + S_l) - res11_E) <= 
                   (beta (INR k) + beta (INR m) + (beta (INR  k)) * (beta (INR m)))* res11_E.
Proof.
have upos: 0 < u by apply/bpow_gt_0.
have uipos: 0 < /u  by apply/Rinv_0_lt_compat.
move:k1; rewrite leq_eqVlt.
case/orP.
  move =>/eqP hk.
  case:Prop4_4_k1 =>//.
  move=>  <- <-.
  apply/(Rle_trans _ (((2 * INR n - 1) * u ^ 2 + (INR n + 2) * u ^ 3 + (2 * INR n - 2) * u ^ 4 +
   (7 * INR n - 7) * u ^ 5 + (3 * INR n - 3) * u ^ 6) * 
  res_E n y)).
    apply/algo_10_res_approx_1_pre=>//.
    + by apply/(@ltn_trans 1 0 n).
    + by move=>i; apply/DW_F2P.
    + move => i; have->: 0 = RN 0 by rewrite round_0.
      by apply/round_le; rewrite -/(Rsqr (a i)); apply/Rle_0_sqr.
    by rewrite -nkm -hk mul1n.
  apply/Rmult_le_compat_r.
    apply/res_E_ge_0.
      by move=>i; apply/DW_F2P.
    move => i; have->: 0 = RN 0 by rewrite round_0.
    by apply/round_le; rewrite -/(Rsqr (a i)); apply/Rle_0_sqr.
  rewrite -/(beta (INR n)).
  have -> : (m = n)%nat by rewrite -nkm -hk mul1n.
  have: 0 < (beta (INR k) ) .
    apply/beta_gt0 =>//.
    by rewrite -hk /=; lra.
  suff: 0 < beta (INR n) by nra.
  apply/beta_gt0=>//.
  by change (INR 1 <= INR n); apply/Rlt_le/lt_INR/leP.
move=> k2.
move:m1; rewrite leq_eqVlt.
case/orP.
  move =>/eqP hm.
  case:Prop4_4_m1 =>//.
  move=>  <- <-.
  apply/(Rle_trans _ (((2 * INR n - 1) * u ^ 2 + (INR n + 2) * u ^ 3 + (2 * INR n - 2) * u ^ 4 +
   (7 * INR n - 7) * u ^ 5 + (3 * INR n - 3) * u ^ 6) * 
  res_E n y)).
    apply/algo_10_res_approx_1_pre=>//.
    + by apply/(@ltn_trans 1 0 n)=>//.

    + by move=>i; apply/DW_F2P.
    + move => i; have->: 0 = RN 0 by rewrite round_0.
      by apply/round_le; rewrite -/(Rsqr (a i)); apply/Rle_0_sqr.
    by rewrite -nkm -hm muln1.
  apply/Rmult_le_compat_r.
    apply/res_E_ge_0.
      by move=>i; apply/DW_F2P.
    move => i; have->: 0 = RN 0 by rewrite round_0.
    by apply/round_le; rewrite -/(Rsqr (a i)); apply/Rle_0_sqr.
  rewrite -/(beta (INR n)).
  have -> : (k = n)%nat by rewrite -nkm -hm muln1.
  have: 0 < (beta (INR m) ) .
    apply/beta_gt0 =>//.
    by rewrite -hm /=; lra.
  suff: 0 < beta (INR n) by nra.
  apply/beta_gt0=>//.
  by change (INR 1 <= INR n); apply/Rlt_le/lt_INR/leP.
move=> m2.
have M2:=(leP m2).
have K2:= (leP k2).
have h52_0: forall j , (j <= k)%nat  -> Rabs (Z_h j + Z_l j - res_aux j) <= (beta (INR m)) * (res_aux j).
  move=>j hj.
  rewrite res_aux_E.
  by apply/algo_10_res_approx_1.
pose f := fun j => (Z_h j + Z_l j - res_aux j).
have : Rabs (sum_f_R0 f (k -1)) <= (beta (INR m)) * (sum_f_R0 (fun i =>  Rsqr (a i)) (n - 1)).
  apply/(Rle_trans _ (sum_f_R0 (fun l : nat => Rabs (f l)) (k -1))).
    by apply/Rsum_abs.
  apply/(Rle_trans _ (sum_f_R0  (fun l =>   beta (INR m) * res_aux l) (k -1))).
    apply/sum_Rle=> j hj.
    apply/h52_0.
    apply/(@leq_trans (k - 1)%N); first by move/leP: hj.
    by rewrite leq_subr.
  have-> : sum_f_R0 (fun l : nat => beta (INR m) * res_aux l) (k - 1)  =  beta (INR m)  * 
    sum_f_R0 (fun l : nat => res_aux l) (k - 1) .
    rewrite scal_sum.
    apply/PartSum.sum_eq=> i hi; ring.
  apply/Rmult_le_compat_l.
  (* 0 <= beta (INR m) *)
    apply/Rlt_le/beta_gt0=>//.
    by change (INR 1 <= INR  m); apply/le_INR/leP/ltnW.
  rewrite /= /res_aux sum_decomp -?nkm //.
  lra.
rewrite /f minus_sum /res_aux sum_decomp; try (apply/leP; lia).
move=> h52.
pose g := (fun j : nat => (Z_h j) + (Z_l j)).
have h53: Rabs (S_h + S_l -         
  sum_f_R0 g (k - 1)) <= beta (INR k) * sum_f_R0 g  (k - 1) .
  set y := (fun j=> Z j).
  move:  ( algo_10_res_approx_1_pre k k1 y).
  have ->:   ((2 * INR k - 1) * u ^ 2 + (INR k + 2) * u ^ 3 + (2 * INR k - 2) * u ^ 4 +
    (7 * INR k - 7) * u ^ 5 + (3 * INR k - 3) * u ^ 6) = beta (INR k) by [].
  have ->:  res_E k y =  sum_f_R0 g (k - 1).
    rewrite /res_E /sigma -minus_n_O.
    apply/PartSum.sum_eq => i hi.
    rewrite plus_O_n.
    by [].
  apply=>//.
    move=>i; rewrite /y res10_Sn //.
    apply/DW_DWPlusFP=>//.
      apply/DWx_i=>//.
      by move=> i0; apply/DW_F2P; apply/generic_format_round.
    apply/Fsigma_i.
    by move=> i0; apply/DW_F2P; apply/generic_format_round.
  move=> i; apply/res10_S_pos=>//.
    by move=> i0; apply/DW_F2P; apply/generic_format_round.
  move=> j; have -> : 0 = RN 0 by rewrite round_0.
  apply/round_le.
  by apply/Rle_0_sqr.
have->: (S_h + S_l - res11_E)= (S_h + S_l - sum_f_R0 g (k - 1)) +   
             (sum_f_R0 g (k - 1) -  res11_E) by ring.
apply/(Rle_trans _ _  _  (Rabs_triang _ _)).
apply/(Rle_trans _ (  beta (INR k) * sum_f_R0 g (k - 1) + Rabs (sum_f_R0 g (k - 1) - res11_E))).
  by apply/Rplus_le_compat_r.
have ->:  (beta (INR k) + beta (INR m) + beta (INR k) * beta (INR m)) * res11_E = 
 (beta (INR k) + beta (INR k) * beta (INR m)) * res11_E +  beta (INR m) * res11_E by ring.
apply/Rplus_le_compat; last first.
  rewrite -/g  nkm   in h52.
  by rewrite /res11_E -sum_sigma.
have : Rabs (  sum_f_R0 g (k - 1) ) <= (1 +  beta (INR m) )* sum_f_R0 (fun i : nat => (a i)²) (n - 1).
  have->: sum_f_R0 g (k - 1) = (sum_f_R0 g (k - 1) - res11_E) + res11_E by ring.
  apply/(Rle_trans _ _  _  (Rabs_triang _ _)).
  apply/(Rle_trans _ (beta (INR m) * res11_E + res11_E)); last first.
    by rewrite /res11_E sum_sigma; lra.
  rewrite (Rabs_pos_eq res11_E).
    rewrite -/g  nkm   in h52.
    by rewrite /res11_E -sum_sigma;  lra.
  rewrite -/g  nkm  (sum_sigma _ (n -1)) in h52.
  apply/(Rmult_le_reg_l (beta (INR m))).
    apply/beta_gt0 =>//.
    change (INR 1 <= INR m).
    apply/le_INR/leP.
    by apply/ltnW.
  apply/Rmult_le_compat_l.
    apply/Rlt_le/beta_gt0=>//.
    change (INR 1 <= INR m).
    apply/le_INR/leP.
    by apply/ltnW.
  rewrite /res11_E -sum_sigma.
  by apply/cond_pos_sum => i; apply/Rle_0_sqr.
move=> h.
rewrite -(Rabs_pos_eq (sum_f_R0 g (k - 1))).
  have->:   (beta (INR k) + beta (INR k) * beta (INR m)) * res11_E = beta (INR k)* 
            (1 + beta (INR m)) * sum_f_R0 (fun i : nat => (a i)²) (n - 1).
    rewrite /res11_E sum_sigma; ring.
  rewrite Rmult_assoc; apply/Rmult_le_compat_l=>//.
  apply/Rlt_le/beta_gt0=>//.
  change (INR 1 <= INR k).
  apply/le_INR/leP.
  by apply/ltnW.
apply/(Rmult_le_reg_l (beta (INR k))).
  apply/beta_gt0=>//.
  change (INR 1 <= INR k).
  apply/le_INR/leP.
  by apply/ltnW.
rewrite Rmult_0_r; 
 apply/(Rle_trans _ (Rabs (S_h + S_l - sum_f_R0 g (k - 1))))=>//.
apply/Rabs_pos.
Qed.

Notation K := (INR k).
Notation M := (INR m).
Notation N := (INR n).

Hypothesis nu : INR n <= /u.

Fact kn : ( k <= n)%N .
Proof.
rewrite -nkm.
move:m1; case:m=>//. 
move=> m0 _.
by apply/leq_pmulr.
Qed.

Fact mn : ( m <= n)%N .
Proof.
rewrite -nkm.
move:k1; case:k=>// k0 _;  
by apply/leq_pmull.
Qed.

Lemma  res11_approx2_aux : (u <= /32) -> 
      (beta (INR k) + beta (INR m) + (beta (INR  k)) * (beta (INR m))) <= 
      (2 * K + 2 * M - 2) * u^2 + (129/1000 * N + 9465/10000 * ( K + M) + 1003/100 )* u^3.
Proof.
move=>  u32.
have upos: 0 < u by apply/bpow_gt_0.
have uipos: 0 < /u  by apply/Rinv_0_lt_compat.
have M1 : (1 <= INR m ) by change (INR 1 <= INR m); apply/le_INR/leP.
have K1 : (1 <= INR k) by change (INR 1 <= INR k); apply/le_INR/leP.
have N2 : (2 <= INR n) by change (INR 2 <= INR n); apply/le_INR/leP.
have hK: K <= /u by move/le_INR: (leP kn);lra.
have hM: M <= /u by move/le_INR: (leP mn);lra.
pose beta' t :=  (2 * t - 1) * u ^ 2 + (t + 5) * u ^ 3.
have hbk : beta K < beta' K by apply/b1_lt_b2.
have hbm : beta M < beta' M by apply/b1_lt_b2.
apply/(Rle_trans _ ( beta' K + beta' M + beta' K * beta' M)).
  apply/Rplus_le_compat; try lra.
  apply/Rmult_le_compat.
  + apply/Rlt_le/beta_gt0; move/le_INR: (leP k1);  rewrite /= ;  lra.
  + apply/Rlt_le/beta_gt0; move/le_INR: (leP m1);  rewrite /= ;  lra.
  + lra.
  lra.
rewrite {1 2}/beta'.
have->: (2 * K - 1) * u ^ 2 + (K + 5) * u ^ 3 + ((2 * M - 1) * u ^ 2 + 
      (M + 5) * u ^ 3) + 
     beta' K * beta' M =  (2 * K + 2 * M - 2) * u ^ 2 + 
   ((K + 5) * u ^ 3 + 
   (M + 5) * u ^ 3 +
     beta' K * beta' M) by ring.
apply/Rplus_le_compat_l.
rewrite /beta'.
have ->:   (K + 5) * u ^ 3 + (M + 5) * u ^ 3 +
  ((2 * K - 1) * u ^ 2 + (K + 5) * u ^ 3) * ((2 * M - 1) * u ^ 2 + (M + 5) * u ^ 3)=

((K + M + 10) + ( ((2 * K - 1)* (2 * M - 1) )* u
+ ((2 * K - 1) *  (M + 5) + (K + 5) * (2*M  -1)) * u ^ 2  + (K + 5) * ( M + 5) * u ^ 3))* u ^ 3 by ring.
apply/Rmult_le_compat_r; first nra.
apply/(Rle_trans _  ( K + M + 10 +
  ((2 * K - 1) * (2 * M - 1) /32 +
   ((2 * K - 1) * (M + 5) + (K + 5) * (2 * M - 1)) /1024 +
   (K + 5) * (M + 5) / 32768))).
  by do !(apply/Rplus_le_compat; try lra); try (apply/Rmult_le_compat_l; nra).
field_simplify(  K + M + 10 +
  ((2 * K - 1) * (2 * M - 1) / 32 +
   ((2 * K - 1) * (M + 5) + (K + 5) * (2 * M - 1)) / 1024 +
   (K + 5) * (M + 5) / 32768)).
have-> : 4225 * K * M = 4225 * N.
  suff : K*M = N by lra.
  by rewrite -mult_INR -nkm.
rewrite /Rdiv 3!Rmult_plus_distr_r Rmult_plus_distr_l -Rplus_assoc.
by do!(apply/Rplus_le_compat; last lra); nra.
Qed.

(* Lemma 4.3 *)
Lemma algo_11_res_approx_2 (u32: u <= /32)  (kiu: INR k <= /u) (miu: INR m <= /u):  
 Rabs ((S_h + S_l) - res11_E ) <= 
  ((2 * K + 2 * M - 2) * u^2 + (129/1000 * N + 9465/10000 * ( K + M) + 1003/100 )* u^3) * res11_E.
Proof.
apply/(Rle_trans _ ((beta (INR k) + beta (INR m) + (beta (INR  k)) * (beta (INR m))) * res11_E)).
  by apply/algo_11_res_approx_1=>//;  lra.
apply/Rmult_le_compat_r; last by apply/res11_approx2_aux.
rewrite /res11_E.
by apply/cond_pos_sum=>i;  apply/Rle_0_sqr.
Qed.

Fact Property4_4_2 (n3: (3 <= n)%nat) (u32: u <= /32) (n3_u2: 3 * (INR n - 1) * u^2 < 1) : 
             (2 * INR n - 1 )* u^2 + (INR n + 5)* u^3 <= 3 * (INR n -1)* u^2 /( 1 - 3 * (INR n - 1) * u^2).
Proof.
have upos: 0 < u by apply/bpow_gt_0.
have uipos: 0 < /u  by apply/Rinv_0_lt_compat.
have u2pos: 0 < u^2 by nra.
have ult1: u < 1 by rewrite -pow0 /u; apply/bpow_lt; lia.
have N3 : 3 <= N.
  have<-: INR 3 = 3 by rewrite /=; lra.
  by apply/le_INR/leP.
have hden: 1 <= / (1 - 3 * (N - 1) * u ^ 2) .
  by rewrite -{1}Rinv_1 /=; apply/Rle_Rinv; nra.
suff : 0 <=  3 * (N - 1) * u ^ 2 / (1 - 3 * (N - 1) * u ^ 2) -  
            ((2 * N - 1) * u ^ 2 + (N + 5) * u ^ 3) by lra.
apply/(Rle_trans _ (3 * (N - 1) * u ^ 2  - ((2 * N - 1) * u ^ 2 + (N + 5) * u ^ 3))); try nra.
  have->:  3 * (N - 1) * u ^ 2 - ((2 * N - 1) * u ^ 2 + (N + 5) * u ^ 3) = 
                (N -2 )* u^2 - (N + 5) * u^3 by ring.
  apply/(Rle_trans _ (( N -2)* u^2 - ( N/32 + 5/32)* u^2)); try nra.
  suff : (N + 5) * u ^ 3 <= (N  + 5 )/32  * u ^ 2 by nra.
  have->: u^3 = u * u^2 by ring.
  rewrite -Rmult_assoc; apply/Rmult_le_compat_r; nra.
apply/Rplus_le_compat_r.
rewrite -[X in X <= _]Rmult_1_r.
apply/Rmult_le_compat_l; nra.
Qed.

Fact Property4_4_3 (k2 : (2 <= k)%nat) (m2 : (2 <= m)%nat)   (u32: u <= /32) (Niu: INR n <= /u): 
(2*(INR k) + 2* (INR m) -2)*u^2 +(129/1000 * N + 9465/10000 * ( K + M) + 1003/100 )* u^3  <= 
 (2 * INR n - 1 )* u^2 + (INR n + 5)* u^3.
Proof.
have upos: 0 < u by apply/bpow_gt_0.
have uipos: 0 < /u  by apply/Rinv_0_lt_compat.
have u2pos: 0 < u^2 by nra.
have ult1: u < 1 by rewrite -pow0 /u; apply/bpow_lt; lia.
have M2 : (2 <= INR m ) by change (INR 2 <= INR m); apply/le_INR/leP.
have K2 : (2 <= INR k ) by change (INR 2 <= INR k); apply/le_INR/leP.
have N4 : 4 <= N by rewrite -nkm mult_INR; nra.

suff: 0 <= -((2 * K + 2 * M - 2) * u ^ 2 + (129 / 1000 * N + 9465 / 10000 * (K + M) + 1003 / 100) * u ^ 3) 
+ ((2 * N - 1) * u ^ 2 + (N + 5) * u ^ 3) by lra.

field_simplify (((2 * K + 2 * M - 2) * u ^ 2 + (129 / 1000 * N + 9465 / 10000 * (K + M) + 1003 / 100) * u ^ 3) -
  ((2 * N - 1) * u ^ 2 + (N + 5) * u ^ 3)).
have->:  - ((2 * K + 2 * M - 2) * u ^ 2 + (129 / 1000 * N + 9465 / 10000 * (K + M) + 1003 / 100) * u ^ 3) +
  ((2 * N - 1) * u ^ 2 + (N + 5) * u ^ 3) = 
(2*N  -2*(K + M) + 1)*u^2 + (871/1000 * N - 503/100 - 9465/10000* (K +M ))* u^3 .
  field.
have kmn: K + M <= N/2 + 2.
  case:(@L4_2 n k m)=>//.
  by rewrite -mult_INR -nkm.
have h : -(N / 2 + 2)<= - (K + M) by lra. 
have h2: - (N / 2 + 2) * u ^ 2 <= - (K + M) * u ^ 2.
  apply/Rmult_le_compat_r; nra.
have h3: - (N / 2 + 2) * u ^ 3 <= - (K + M) * u ^ 3.
  apply/Rmult_le_compat_r; nra.
apply/(Rle_trans _ 
 ((2 * N - 2 * ( N / 2 + 2) +1)* u^2 + (871 / 1000 * N - 503 / 100  -  9465 / 10000 * (N/2 +2))* u^3)); 
last first.
  nra.
have->: (2 * N - 2 * (N / 2 + 2) + 1)  = (N - 3) by lra.
have->: (871 / 1000 * N - 503 / 100 - 9465 / 10000 * (N / 2 + 2))= 
         (39775/100000 * N  - 6923/1000) by field.
apply/(Rle_trans _ ( (N - 3) * u ^ 2  - 6923/1000 * u^3)); last first.
  apply/Rplus_le_compat_l.
  rewrite Rmult_minus_distr_r  -[X in X <= _]Rplus_0_l.
  apply/Rplus_le_compat; try nra.
  have : 0 < u^3 by nra.
  nra.
suff: 0<= (N - 3) - 6923 / 1000 * u  by nra.
suff:  6923 / 1000 * u <= N -3 by lra.
apply/(Rle_trans _ 1).
  by apply/(Rle_trans _ ( 6923 / 1000 * (/32)));  nra.
lra.
Qed.

End Algo_11.

End  Algo_10_11.

(* Th 4.5 no overflow/underflow MED case *)
Variable n: nat.
Hypothesis n1: (2 <= n)%N. 

Variable a : nat -> R.
(* Hypothesis Fa : forall i,  format (a i). *)

Section TH4_5_with_algo10.

Notation res_S := (res10_S n (fun i =>  (Fast2Mult (a i) ( a i)))).

Notation S_h := (fst res_S).
Notation S_l := (snd res_S).

Notation  R10 := (SQRTDWToFP p choice S_h S_l) .
Notation res10 := (sqrt (sigma (fun i => Rsqr (a i)) 0 (n-1))).
Notation  nu := ((beta (INR n))/(u^2)).


Theorem Th4_5_10 (u32: u <= /32) : ( INR n ) <= /u ->
  nu < /(2*u) ->
  Rabs (R10 - res10) <= (/2 + u * ( 7/4 + nu /(1 - nu * u^2)))* ulp (res10).
Proof.
move=> n_u nu_u.
have upos: 0 < u by apply/bpow_gt_0.
have p5: (5  <= p)%Z.
  suff: (-p <= -5)%Z by lia.
  apply/(le_bpow two); by rewrite -/u /=;  lra.
have N2 : (2 <= INR n) by change (INR 2 <= INR n); apply/le_INR/leP.
have sigpos:  0 <= sigma (fun i : nat => (a i)²) 0 (n - 1).
  by apply/cond_pos_sum=> i; apply/Rle_0_sqr.
rewrite /u -Rsqr_pow2.
apply/Th_3_10' =>//.
+ apply/Rmult_le_pos.
    apply/Rlt_le/beta_gt0; lra.
  by apply/Rlt_le/Rinv_0_lt_compat/Rlt_0_sqr; rewrite -/u; lra.
+ rewrite res10_Sn //.
  apply/DW_DWPlusFP; try lia.
    apply/DWx_i.
      by move=>i; apply/DW_F2P.
    lia.
  apply/Fsigma_i.
  by move=>i; apply/DW_F2P. 
+ apply/(Rle_trans _ (  ((2 * INR n - 1) * u ^ 2 + (INR n + 2) * u ^ 3 + (2 * INR n - 2) *    u ^ 4 +
   (7 * INR n - 7) * u ^ 5 + (3 * INR n - 3) * u ^ 6) *
  sigma (fun i : nat => (a i)²) 0 (n - 1))); last first.
    apply/(Rmult_le_compat_r)=>//.
    rewrite -/(beta (INR n)).
    by rewrite Rmult_assoc Rinv_l /Rsqr -/u;  nra.
  by apply/algo_10_res_approx_1 =>//;  try lia; try lra.
+ apply/res10_S_pos =>//; try lia; try lra.
    - by apply/leP; move/leP: n1; lia.
    - by move=> i; apply/DW_F2P.
  move=>i; have->: 0 = RN 0 by rewrite round_0.
  by apply/round_le/Rle_0_sqr.
have->:  2 * pow (- p) * (beta (INR n) / (pow (- p))²)  = 2 * beta (INR n) /u.
  by rewrite -/u /Rsqr; field;  lra.
move:nu_u; rewrite /nu -/u.
have ->: 2 * beta (INR n) / u =  beta (INR n) * / u ^ 2 * (2 * u) .
  field; nra.
move/(Rmult_lt_compat_r (2 *u)).
by rewrite !Rmult_assoc Rinv_l; lra.
Qed.

End TH4_5_with_algo10.

Section TH4_5_with_algo11.
Variables  k m: nat.
Hypothesis nkm: (k * m = n)%nat.

Fact km0: (1 <= k)%nat /\ (1 <= m)%nat.
Proof.
move : nkm n1; case: k.
  by rewrite mul0n => <- .
move=> i; case: m=>//.
by rewrite muln0 =><- .
Qed.



Notation res_S := (res11_S k m a).

Notation S_h := (fst res_S).
Notation S_l := (snd res_S).


Notation  R11 := (SQRTDWToFP p choice S_h S_l) .
Notation res11 := (sqrt (sigma (fun i => Rsqr (a i)) 0 (n-1))).

Notation  nu := ((beta (INR k)+(beta (INR m)) + (beta (INR k)) * (beta (INR m)))/(u^2)).



Theorem Th4_5_11 (u32: u <= /32) (ku: INR k <= /u) (mu: INR m <= /u):
  nu < /(2*u) ->
  Rabs (R11 - res11) <= (/2 + u * ( 7/4 + nu /(1 - nu * u^2)))* ulp (res11).
Proof.
move=>  nu_u.
have upos: 0 < u by apply/bpow_gt_0.
have u2pos: 0 < u^2 by nra.
have ult1: u < 1 by rewrite -pow0 /u; apply/bpow_lt; lia.
have p5: (5  <= p)%Z.
  suff: (-p <= -5)%Z by lia.
  apply/(le_bpow two); by rewrite -/u /=;  lra.
have N2 : (2 <= INR n) by change (INR 2 <= INR n); apply/le_INR/leP.
have sigpos:  0 <= sigma (fun i : nat => (a i)²) 0 (n - 1).
  by apply/cond_pos_sum=> i; apply/Rle_0_sqr.
case:km0=>/leP k1 /leP m1.
have bk0: 0 < beta (INR k) by apply/beta_gt0; try lra;  apply/(le_INR 1).
have bm0: 0 < beta (INR m) by apply/beta_gt0; try lra;  apply/(le_INR 1).
rewrite /u -Rsqr_pow2.
apply/Th_3_10' =>//.
+ apply/Rmult_le_pos; try nra.
  apply/Rlt_le/Rinv_0_lt_compat/Rlt_0_sqr; rewrite -/u; lra.
+ apply/DW_res11_S; lia.
+ have ->:   (beta (INR k) + beta (INR m) + beta (INR k) * beta (INR m)) /
  (pow (- p))² * (pow (- p))² *
  sigma (fun i : nat => (a i)²) 0 (n - 1) =   (beta (INR k) + beta (INR m) + beta (INR k) * beta (INR m))*
  sigma (fun i : nat => (a i)²) 0 (n - 1).
    field.
    suff:  0 < (pow (- p)) .
      rewrite /Rsqr; nra.
    by apply/bpow_gt_0.
  apply/algo_11_res_approx_1=>//; try lia; try (apply/leP; lia); lra.
+ apply/res10_S_pos =>//; try lia; try lra.
  - by apply/leP; move/leP: n1; lia.
  - move=>i; apply DWres10_S.
     by move=> j; apply/DW_F2P.
    lia.
  move=> i; apply/res10_S_pos=>//; try (apply/leP; lia).
  + by move=> j; apply/DW_F2P.
  + move=>j; have->: 0 = RN 0 by rewrite round_0.
    by apply/round_le/Rle_0_sqr.
  + lia.
  lra.
rewrite /Rsqr; field_simplify; last first.
rewrite -/u; lra.
move:nu_u; rewrite /nu -/u.
move=>h.
apply/(Rmult_lt_reg_r u)=>//.
rewrite !Rmult_assoc Rinv_l; try lra.
suff: (beta (INR k) + beta (INR m) + beta (INR k) * beta (INR m)) < u/2 by lra.
apply/(Rmult_lt_reg_r (/ u^2)).
  apply/Rinv_0_lt_compat; nra.
have->: u / 2 * / u ^ 2 = / (2 * u).
  field.
  lra.
by [].
Qed.

End TH4_5_with_algo11.


End Main.

