(* Copyright (c) ENS de Lyon and Inria. All rights reserved. *)

Require Import Reals  Psatz.
From Flocq Require Import Core Relative Sterbenz Operations Mult_error  Plus_error.
Require Import mathcomp.ssreflect.ssreflect.
Require Import Bayleyaux.
From Flocq Require Import Pff2Flocq.
Require Import F2Sum DWPlus.

From Flocq Require Import Div_sqrt_error Sqrt.

From Interval Require Import Tactic.
Require Import square_root_aux.

Open Scope R_scope.

Set Implicit Arguments.

Section RN_sym.
Variable r:radix.
Variable fexp:  Z -> Z.
Hypothesis valid_fexp: Valid_exp fexp.

Variable choice : Z -> bool.
Hypothesis choice_sym: forall x, choice x  = negb (choice (- (x + 1))).

Notation RN :=(round r  fexp (Znearest choice) ).

Lemma RN_sym x: RN (-x) = - (RN  (x)).
Proof.
suff : - RN (- x) =  RN x by lra.
by rewrite round_N_opp Ropp_involutive /round  /Znearest -choice_sym.
Qed.

Lemma RN_abs x: RN (Rabs x) = Rabs (RN x).
Proof.
case:(Rlt_le_dec x 0)=> x0.
  have: RN x <= 0.
    have -> : 0 = RN 0 by rewrite round_0.
    by apply:round_le;lra.
  by rewrite Rabs_left // => h ; rewrite Rabs_left1 // RN_sym.
move/(round_le r fexp (Znearest choice)):(x0); rewrite round_0=> h.
by rewrite !Rabs_pos_eq.
Qed.
End RN_sym. 


Section Main.

Notation two  := radix2.
Notation pow e := (bpow two e).


Variable p: Z.
Context { prec_gt_0_ : Prec_gt_0 p }.
Variable choice : Z -> bool.
Hypothesis choice_sym: forall x, choice x  = negb (choice (- (x + 1))).


(***********************   FLX ************************************)
Notation fexp := (FLX_exp   p).
Notation format := (generic_format two fexp ).
Notation RN :=(round two  fexp (Znearest choice) ).

Hypothesis pGt1: (1 < p)%Z.

Fact pow1 : pow 1 = 2 . by []. Qed.
Fact pow2 : pow 2 = 4 . by []. Qed.
Fact pow0: pow 0 = 1 . by []. Qed.
Fact powm1: pow (-1) = /2 .  by []. Qed.
Fact pow3 : pow 3 = 8 . by []. Qed.
Fact pow4 : pow 4 = 16 . by []. Qed.

Let u := pow (-p).


Local Notation ulp := (ulp two fexp).


(* Section No_overflow. *)

Fact diff_max x y (xpos: 0 <= x) (ypos : 0 <= y) : Rabs (x - y) <= Rmax x y.
Proof.
case:(Rle_lt_dec x y)=> xy.
  have ->: (x -y) = -(y -x) by ring.
  rewrite Rabs_Ropp Rabs_pos_eq ?Rmax_right; lra.
by rewrite  Rabs_pos_eq ?Rmax_left; lra.
Qed.


Section DW_sqrt.


Notation TwoSum_sum := (TwoSum_sum p  choice).
Notation TwoSum_err := (TwoSum_err p choice).

Notation  F2Sum  := (Fast2Sum  p  choice ).

Notation  F2Sum_sum a b  := (fst (F2Sum a b)).
Notation  F2Sum_err a b  := (snd (F2Sum a b)).

Notation double_word xh xl :=  (((format xh) /\ (format xl)) /\ 
                                 ( xh = RN (xh + xl))%R).



Notation sh xh := (RN (sqrt xh)).
Notation ro1  xh  := (RN (xh - (sh xh )^2)).
Notation ro2 xh xl := (RN (xl + (ro1 xh))).
Notation sl xh xl := (RN ((ro2 xh xl)/ (2*(sh xh)))).

Fact shS x k (xpos: 0<= x): sh (x * pow (2*k))= sh x * pow k.
rewrite sqrt_mult //; try apply/bpow_ge_0.
have-> : (2*k = k +k)%Z by ring.
rewrite bpow_plus sqrt_square; last by apply/bpow_ge_0.
by rewrite round_bpow.
Qed.

Fact ro1S x k (xpos: 0<= x): ro1 (x * pow (2*k))= (ro1 x) * pow(2* k).
Proof.
rewrite shS // Rpow_mult_distr .
have ->:  pow k ^ 2 =  pow (2 * k).
rewrite -Z.add_diag bpow_plus; ring.
by rewrite -Rmult_minus_distr_r round_bpow.
Qed.

Fact ro2S xh xl  k (xpos: 0<= xh): ro2 (xh * pow (2*k)) (xl * pow (2*k)) = 
(ro2 xh xl) * pow(2* k).
Proof.
by rewrite ro1S //  -Rmult_plus_distr_r round_bpow.
Qed.

Fact slS xh xl  k (xpos: 0< xh): sl (xh * pow (2*k)) (xl * pow (2*k)) = 
(sl xh xl) * pow k.
Proof.
have xhpos: 0 <= xh by lra.
rewrite ro2S ?shS //.
rewrite -Z.add_diag bpow_plus.
suff->:ro2 xh xl * (pow k * pow k) / (2 * (sh xh * pow k ))= 
   (ro2 xh xl)/ (2*(sh xh)) * pow k by rewrite round_bpow.
field; split; try (move: (bpow_gt_0 two k); lra).
apply/round_neq_0_negligible_exp ; try apply/negligible_exp_FLX.
move=>h.
have: xh <> 0 by lra.
by apply; apply/sqrt_eq_0.
Qed.



Notation z xh xl :=(F2Sum (sh xh) (sl xh xl)).
(* Notation zh xh xl  :=  (F2Sum_sum (sh xh) (sl xh xl)). *)
(* Notation zl xh xl :=  (F2Sum_err (sh xh) (sl xh xl)). *)
(* Algo 6*)
Definition SQRTDWToDW  xh xl := z xh xl.
(* Algo 7*)
Definition SQRTDWToFP xh xl := RN ((sh xh) + (sl xh xl)).

Remark R4_6 xh xl k  (DWx : double_word xh xl): 
      xh = pow (2*k) -> SQRTDWToFP xh xl  = pow k.
Proof.
move=>xhE.
case:(DWx)=>[[Fxh Fxl]xE].
rewrite /SQRTDWToFP.
have ->: sh xh = pow k.
  rewrite xhE.
  have->: (2 * k  = k + k)%Z by ring.
  rewrite bpow_plus sqrt_square ?round_generic //.
    by apply/format_bpow.
  by apply/bpow_ge_0.
have h1: 0 < pow (k) by apply/bpow_gt_0.
rewrite xhE.
have->: (pow (2 * k) - pow k ^ 2) = 0
   by rewrite -Z.add_diag bpow_plus; ring.
rewrite round_0 Rplus_0_r (round_generic _ _ _ xl) //.
apply/(Rmult_eq_reg_r (pow k)); last lra.
rewrite -round_bpow Rmult_plus_distr_r  -bpow_plus Z.add_diag.
rewrite -xhE /Rdiv.
have->: / (2 * pow k) = pow (-k -1).
  apply/(Rmult_eq_reg_r (pow k)); last lra.
  rewrite -bpow_plus; ring_simplify(-k-1+k)%Z.
  rewrite powm1;field; lra.
rewrite round_bpow (round_generic _ _ _ xl) //.
rewrite  Rmult_assoc -bpow_plus; ring_simplify(-k-1+k)%Z.
rewrite powm1.
case:(Rle_lt_dec 0 xl)=> xl0; apply/Rle_antisym.
+ rewrite {2}xE; apply/round_le; lra.
+ have {1}->: xh = RN xh by rewrite round_generic.
  by apply/round_le; lra.
+ have {2}->: xh = RN xh by rewrite round_generic.
  by apply/round_le; lra.
rewrite {1}xE; apply/round_le;lra.
Qed.

Notation SQRTDW2_1 xh xl := (fst (SQRTDWToDW xh xl)).
Notation SQRTDW2_2 xh xl := (snd (SQRTDWToDW xh xl)).

Definition rel_err_sqrtDW2 xh xl :=  
    Rabs(((SQRTDW2_1 xh xl + SQRTDW2_2 xh xl) - (sqrt (xh +xl)))/sqrt(xh + xl)).

Definition rel_err_sqrtDWFP xh xl :=  
    Rabs(( (SQRTDWToFP xh xl) - (sqrt (xh +xl))) /sqrt(xh + xl)).

Definition err_sqrtDWFP xh xl :=  
    Rabs(SQRTDWToFP xh xl  - (sqrt (xh +xl))).

Fact  rel_err_sqrtDW2_scale xh xl k (xhpos: 0 < xh) (DWx: double_word xh xl):
 rel_err_sqrtDW2 (xh * pow (2* k)) (xl * pow (2*k)) = rel_err_sqrtDW2  xh xl.
Proof.
case:(DWx)=>[[Fxh Fxl] xE].
rewrite /rel_err_sqrtDW2 /SQRTDWToDW.
rewrite slS ?shS ?Fast2SumS //; try lra; try apply/generic_format_round.
set fs:= F2Sum_sum _ _.
set fe:= F2Sum_err _ _.
set d:= xh*_  +xl *_.
rewrite /= /d.
case:(Rlt_le_dec 0 (xh +xl))=> xpos; last first.
  have : RN (xh +xl) <= RN 0 by apply/round_le.
  rewrite round_0 -xE; lra.
rewrite -!Rmult_plus_distr_r sqrt_mult; try apply/bpow_ge_0; try lra.
rewrite -Z.add_diag bpow_plus sqrt_square; try apply/bpow_ge_0.
rewrite -Rmult_minus_distr_r.
field_simplify
 ((fs + fe - sqrt (xh + xl)) * pow k / (sqrt (xh + xl) * pow k))=>//.
split;first by move:(bpow_gt_0 two k); lra.
move=>h.
have : xh + xl <> 0 by lra.
apply; apply/sqrt_eq_0; lra.
Qed.

Fact  rel_err_sqrtDWFP_scale xh xl k (xhpos: 0 < xh) (DWx: double_word xh xl):
 rel_err_sqrtDWFP (xh * pow (2* k)) (xl * pow (2*k)) = rel_err_sqrtDWFP  xh xl.
Proof.
case:(DWx)=>[[Fxh Fxl] xE].
rewrite /rel_err_sqrtDWFP /SQRTDWToFP.
rewrite slS ?shS  ;  try lra.

(* set d:= xh*_  +xl *_. *)
(* rewrite /= /d. *)
case:(Rlt_le_dec 0 (xh +xl))=> xpos; last first.
  have : RN (xh +xl) <= RN 0 by apply/round_le.
  rewrite round_0 -xE; lra.
rewrite -!Rmult_plus_distr_r sqrt_mult; try apply/bpow_ge_0; try lra.
rewrite -Z.add_diag bpow_plus sqrt_square; try apply/bpow_ge_0.
rewrite round_bpow -Rmult_minus_distr_r.
set a:= _ - sqrt _.
field_simplify  (a * pow k / (sqrt (xh + xl) * pow k)) =>//.
split.
 by move:(bpow_gt_0 two k); lra.
move=>h.
have : xh + xl <> 0 by lra.
apply; apply/sqrt_eq_0; lra.
Qed.

Fact err_sqrtDWFP_scale xh xl k (xhpos: 0 < xh) (DWx: double_word xh xl):
   err_sqrtDWFP (xh * pow (2* k)) (xl * pow (2*k)) = 
    (err_sqrtDWFP  xh xl )* (pow k).
Proof.
case:(DWx)=>[[Fxh Fxl] xE].
rewrite /err_sqrtDWFP /SQRTDWToFP.
rewrite slS ?shS  ;  try lra.
case:(Rlt_le_dec 0 (xh +xl))=> xpos; last first.
  have : RN (xh +xl) <= RN 0 by apply/round_le.
  rewrite round_0 -xE; lra.
rewrite -!Rmult_plus_distr_r sqrt_mult; try apply/bpow_ge_0; try lra.
rewrite -Z.add_diag bpow_plus sqrt_square; try apply/bpow_ge_0.
rewrite round_bpow -Rmult_minus_distr_r Rabs_mult (Rabs_pos_eq (pow k)) //;
 apply/bpow_ge_0.
Qed.


Lemma scale_generic_14_strict  x (Fx: format x) (xpos:  0 < x):(exists e, (Zeven e) /\ 1 <= x * pow   e < 4).
Proof.
have xn0: x <> 0 by lra.  
case: (mag two x)=>ex /(_ xn0).
rewrite  Rabs_pos_eq;( try lra) => [[hex1 hex2]].
case:(Zeven_odd_dec ex)=>ex_even.
  exists (-ex +2)%Z.
  split.
    suff->: (-ex +2 = Z.succ (Z.succ (-ex)))%Z.
      apply/Zeven_Sn/Zodd_Sn.
      by rewrite Z.opp_eq_mul_m1; apply/Zeven_mult_Zeven_l.
    have->: (-ex +2 = -ex+ 1+1)%Z by ring.
    by rewrite-!Z.add_1_r.
  split.
    apply/(Rmult_le_reg_r (pow (ex -2))).
      by apply/bpow_gt_0.
    apply/(Rle_trans _ (pow (ex -1))).
      by rewrite Rmult_1_l; apply/bpow_le; lia.
    rewrite Rmult_assoc -bpow_plus.
    ring_simplify (- ex + 2 + (ex - 2))%Z.
    rewrite pow0; lra.
  rewrite bpow_plus pow2.
  suff:  x * (pow (- ex) ) < 1by lra.
  apply/(Rmult_lt_reg_r (pow (ex ))).
    apply/bpow_gt_0.
  rewrite Rmult_assoc -bpow_plus.
  ring_simplify(-ex +ex)%Z.
  rewrite pow0; lra.
exists (-ex +1)%Z; split.
  apply/Zeven_Sn.
  by rewrite Z.opp_eq_mul_m1; apply/Zodd_mult_Zodd =>//.
split.
  apply/(Rmult_le_reg_r (pow (ex -1))).
    by apply/bpow_gt_0.
  rewrite Rmult_1_l Rmult_assoc -bpow_plus.
  ring_simplify(- ex + 1 + (ex - 1))%Z.
  by rewrite pow0; lra.
apply/(Rmult_lt_reg_r (pow (ex -1))).
  apply/bpow_gt_0.
rewrite Rmult_assoc -bpow_plus.
ring_simplify (- ex + 1 + (ex - 1))%Z.
rewrite pow0; apply/(Rlt_trans _ (pow ex))=>//; try lra.
rewrite bpow_plus powm1; suff : 0< (pow ex) by lra.
apply/bpow_gt_0.
Qed.


Lemma scale_generic_12_strict  x (Fx: format x) (xpos:  0 < x):(exists e, 1 <= x * pow e < 2).
Proof.
have xn0: x <> 0 by lra.  
case: (mag two x)=>ex /(_ xn0).
rewrite  Rabs_pos_eq;( try lra) => [[hex1 hex2]].

(* have h := (id_p_ulp_le_bpow two fexp x ex xpos Fx hex2). *)
(* have hpow := (bpow_gt_0 two (ex -1)). *)
exists (-(ex -1))%Z.
rewrite bpow_opp;split.
 apply:(Rmult_le_reg_r ((pow (ex - 1)%Z)))=>//.
    apply/bpow_gt_0.

  (* by field_simplify; try lra. *)
rewrite Rmult_assoc.
  suff-> : (/ pow (ex - 1) * pow (ex - 1))=1 by lra.
  apply/Rinv_l; move:(bpow_gt_0 two (ex -1)); lra.
 apply:(Rmult_lt_reg_r ((pow (ex - 1)%Z))); try apply/bpow_gt_0.
rewrite Rmult_assoc.
  suff-> : (/ pow (ex - 1) * pow (ex - 1))=1.
rewrite bpow_plus powm1; lra.
rewrite Rinv_l //;  move:(bpow_gt_0 two (ex -1)); lra.
Qed.



Fact DWpos xh xl (DWx: double_word xh xl): 0 < xh -> 0 < xh + xl.
Proof.
move=> xhpos.
have Hp1:=pGt1.
case:(DWx)=>[[Fxh Fxl]xE].
suff : (0 <= xh + xl ) /\ (xh + xl <> 0) by lra.
case:(Rle_lt_dec 0 (xh + xl))=>hx0.
  split=>// x0.
  by move:xE; rewrite x0 round_0; lra.
move/Rlt_le: hx0=>hx0.
 suff: RN (xh + xl) <= 0 by lra.
have ->: 0 = RN 0 by rewrite round_0.
by apply/round_le.
Qed.


Fact mag_pos_bounds x (xpos: 0<  x) : pow (mag two x -1) <= x < pow (mag two x).
Proof.
by case:mag=> magx /=;rewrite Rabs_pos_eq; lra.
Qed.


Fact ulp_eq_eps  x (Fx : format x) (xpos: 0 < x) z : z = x + /2* ulp x ->
 (ulp z) =( ulp x).
Proof.
move=> zE.
have zpos: 0 <z by move: (ulp_ge_0 two fexp x); lra.
rewrite !ulp_neq_0 /cexp; try lra.
rewrite zE.
rewrite  (mag_plus_eps _ fexp) //.
suff: 0<ulp x by lra.
rewrite ulp_neq_0; move:(bpow_gt_0 two (cexp two fexp x)); lra.
Qed.


Fact not_sqrt_midp x (Fx : format x) (xpos: 0 < x) z : z = x + /2* ulp x -> 
                     ~(format (z^2)).
Proof.
move=>zE' Fy.
move:(zE').
move:(Fx); rewrite {1}/generic_format.
set mx := Ztrunc _ .
set ex := cexp _ _ _ .
rewrite /F2R /=  => xE.
rewrite ulp_neq_0 ; last lra.
rewrite {1} xE /ex -Rmult_plus_distr_r  /cexp/fexp.
pose ez := (mag two x -1)%Z.
have magxE : mag two x = (ez +1)%Z :>Z by  rewrite /ez; ring.
have ezE : (ez = ex + p -1)%Z by rewrite /ex /ez /cexp /fexp; ring.
rewrite magxE.
set Z := (IZR mx).
rewrite -{2}/Z.
have Zb: pow (p -1) <= Z < pow p.
  suff:  pow (p - 1) * (pow ex) <= Z  * pow ex < pow p * pow ex.
    move: (bpow_gt_0 two ex); nra.
  move: (mag_pos_bounds xpos).
  rewrite magxE xE -/Z -!bpow_plus ezE.
  ring_simplify (ex + p - 1 + 1 - 1)%Z; ring_simplify (ex + p - 1 + 1)%Z.
  have->: (ex + p - 1 = p - 1 + ex)%Z by ring.
  by have->:( ex + p = p +ex)%Z by ring.
move=> zE.
move: (Fy); set y := z^2.
rewrite /generic_format.
set my := Ztrunc _ .
set ey' := (cexp two fexp y).
rewrite /F2R /=.
rewrite /ey' /cexp/fexp.
have->:  ((mag two y - p) =  (mag two y  -1 + (1 - p)))%Z by ring.
set ey := (mag two y - 1 )%Z.
set Y := IZR my.
rewrite /y zE.
move/(Rmult_eq_compat_l 4).
have->:  4 * ((Z + / 2) * pow (ez + 1 - p)) ^ 2 = (2*Z +1)^2  *  pow (ez + 1 - p)^2 by field.
move/(Rmult_eq_compat_r ((pow (p -1 -ez)) * (pow (p -1 -ez)))).
have ->:  pow (ez + 1 - p) ^ 2 =  (pow (ez + 1 - p)) * (pow (ez + 1 - p)) by ring.
rewrite 4!Rmult_assoc  -!bpow_plus.
ring_simplify (ez + 1 - p + (ez + 1 - p + (p - 1 - ez + (p - 1 - ez))))%Z.
ring_simplify  (ey + (1 - p) + (p - 1 - ez + (p - 1 - ez)))%Z.
rewrite Rmult_1_r.
rewrite -(IZR_Zpower two). 
  have-> : (2 * Z + 1) ^ 2 = (2 * Z + 1)*(2 * Z + 1) by ring.
  rewrite -!mult_IZR -plus_IZR -mult_IZR.
  move/eq_IZR=> h.
  have : Zodd ((2 * mx + 1) * (2 * mx + 1))%Z.
    by apply/Zodd_mult_Zodd; apply/Zodd_2p_plus_1.
  have : Zeven (4 * (my * two ^ (ey + p - 2 * ez - 1)))%Z.
    have->: (4 = 2*2)%Z by ring.
    rewrite -Zmult_assoc.
    by apply/Zeven_2p.
  by move/Zeven_not_Zodd; rewrite h.
suff : (2*ez <= ey)%Z by lia.
rewrite /ez /ey /y.
have-> : z^2 = z* z by ring.
have ulp0: 0 < ulp x.
  rewrite ulp_neq_0; try lra.
  by apply/bpow_gt_0.
apply/(Z.le_trans _  (2 * (mag two z - 1))).
  apply/Z.eq_le_incl.
  congr Zmult; congr Zminus.
  rewrite zE' (mag_plus_eps _  fexp) //.
  lra.
have zpos: 0 < z.
  rewrite zE'; lra.
case:(mag_mult two z z); try lra.
lia.
Qed.


Fact pred_pos x (Fx : format x) : 0 < x -> 0 < pred two fexp x.
Proof.
move=> xpos.
have: pred two fexp  x = 0 -> x = 0.
move=> h0.
have : succ two fexp ( pred two fexp x ) = 0.
  by rewrite  h0 succ_0 ulp_FLX_0 // .
by rewrite  succ_pred.
move=>h.
have : 0 <= pred two fexp x by apply/pred_ge_0.
case=>//.
lra.
Qed.


Fact sqrtF_no_midpoint_aux z (zpos: 0 < z):
 Rabs (RN z - z)= /2* ulp z-> 
   z = round two fexp Zfloor z + /2 * ulp (round two fexp Zfloor z).
Proof.
move=>herr.
have ce_pos :=(bpow_gt_0 two (cexp two fexp z)).
have [rf rc]:=( round_DN_UP_le p two z).
case: rf => rf; last first.
  move : herr ; rewrite rf round_generic ?Rminus_diag_eq //.
  rewrite Rabs_R0 ulp_neq_0;lra.
  by rewrite - rf; apply/generic_format_round.
case: rc => rc; last first.
  move : herr; rewrite  round_generic ?Rminus_diag_eq //.
    rewrite Rabs_R0 ulp_neq_0; lra.
  by rewrite rc; apply/generic_format_round.
have nF: ~ format z by move/round_DN_UP_generic; lra.
case:(round_DN_or_UP two fexp (Znearest choice) z)=>rE;
move: (herr); rewrite rE.
  rewrite -Rabs_Ropp Rabs_pos_eq; try lra.
  by rewrite ulp_DN; lra.
rewrite Rabs_pos_eq ; last lra.
rewrite ulp_DN ?round_UP_DN_ulp //; lra.
Qed.

Fact sqrtF_no_midpoint x (Fx: format x) (xpos : 0 < x) : Rabs (RN (sqrt x) - (sqrt x))< /2* ulp (sqrt x).
Proof.
have sqrt0: 0 < sqrt x.
  case:(sqrt_pos x)=>//  h.
  suff  : x = 0 by lra.
  by apply/sqrt_eq_0; lra.
have :  Rabs (sh x - sqrt x) <= / 2 * ulp ( (sqrt x)).
  by apply/error_le_half_ulp.
case=>//.
have zE : x = (sqrt x)^2 by rewrite -Rsqr_pow2 /Rsqr sqrt_sqrt; lra.
move/(sqrtF_no_midpoint_aux   sqrt0).
 move=> h.
suff : ~format x by [].
rewrite zE; apply/(@not_sqrt_midp  (round two fexp Zfloor (sqrt x))) =>//.
  by apply/generic_format_round.
have: 0<=  round two fexp Zfloor (sqrt x).
have->:  0 = round two fexp Zfloor 0 by rewrite round_0.
  apply/round_le; lra.
case=>// =>h0.
suff : sqrt x = 0  by lra.
apply:(eq_0_round_0_FLX two p Zfloor ); lra.
Qed.

 Lemma dw_ulp_flx xh xl: double_word xh xl -> Rabs xl <= /2 * ulp xh.
Proof.
case=>[[Fxh Fxl] xhe].
have -> : xl = - (RN (xh + xl)  - (xh + xl)) by rewrite -xhe; ring.
rewrite Rabs_Ropp; apply:(Rle_trans _  (/ 2 * ulp (RN (xh +xl)))).
  by apply: error_le_half_ulp_round.
by rewrite -xhe; lra.
Qed.


Lemma dw_ulpx_flx xh xl: double_word xh xl -> Rabs xl <= /2 * ulp (xh +xl).
Proof.
case=>[[Fxh Fxl] xhe].
have {1} -> : xl = - (RN (xh + xl)  - (xh + xl)) by rewrite -xhe; ring.
rewrite Rabs_Ropp;
  by apply: error_le_half_ulp.
Qed.

Lemma error_le_half_ulp'_flx (t:R) (k:Z): 
        ((Rabs t) <= pow k)%R -> 
                       (Rabs ((RN t) - t)) <= /2 * u  * (pow k).
Proof.
have rge0: 0 <=  / 2 * u * pow k.
  by rewrite  /u Rmult_assoc -bpow_plus; move:(bpow_ge_0 two (- p + k)); lra.
case:(Req_dec t 0)=>[-> |t0].
  by rewrite round_0 Rminus_0_r Rabs_R0. 
case=>Rteq.
  apply: (Rle_trans _  (/2 * Ulp.ulp two fexp t)); 
    first exact:error_le_half_ulp.
  rewrite Rmult_assoc; apply:Rmult_le_compat_l; first lra.
  rewrite /u ulp_neq_0 // /cexp /fexp Rmult_comm.  
  rewrite bpow_plus.
  apply:Rmult_le_compat_r; first exact: bpow_ge_0.
  apply/bpow_le.
  by apply/mag_le_bpow=>//.
rewrite round_generic.
  by ring_simplify ( t - t); rewrite Rabs_R0 //.
move:Rteq; rewrite -(Rabs_pos_eq (pow k)); last by apply/bpow_ge_0.
case/Rabs_eq_Rabs => ->; try apply/generic_format_opp;
apply/generic_format_bpow; rewrite /fexp;lia.
Qed.

(*Th 3.6 & 3.7 *)
Lemma SQRTDW_bounds xh xl (DWx: double_word  xh xl) (xhpos: 0 < xh)
                           (Hp5: (5 <= p)%Z) : 
rel_err_sqrtDW2 xh xl <= 25/8* u^2 /\ 
((err_sqrtDWFP xh xl  <= (1/2 + 175/100 * u) *  (ulp (sqrt (xh + xl)))) /\
 rel_err_sqrtDWFP xh xl <= u + 17/8*u^2 + 33/8*u^3).
Proof.
wlog binx: xh  xl DWx xhpos  / 1 <= xh <= 4 -4*u.
  move=> Hwlog.
  have [[Fxh Fxl] Exh] := DWx.
  case: (scale_generic_14_strict Fxh xhpos)=> exh [ Hxhse Hxhs].
  case:(Zeven.Zeven_ex _  Hxhse)=> e exhE.
  case:(Hwlog  (xh * pow exh) (xl * (pow exh))).
  + split.
      split; by apply/formatS.
    by rewrite -Rmult_plus_distr_r round_bpow  {1}Exh.
  + move:(bpow_gt_0 two exh); nra.
  + split; try lra.
    suff-> : 4 -4*u = pred two fexp 4.
      apply/pred_ge_gt; last lra.
        by apply/formatS.
      rewrite -pow2; apply/generic_format_bpow; rewrite /fexp; lia.
    by rewrite  -pow2 pred_bpow /fexp bpow_plus.
  rewrite exhE rel_err_sqrtDW2_scale ?rel_err_sqrtDWFP_scale 
     ?err_sqrtDWFP_scale //.
  move=>h1 [h2 h3].
  split=>//; split=>//.
  case:(Rlt_le_dec 0 (xh +xl))=> xpos; last first.
    have : RN (xh +xl) <= RN 0 by apply/round_le.
    rewrite round_0 -Exh; lra.
  have ppos:=(bpow_gt_0 two ( (e+e))).
  apply/(Rmult_le_reg_r (pow e)); first by apply/bpow_gt_0.
  move:h2; rewrite -Rmult_plus_distr_r sqrt_mult -?Z.add_diag; try lra.
  rewrite bpow_plus sqrt_square; last by apply/bpow_ge_0.
  by rewrite  ulp_FLX_exact_shift; lra.

(* particular case : 1 <= xh <= 4 - 4 * u *)
set x := xh +xl.
have upos: 0<u by rewrite /u; apply/bpow_gt_0.
have u2pos: 0 < u^2 by nra.
have ult1: u < 1 by rewrite -pow0 /u; apply/bpow_lt; lia.
case: (DWx) =>[[Fxh Fxl] xE].
have : xh < 4 -4*u + u^2 by lra.
have ->: 4 -4*u + u^2 = (2-u)^2 by ring.
move=>xh_ub.
have: sqrt xh <  sqrt ((2 - u) ^ 2) by apply/sqrt_lt_1_alt;lra.
rewrite sqrt_pow2; last lra.
have p2: 2- 2*u = pred two fexp (pow 1).
  by rewrite pred_bpow /fexp bpow_plus pow1 /u.
move=>sqrt_ub.
have shub: sh xh <= 2-2*u .
  apply/round_N_le_midp.
    by rewrite p2; apply/generic_format_pred/format_bpow.
  suff -> : succ two fexp (2 - 2 * u)= 2 by lra.
  rewrite p2 succ_pred pow1 //.
  by rewrite -pow1; apply/format_bpow.
have shlb: 1 <= sh xh.
  have->: 1 = RN 1 by rewrite round_generic // -pow0; apply/format_bpow.
  by apply/round_le; rewrite -sqrt_1; apply/sqrt_le_1; lra.
have h9': Rabs (sh xh - sqrt xh) < u.
  have  sqrtx_lb: 1 <= sqrt xh.
    by rewrite -sqrt_1; apply/sqrt_le_1_alt; lra.
  have ->: u = /2* ulp(sqrt xh).
    rewrite ulp_neq_0; last lra.
    rewrite /cexp /fexp  (mag_unique_pos _ _ 1).
      by rewrite bpow_plus -/u pow1; lra.
    by rewrite pow0 pow1; lra.
  by apply/ sqrtF_no_midpoint.
have ro1E: ro1 xh = xh - (Rsqr (sh xh) ).
  rewrite round_generic //.
    by rewrite -Rsqr_pow2.
  by rewrite -Rsqr_pow2; apply/ sqrt_error_FLX_N.
have h10: Rabs (sqrt x - sh xh) <= 
         Rabs (sqrt xh -sh xh) + Rabs (sqrt x - sqrt xh).
  have ->: sqrt x  -sh xh =  (sqrt xh -sh xh) +(sqrt x - sqrt xh) by ring.
  by apply/Rabs_triang.
have xpos := (DWpos DWx xhpos).
rewrite -/x in xpos.
case:(MVT_abs (fun x =>  sqrt x) (fun  x  => / (2* sqrt x)) xh x).
  move=> c hc;  apply/derivable_pt_lim_sqrt.
  apply/(Rlt_le_trans _ (Rmin xh x)); last  lra.
  by apply/Rmin_glb_lt; move:(DWpos DWx); rewrite /x; lra.
move=> zeta.
have ->: x -xh = xl by rewrite /x; ring.
move=>[h11 hzeta].
have z0 : 0 < zeta.
  apply/(Rlt_le_trans _ (Rmin xh x)); last  lra.
  by apply/Rmin_glb_lt;lra.
have sqrtz0:= (sqrt_lt_R0 _ z0).
have sqrtzi0:= (Rinv_0_lt_compat _  sqrtz0).
have cas1: (xh = 1 ) -> (xl < 0) -> Rabs (sqrt x - sqrt xh ) <= u/2.
  move=> xh1 xl0 .
  have xllb : -u <= xl.
    move:( dw_ulp_flx DWx).
    rewrite ulp_neq_0 /cexp /fexp ?(mag_unique_pos _ _ 1); try lra; 
      last first.
      by rewrite pow0 pow1; lra.
    rewrite -Rabs_Ropp Rabs_pos_eq ; try lra.
    by rewrite bpow_plus pow1 -/u; lra.
  pose d:= pred two fexp xh.
  have dE: d = 1 -u .
    rewrite /d xh1 -pow0 pred_bpow /fexp bpow_plus pow0 -/u; ring.
  have xd:  round two  fexp Zfloor x = d.
    apply/round_DN_eq.
      by rewrite /d; apply/generic_format_pred.
    by rewrite {2}/d succ_pred //   dE /x; split;  lra.
  have [rf rc]:=( round_DN_UP_le p two x).
  case: rf => hf; last first.
    have :  RN (round two fexp Zfloor x) =  RN x by rewrite hf.
    rewrite round_generic; try apply /generic_format_round.
    by rewrite xd /x -xE; lra.
  have xu: round two fexp Zceil x = xh.
    apply/round_UP_eq=>//.
    rewrite -/d dE -/x; split; try lra.
    by rewrite/x; lra.
  case: rc => rc; last first.
    have : RN (round two fexp Zceil x) = RN x by rewrite -rc.
    rewrite round_generic; try apply /generic_format_round.
    by rewrite -rc /x -xE; lra.
  have nF: ~ format x by move/round_DN_UP_generic; lra.
  case:(Rle_lt_dec (-u/2) xl )=>// h; last first.
    have : RN x = round two fexp Zfloor x.
      apply/round_N_eq_DN.
      by rewrite xd xu /x; lra.
    by rewrite xd /x -xE; lra.
  rewrite h11 -[X in _ <= X]Rmult_1_l.
  apply/Rmult_le_compat; try apply/Rabs_pos.
    rewrite Rabs_pos_eq.
      rewrite -Rinv_1.
      apply/Rinv_le; try lra.
      suff :/2 <= sqrt zeta by lra.
      apply/Rsqr_incr_0_var.
        rewrite Rsqr_sqrt.
          apply/(Rle_trans _ ( Rmin xh x)); try lra.
          rewrite Rmin_right; try lra.
          have ->: Rsqr (/2) = /4 by  rewrite /Rsqr; lra.
          lra.
        apply/(Rle_trans _ ( Rmin xh x)); try lra.
        by rewrite Rmin_right; lra.
      by apply/sqrt_ge_0.
    apply/Rlt_le/Rinv_0_lt_compat.
    suff : 0< sqrt zeta by lra.
    apply/sqrt_lt_R0.
    apply/(Rlt_le_trans _ ( Rmin xh x)); try lra.
    by rewrite Rmin_right;  lra.
  by rewrite -Rabs_Ropp Rabs_pos_eq; lra.
have cas12: (xh = 1 ) -> (0 <= xl ) -> Rabs (sqrt x - sqrt xh ) <= u/2.
  move=>xh1 xl0.
  have xlu:  xl <= u.
    move:(dw_ulp_flx DWx); rewrite Rabs_pos_eq; try lra.
    rewrite ulp_neq_0 /cexp /fexp ; try lra.
    rewrite (mag_unique_pos _ _ 1).
      by rewrite bpow_plus pow1 -/u; lra.
    by rewrite pow0 pow1; lra.
  have  sqrzt_1 : 1 <= sqrt zeta.
    rewrite -sqrt_1; apply/sqrt_le_1_alt; try lra.
    apply/(Rle_trans _ ( Rmin xh x)); try lra.
    by rewrite Rmin_left ; rewrite /x; lra.
  have isqrt_pos : 0 < /( sqrt zeta ).
    by apply/Rinv_0_lt_compat; lra.
  have sqrtzi_1: /(sqrt zeta)<=1.
    rewrite -Rinv_1.
    by apply/Rinv_le; lra.
  rewrite h11 (Rabs_pos_eq xl); try lra.
  rewrite Rinv_mult_distr; try lra.
  rewrite Rabs_mult !Rabs_pos_eq; try lra.
  nra.
have cas2': 1 < xh <2 ->  Rabs (sqrt x - sqrt xh) <= u / 2.
  move=> [xh1 xh2].
  have  sqrzt_1 : 1 < sqrt zeta.
    rewrite -sqrt_1; apply/sqrt_lt_1_alt; try lra.
    split; try lra.
    apply/(Rlt_le_trans _ ( Rmin xh x)); try lra.
    apply/Rmin_glb_lt; try lra.
    case:(Rlt_le_dec 1 x)=>// x1.
    have: RN x <= RN 1 by apply/round_le.
    rewrite /x -xE -pow0 round_generic; last by apply/format_bpow.
    by rewrite pow0;lra.
  move:(sqrzt_1); rewrite -sqrt_1=> sqrtz_1.
  have :/sqrt zeta < /sqrt 1.
    apply/Rinv_lt; try lra.
    by rewrite sqrt_1; lra.
  rewrite sqrt_1 Rinv_1=> sqrtiz_1.
  have sqrtiz_0: 0 < /(sqrt zeta).
    by apply/Rinv_0_lt_compat; lra.
  rewrite h11 Rinv_mult_distr; try lra.
  rewrite Rabs_mult Rabs_pos_eq;try lra.
  rewrite Rabs_pos_eq;try lra.
  suff : 0<= Rabs xl <= u by nra.
  split; first by apply/Rabs_pos.
  move:(dw_ulp_flx DWx).
  rewrite ulp_neq_0 /cexp /fexp ; try lra.
  rewrite bpow_plus -/u. 
  suff: 0<= pow (mag two xh)<= 2 by nra.
  split; first by apply/bpow_ge_0.
  rewrite (mag_unique_pos _ _ 1).
    by rewrite pow1;lra.
  by rewrite pow0 pow1; lra.
have cas3 : xh = 2 -> xl <0 ->  Rabs (sqrt x - sqrt xh) <= u / (sqrt 2).
  move=> xh2 xl0.
  apply/(Rle_trans _ (u /(2 * sqrt (2 -u)))); last first.
    apply/(Rmult_le_reg_l (/u)) .
      by apply/Rinv_0_lt_compat; lra.
    rewrite  -!Rmult_assoc Rinv_l ?Rmult_1_l; try lra.
    apply/Rinv_le.
      by rewrite -sqrt_0; apply/sqrt_lt_1_alt; lra.
    have->:  2 * sqrt (2 - u) = sqrt (4 * (2 -u)).
      rewrite sqrt_mult; try lra.
      have ->:  4 = 2*2 by [].
      by rewrite sqrt_square //; lra.
    apply/sqrt_le_1_alt.
    suff : 1 <= 4 -2*u by lra.
    lra.
(* have xllb: -u <= xl. *)
  have pxhE: pred two fexp xh = 2 -2*u.
    by rewrite xh2 -pow1 pred_bpow /fexp bpow_plus pow1 -/u; ring.
  have [h1 h2] : pred two fexp xh <= x < xh.
    split;last by   rewrite /x ;lra.
    rewrite pxhE;    move:(dw_ulp_flx DWx).
    rewrite ulp_neq_0 /cexp /fexp; try lra.
    rewrite (mag_unique_pos _ _ 2); last first.
      by rewrite pow1 pow2; lra.
    by rewrite bpow_plus pow2 -Rabs_Ropp Rabs_pos_eq -/u /x;  lra.
  case: h1=>h1; last first.
    have :  RN (pred two fexp xh)  = RN x by rewrite h1.
    rewrite round_generic;  last  by apply/generic_format_pred.
    by rewrite /x ;lra.
  have cex : cexp two fexp x  = (1-p)%Z.
    rewrite /cexp /fexp.
    by rewrite (mag_unique_pos _ _ 1) // pow0 pow1 -/u;split; lra.
  have rfxh: round two fexp Zfloor x =  pred two fexp xh .
    apply/round_DN_eq; first by apply/generic_format_pred.
    by rewrite succ_pred //; lra.
  have nF : ~ format x.
    by move=>Fx; move:xE; rewrite round_generic //; lra.
  case: (Rle_lt_dec (-u) xl )=> // xlu; last first.
    have: RN x = round two fexp Zfloor x.
      apply/round_N_eq_DN.
      rewrite -succ_DN_eq_UP // ?rfxh.
      by rewrite succ_pred /x //; lra.
    by rewrite rfxh /x -xE; lra.
  rewrite h11 Rinv_mult_distr; try lra.
  rewrite Rabs_mult Rabs_pos_eq;try lra.
  rewrite Rabs_pos_eq;try lra.
  rewrite /Rdiv (Rmult_comm u).
  apply/Rmult_le_compat; try lra.
  + by apply/Rabs_pos.
  + have h0: 0 < sqrt (2 -u) by apply/sqrt_lt_R0; lra.
    rewrite Rinv_mult_distr; try lra.
    apply/Rmult_le_compat_l;try lra.
    apply/Rinv_le; try lra.
    apply/sqrt_le_1; try lra.
    apply/(Rle_trans _ (Rmin xh x)); try lra.
    by rewrite /x Rmin_right;  lra.
  by rewrite -Rabs_Ropp Rabs_pos_eq; lra.
have cas3': xh = 2 -> 0 <= xl ->  Rabs (sqrt x - sqrt xh) <= u / sqrt 2.
  move=> xh2 xl0.
  rewrite h11.
  have h :  sqrt 2  <= sqrt zeta.
    apply/sqrt_le_1_alt/(Rle_trans _ (Rmin xh x)); try lra.
    by rewrite Rmin_left /x ; lra.
  have: /(sqrt zeta) <= /(sqrt 2).
    apply/Rinv_le; try lra.
    by apply/Rlt_sqrt2_0.
  move=>h2.
  rewrite Rabs_Rinv; try lra.
  rewrite !Rabs_pos_eq; try lra.
  rewrite Rinv_mult_distr; try lra.
  rewrite Rmult_comm -Rmult_assoc.
  apply/Rmult_le_compat; try lra.
  move:(dw_ulp_flx DWx).
  rewrite Rabs_pos_eq ?ulp_neq_0 /cexp /fexp ; try lra.
  rewrite (mag_unique_pos _ _ 2).
    by rewrite bpow_plus pow2 -/u; lra.
  by rewrite pow1 pow2; lra.
have cas4: 2 < xh ->  Rabs (sqrt x - sqrt xh) <= u / sqrt 2.
  move=>xh2.
  have xhulp: ulp  xh = 4*u.
    rewrite ulp_neq_0 /cexp /fexp; last lra.
    rewrite (mag_unique_pos _ _ 2).
      by rewrite bpow_plus pow2.
    by  rewrite pow1 pow2; lra.
  have xlub: Rabs xl <= 2*u.
    by move:(dw_ulp_flx DWx); lra.
  rewrite h11.
  have xh2': 2 + 4*u <= xh.
    have-> : 2+4*u = succ two fexp 2.
      rewrite succ_eq_pos ? ulp_neq_0 /cexp /fexp ; try lra.
      by rewrite( mag_unique_pos _ _ 2) ?pow1 ?bpow_plus pow2 -/u; lra.
    apply/succ_le_lt=>//.
    by rewrite -pow1; apply/format_bpow; lia.
  rewrite Rabs_Rinv; try lra.
  rewrite Rabs_pos_eq; try lra.
  rewrite Rinv_mult_distr; try lra.
  rewrite Rmult_comm -Rmult_assoc.
  apply/Rmult_le_compat; try lra.
    suff: 0<= Rabs xl by lra.
    by apply/Rabs_pos.
  apply/Rinv_le; try lra.
    by apply/Rlt_sqrt2_0.
  apply/sqrt_le_1; try lra.
  apply/(Rle_trans _ (Rmin xh x)); try lra.
  apply/Rmin_glb ; rewrite /x ;try  lra.
  by move/Rabs_le_inv: xlub; lra.
have h12: xh <= 2-2*u -> Rabs (sqrt x -sh xh ) < 3/2*u.
  move=> xhub; apply/(Rle_lt_trans _ 
                     (Rabs (sqrt xh - sh xh) + Rabs (sqrt x - sqrt xh)))=>//.
  have ->: 3/2 * u = u + u/2 by lra.
  apply/Rplus_lt_le_compat.
    by rewrite -Rabs_Ropp Ropp_minus_distr.
  move:binx=>[[xh1 | xh1] _].
    by apply/  cas2'; lra.
  by case:(Rle_lt_dec 0 xl);[apply/ cas12| apply/ cas1];lra.
have sqrt2_pos := Rlt_sqrt2_0.
have h12': 2- 2*u < xh -> Rabs (sqrt x - sh xh) < u * (1 + (sqrt 2)/2).
  have-> : 2 - 2*u = pred two fexp 2.
    by rewrite  -[X in pred _ _ X]pow1 pred_bpow /fexp bpow_plus pow1 -/u.
  move=> h.
  have F2: format 2 by rewrite -pow1;  apply/format_bpow.
  have : succ two fexp (pred two fexp 2) <= xh.
    apply/succ_le_lt=>//.
    by apply/generic_format_pred.
  rewrite succ_pred // Rmult_plus_distr_l Rmult_1_r=> xh2.
  apply/(Rle_lt_trans _ 
                     (Rabs (sqrt xh - sh xh) + Rabs (sqrt x - sqrt xh)))=>//.
  apply/Rplus_lt_le_compat.
    by rewrite -Rabs_Ropp Ropp_minus_distr. 
  have -> : sqrt 2 /2 = /(sqrt 2).
    apply/(Rmult_eq_reg_l (sqrt 2)); try lra.
    rewrite -Rmult_assoc Rinv_r; try lra.
    by rewrite sqrt_def; lra.
  case: xh2 => xh2.
    by  apply/cas4.
  case: (Rle_lt_dec 0 xl)=>xl0.
    by apply/cas3'.
  by apply/cas3.
(* line 3 *)
have hro1 : Rabs (ro1 xh) <  u*(2* (sh xh) + u).
  rewrite ro1E.
  have->: (xh - (sh xh)²)= (sqrt xh - (sh xh))* (sqrt xh + (sh xh)).
    ring_simplify.
    by rewrite -Rsqr_pow2 Rsqr_sqrt /Rsqr; first  ring; lra.
  rewrite Rabs_mult;apply/Rmult_lt_compat; try apply/Rabs_pos.
    by rewrite -Rabs_Ropp Ropp_minus_distr.
  have ->: (sqrt xh + sh xh) = 2*(sh xh) + - (sh xh - sqrt xh) by ring.
  apply/(Rle_lt_trans _ _  _  (Rabs_triang _ _)).
  rewrite Rabs_pos_eq; try lra.
  apply/Rplus_lt_compat_l.
  by rewrite Rabs_Ropp.
have h13: Rabs (xl + (ro1 xh)) < Rabs xl + u*(2* (sh xh) + u).
   by apply/(Rle_lt_trans _ _  _  (Rabs_triang _ _)); lra.
have h13_1: xh<= 2-2*u -> Rabs xl <= u /\ (sh xh < sqrt 2 + u/2).
  move=> xh2.
  split.
    move: (dw_ulp_flx DWx); rewrite ulp_neq_0 /cexp /fexp; last lra.
    by rewrite (mag_unique_pos _ _ 1) ?pow0 ?bpow_plus pow1 -/u;  lra.
  apply/(Rlt_trans _ (sqrt 2 *(1 -u/2) + u)); last first.
    rewrite Rmult_minus_distr_l Rmult_1_r.
    suff: u * ( 1 - sqrt 2 * /2) < u/2 by lra.
    apply/Rmult_lt_compat_l =>//.
    suff : 1 < sqrt 2  by lra.
    by rewrite -sqrt_1; apply/sqrt_lt_1; lra.
  apply/(Rlt_trans _ (sqrt ( 2 - 2*u) +u)); last first.
    apply/Rplus_lt_compat_r.
    have ->:  (2 - 2 * u) = 2 * ( 1 -u) by ring.
    rewrite sqrt_mult; try lra.
    apply/Rmult_lt_compat_l =>//.
    apply/Rsqr_incrst_0; try lra; try apply/sqrt_pos.
    rewrite Rsqr_sqrt; last lra.
    have->: (1 - u / 2)² = 1 -u + ( u / 2)² by  rewrite /Rsqr; field.
    by rewrite /Rsqr; lra.
  apply/(Rlt_le_trans _ (sqrt xh + u)).
    by move/Rabs_lt_inv:(h9'); lra.
  apply/(Rplus_le_compat_r).
  by apply/sqrt_le_1; lra.
have h13_2:  xh<= 2-2*u  -> (4 <= p)%Z -> Rabs (xl + ro1 xh) < 4*u.
  move=> xh2 Hp4.
  apply/(Rlt_trans _ (u*(1 + 2* (sqrt 2))+ 2* (Rsqr u))); last first.
    rewrite Rmult_comm /Rsqr -Rmult_assoc  -Rmult_plus_distr_r.
    apply/(Rmult_lt_compat_r); first  lra.
    suff: (sqrt 2)  < 3/2 -u  by lra.
    apply/Rsqr_incrst_0; try lra.
    rewrite Rsqr_sqrt; last lra.
    rewrite /Rsqr;
      suff : u <= /16 by lra.
    change ((pow (-p)) <= pow (-4)).
    by apply/bpow_le; lia.
  apply/(Rlt_le_trans _ (Rabs xl + u * (2 * sh xh + u)))=>//.
  case:h13_1=>//=>*.
  rewrite /Rsqr !Rmult_plus_distr_l Rmult_1_r Rplus_assoc.
  apply/Rplus_le_compat; try lra.
  rewrite -Rmult_plus_distr_l.
  have->:  u * (2 * sqrt 2) + 2 * (u * u) = u*((2 * sqrt 2) + 2 * u)by ring.
  apply/Rmult_le_compat_l; first lra.
  lra.
have h13_3: 2 <= xh -> Rabs xl <= 2*u /\ sh xh <= 2.
  move => xh2; split.
    move:(dw_ulp_flx DWx); rewrite ulp_neq_0 /cexp /fexp; last lra.
    rewrite (mag_unique_pos _ _ 2).
      by rewrite bpow_plus pow2 -/u; lra.
    by rewrite pow1 pow2; lra.
  lra.
have h13_4:  2 <= xh -> Rabs (xl + ro1 xh ) < 6*u + Rsqr u.
  move=> xh2.    
  apply/(Rlt_le_trans _ (Rabs xl + u * (2 * sh xh + u)))=>//.
  have->: 6*u + Rsqr u = 2*u + (u*4 + Rsqr u) by ring.
  apply/Rplus_le_compat; try lra.
  case: h13_3=> // *.
  rewrite Rmult_plus_distr_l /Rsqr; apply/Rplus_le_compat_r ; try lra.
  by clear -upos    shlb  shub; nra.
pose e1 := (ro2 xh xl) - x + Rsqr (sh xh).
have h14: ro2 xh xl = x -Rsqr (sh xh) + e1 by rewrite /e1; lra.
have h15: ((xh <= 2 -2*u )/\ Rabs e1 <= 2* Rsqr u)\/ 
          ( (2 -2*u < xh) /\ (Rabs e1 <= 4* Rsqr u)).
  have e1E: e1 = RN(xl + ro1 xh) - (xl + ro1 xh).
    rewrite /e1 ro1E /x; ring.
  case:(Req_dec (xl + ro1 xh) 0) => e10.
    rewrite e1E e10 round_0 Rminus_0_r Rabs_R0.
    by case:(Rle_lt_dec xh (2 -2*u));[left|right]; rewrite /Rsqr;lra.
  have h : Rabs e1 <= /2 * (ulp (xl + ro1 xh)).
    rewrite e1E; apply/error_le_half_ulp.
  case :(Rle_lt_dec xh (2-2*u))=> xh2.
    left; split=>//.
    apply/(Rle_trans _ ( / 2 * ulp (xl + ro1 xh)))=>//.
    suff:  ulp (xl + ro1 xh) <= 4 * u²by lra.
    rewrite ulp_neq_0 // /cexp {1}/fexp.
    apply/(Rle_trans _ (pow (2 -p -p))).
      apply/bpow_le.
      suff :  ( (mag two (xl + ro1 xh) ) <= (2 -p))%Z by lia.
      apply/mag_le_bpow =>//.
      rewrite bpow_plus pow2 -/u; apply/ h13_2 =>//;lia.
    rewrite !bpow_plus -/u pow2 /Rsqr; lra.
  have xhge2: 2 <= xh.
    have-> : 2 = succ two fexp (pred two fexp 2).
      by rewrite succ_pred // -pow1; apply/format_bpow.
    apply/succ_le_lt =>//; first apply/generic_format_pred.
      by rewrite -pow1; apply/format_bpow.
    by rewrite -pow1 pred_bpow /fexp bpow_plus pow1 -/u.
  right;split=>//;
   apply/(Rle_trans _ ( / 2 * ulp (xl + ro1 xh)))=>//.
  suff:  ulp (xl + ro1 xh) <= 8 * u² by lra.
  rewrite ulp_neq_0 // /cexp {1}/fexp.
  apply/(Rle_trans _ (pow (3 -p -p))).
    apply/bpow_le.
    suff :  ( (mag two (xl + ro1 xh) ) <= (3 -p))%Z by lia.
    apply/mag_le_bpow =>//.
    rewrite bpow_plus pow3 -/u. 
    apply/(Rlt_trans _ ( 6 * u + u²)).
      by apply/ h13_4.
    suff : Rsqr u <  2*u by lra.
    by rewrite /Rsqr; apply/Rmult_lt_compat_r; lra.
  rewrite !bpow_plus -/u pow3 /Rsqr; lra.
(* line 4*)
have h16_0: Rabs (ro2 xh xl) <= Rabs xl + u * (2* sh xh + u) + Rabs e1.
  rewrite h14.
  apply/(Rle_trans _ _  _  (Rabs_triang _ _)).
  apply/Rplus_le_compat_r.
  have->: (x - (sh xh)²) = xl + (xh -Rsqr (sh xh)) by rewrite /x; ring.
  apply/(Rle_trans _ _  _  (Rabs_triang _ _)).
  apply/(Rplus_le_compat_l).
  by rewrite -ro1E; lra.
have i2sh: 0< /(2 * sh xh).
  by apply/Rinv_0_lt_compat;clear - shlb; lra.
have: Rabs (ro2 xh xl) /(2* sh xh) <=
          (Rabs xl  + u * (2 * sh xh + u) + Rabs e1)/(2*sh xh).
  by apply/Rmult_le_compat_r; lra.
rewrite /Rdiv -(Rabs_pos_eq (/(2*sh xh))); last lra.
rewrite -!Rabs_mult  3!Rmult_plus_distr_r -!Rabs_mult.
rewrite (Rabs_pos_eq (/(2*sh xh))); last lra.
have->: u * (2 * sh xh + u) * / (2 * sh xh) = u * ( 1 + u/(2* sh xh)).
  field; clear -shlb; lra.
move=> h16_1.
have h17_0: Rabs (ro2 xh xl * / (2 * sh xh)) <= 2*u -2* Rsqr u.
  case:(Rle_lt_dec xh (2 -2*u))=> xh2.
    apply/(Rle_trans _ (3/2*u + 3/2*Rsqr u)); last first.
      suff : 8 *u <= 1  by rewrite /Rsqr ; clear -upos; nra.
      by rewrite -pow3 -/u -pow0; rewrite -bpow_plus; apply/bpow_le; lia.
    apply/(Rle_trans _ (Rabs (xl * / (2 * sh xh)) + 
                    u * (1 + u / (2 * sh xh)) +
                    Rabs (e1 * / (2 * sh xh))))=>//.
    case:h15; try lra.
    case=>_ he1.
    have->: 3/2* u + 3/2*Rsqr u= u/2 + u + /2*Rsqr u + Rsqr u.
      by field.
    apply/Rplus_le_compat; last first.
      rewrite Rabs_mult.
      have->: Rsqr u = (2* Rsqr u) */2 by lra.
      apply/Rmult_le_compat =>//; try apply/Rabs_pos.
      rewrite Rabs_pos_eq; try lra.
      by apply/Rinv_le; lra.
    rewrite Rplus_assoc.
    apply/Rplus_le_compat.
      rewrite Rabs_mult /Rdiv; apply/Rmult_le_compat; 
        try apply/Rabs_pos; try lra.
      rewrite Rabs_pos_eq; try lra.
      by apply/Rinv_le; lra. 
    rewrite Rmult_plus_distr_l Rmult_1_r;apply/Rplus_le_compat_l.
    field_simplify; last lra.
    apply/Rmult_le_compat; try lra.
      by rewrite /Rsqr; lra.
    by apply/Rinv_le; lra. 
  have ultsqrt2: u< sqrt 2.
    apply/Rsqr_incrst_0; try lra.
    rewrite Rsqr_sqrt; try lra.
    by rewrite /u /Rsqr -bpow_plus -pow1; apply/bpow_lt; lia.
  case:h15; try lra; case=> _ he1.
  have F2: format 2 by rewrite -pow1; apply/format_bpow.
  have xhge2: 2 <= xh.
    have->: 2 = (succ two fexp (pred two fexp 2)).
      by rewrite succ_pred //.
    apply/succ_le_lt =>//;  try by   apply/generic_format_pred.
    by rewrite -pow1 pred_bpow /fexp bpow_plus pow1 -/u.
  have h17_0: sqrt 2 -u < sh xh .
    apply/(Rlt_le_trans _ (sh 2)); last first.
      apply/round_le.
       apply/sqrt_le_1; lra.
     have : Rabs (sh 2 - sqrt 2) < / 2 * ulp (sqrt 2).
      apply/ sqrtF_no_midpoint=>//; lra.
    rewrite ulp_neq_0 /cexp ? (mag_unique_pos _ _ 1); try lra; 
      last first.
      rewrite pow0 pow1 -sqrt_1; split.
        apply/sqrt_le_1; lra.
      have {2}->: 2 = sqrt 4 .
        have-> : 4 = 2 * 2 by ring.
        rewrite sqrt_square //;  lra.
      apply/sqrt_lt_1; lra.
    rewrite {2}/fexp bpow_plus pow1 -/u -Rmult_assoc Rinv_l // 
           Rmult_1_l.
    by move/Rabs_lt_inv; lra.
  apply/(Rle_trans _ (Rabs (xl * / (2 * sh xh)) + 
                    u * (1 + u / (2 * sh xh)) +
                    Rabs (e1 * / (2 * sh xh))))=>//.
  apply/(Rle_trans _ ((2*u)/(2* (sqrt 2 -u)) +
                    u*(1 + u/(2* (sqrt 2 -u))) +
                    2* Rsqr u)).
    apply/Rplus_le_compat.
      apply/Rplus_le_compat.
        rewrite Rabs_mult /Rdiv;apply/Rmult_le_compat; try apply/Rabs_pos.
          by lra.
        rewrite Rabs_pos_eq; try lra.
        by apply/Rinv_le;  lra.
      apply/Rmult_le_compat_l; try lra.
      apply/Rplus_le_compat_l.
      apply/Rmult_le_compat_l; try lra.
      by apply/Rinv_le;  lra.
    rewrite Rabs_mult.
    have -> : 2 * Rsqr u =( 4* Rsqr u) * /2 by field.
    apply/Rmult_le_compat =>//;try apply/Rabs_pos.
    rewrite Rabs_pos_eq; try lra.
    by apply/Rinv_le;   lra.
  pose phi v := /(sqrt 2 - v) + (1 + v /(2 * (sqrt 2 -v))) + 2*v.
  pose h v:= 2 -2*v.
  have h_decr: decreasing h by move=>z t; rewrite /h; lra.
  have phi_incr:  forall z t ,  0 <= z -> t  < sqrt 2 -> z <= t ->
   phi z <= phi t.
    move=> z t  hz0 ht2 hzt; rewrite /phi.
    apply/Rplus_le_compat; last lra.
    apply/Rplus_le_compat; try  lra.
      by apply/Rinv_le; lra.
    apply/Rplus_le_compat_l.
    apply/Rmult_le_compat; try lra.
      apply/Rlt_le/Rinv_0_lt_compat.
      suff: z < sqrt 2  by lra.
      by apply/(Rle_lt_trans _ t).
    by apply/Rinv_le; lra.
  have ->:  2 * u / (2 * (sqrt 2 - u)) = u* /(sqrt 2 - u) .
    field; lra.
  have->:  2* Rsqr u = u * (2 *u) by rewrite /Rsqr; ring.
  rewrite -!Rmult_plus_distr_l {2}(Rmult_comm 2 u).
  rewrite -Rmult_minus_distr_l.
  apply/Rmult_le_compat_l; try lra.
  suff : phi u <= h u by rewrite /phi /h; lra.
  apply/(Rle_trans _ (phi (/32))).
    apply/phi_incr; try lra.
      apply/Rsqr_incrst_0; try lra.
      by rewrite Rsqr_sqrt /Rsqr; lra.
    change (pow (-p) <= pow (-5)).
    by apply/bpow_le; lia.
  apply/(Rle_trans _ (h (/32))); last first.
    apply/h_decr.
    change (pow (-p) <= pow (-5)).
    by apply/bpow_le; lia.
  by rewrite /phi /h; interval.
have slub: Rabs (sl xh xl) <= 2*u -2*Rsqr u.
  have->:  ( 2 * u - 2 * u²) =RN ( 2 * u - 2 * u²).
    rewrite round_generic //.
    have-> : (2 * u - 2 * u²) = pred two fexp (2*u).
      by rewrite -pow1 /u -bpow_plus pred_bpow /fexp /Rsqr !bpow_plus; ring.
    apply/generic_format_pred.
    by rewrite -pow1 /u -bpow_plus; apply/format_bpow.
  by rewrite -RN_abs //; apply/round_le; lra.
pose e2:= sl xh xl - (ro2 xh xl * / (2 * sh xh)).
have slE: sl xh xl = (ro2 xh xl * / (2 * sh xh)) + e2 by rewrite /e2; ring.
have e2ub: Rabs e2 <= Rsqr u.
  apply/(Rle_trans _ (/2 * ulp ((ro2 xh xl )/ (2 * sh xh)))).
    rewrite /e2; apply/error_le_half_ulp.
  apply/(Rle_trans _ (/2 * ulp (2*u - 2*Rsqr u))).
  apply/Rmult_le_compat_l; first lra.
  apply/ulp_le; rewrite (Rabs_pos_eq (2 * u - 2 * u²)) //; 
    last by clear -upos ult1; rewrite /Rsqr; nra.
   rewrite ulp_neq_0;     last by clear -upos ult1; rewrite /Rsqr; nra.
  rewrite /cexp /fexp (mag_unique_pos _ _ (-p +1)).
    rewrite !bpow_plus /Rsqr /u pow1; lra.
   rewrite !bpow_plus /Rsqr -/u pow1 powm1; split; try lra.
  suff: 2*u*u <= 1*u by lra.
  apply/Rmult_le_compat_r; first lra.
  by rewrite -pow0 -pow1 -bpow_plus; apply/bpow_le; lia.
(* line 5 of 6 *)
have F2SC: (Fast2Sum_correct  p choice (sh xh) (sl xh xl)).
  apply/F2Sum_correct_abs=>//; try apply/generic_format_round.
    by  apply/RN_sym.
  rewrite (Rabs_pos_eq (sh xh)); try lra.
  apply/(Rle_trans _ ( 2 * u - 2 * u²))=>//.
  apply/(Rle_trans _ 1)=>//.
  suff : 2* u <= 1 by  rewrite /Rsqr;  lra.
  by rewrite -pow1 -pow0 -bpow_plus; apply/bpow_le; lia.
have h19_0: sh xh  + sl xh xl = /2*(sh xh  + x/(sh xh)) + e1/(2*(sh xh)) + e2.
  by rewrite /e1 /e2 /Rsqr; field; clear - shlb; lra.
have h19_1: sh xh + sl xh xl - sqrt x = 
            /(2*(sh xh)) * Rsqr (sh xh -sqrt x) + e1/(2*(sh xh))+e2.
  rewrite   h19_0 /Rsqr; field_simplify; try lra.
  by  rewrite -!Rsqr_pow2  Rsqr_sqrt; lra.
have h19: Rabs (sh xh + sl xh xl - sqrt x ) <=
         /(2 * sh xh) * (sh xh - sqrt x)² + Rabs (e1 / (2 * sh xh)) + Rabs e2.
  rewrite h19_1.
   apply/(Rle_trans _ _  _  (Rabs_triang _ _)).
  apply/Rplus_le_compat_r.
  apply/(Rle_trans _ _  _  (Rabs_triang _ _)).
  apply/Rplus_le_compat_r.
  rewrite Rabs_pos_eq //; try lra.
  by apply/Rmult_le_pos; try apply/Rle_0_sqr; lra.
case:binx =>xh1 xh4.
case: xh1=> xh1; last first.
(* xh = 1 essai *) 
  clear xh4 h13_3 h13_4 h12' cas4 cas3' cas3 cas2'.
  have {} cas1 : xl < 0 -> Rabs (sqrt x - sqrt xh) <= u / 2
  by apply/cas1; rewrite -xh1 //.
  have {} cas12:  0 <= xl -> Rabs (sqrt x - sqrt xh) <= u / 2 
  by apply/cas12; rewrite -xh1.
  have   {h13_1}[ h13_11 h13_12]:  
      Rabs xl <= u /\ sh xh < sqrt 2 + u / 2 by apply/h13_1; rewrite -xh1; lra.
  move/Rabs_le_inv:(h13_11)=> h13_1.
  have {}  h13_2 : (4 <= p)%Z -> Rabs (xl + ro1 xh) < 4 * u
    by apply/h13_2; rewrite -xh1; lra.
  have {}  h12 : Rabs (sqrt x - sh xh) < 3 / 2 * u
    by apply/h12; rewrite -xh1; lra.
  case : h15;[ move => h15|]; last lra.
  have R4_6_1: SQRTDWToFP xh xl  = pow 0 by apply/R4_6.
  rewrite /SQRTDWToFP pow0 in R4_6_1.
  have h1 : 1-u <= 1+xl <= 1+u by  lra.
  have h2 : sqrt (1-u) <= sqrt (x) <= sqrt (1+u).
    by split; rewrite /x -xh1;  apply/sqrt_le_1_alt;  lra.
  have h3:  (1 -u ) <sqrt (1 -u) <1.
    split; try( apply/sqrt_more; lra).
    by rewrite -[X in _ <X]sqrt_1; apply/sqrt_lt_1_alt;   lra.
  have h4: 1 < sqrt (1+u) < 1+u.
    split; try apply/sqrt_less_alt; try lra.
    by rewrite -{1}sqrt_1; apply/sqrt_lt_1_alt; (clear -ult1 upos;  lra).
  have h5: Rabs (1 -sqrt x) <= Rabs xl.
    rewrite /x -xh1  -Rabs_Ropp Ropp_minus_distr;apply/sqrt_var_maj.
    apply/(Rle_trans _ u)=>//.
    by rewrite /u -pow0; apply/bpow_le; lia.
  have sh1: sh xh = 1.
    by rewrite -xh1 sqrt_1 round_generic// -pow0; apply/format_bpow.
  have ro1_0: ro1 xh = 0.
  rewrite sh1  -xh1.
    have ->: 1 - 1^2 = 0 by   ring.
    by rewrite round_0.
  have ro2E: ro2 xh xl = xl by  rewrite ro1_0 Rplus_0_r round_generic.
  have e1E: e1 = 0 by rewrite /e1 /x ro2E sh1/Rsqr -xh1; ring.
  have ro2ub:  Rabs (ro2 xh xl * / (2 * sh xh)) <= u/2.
    rewrite ro2E sh1 Rmult_1_r Rabs_mult.
    apply/Rmult_le_compat=>//; try apply/Rabs_pos.
    by rewrite Rabs_pos_eq;lra.
  have {} e2ub: Rabs e2 <= (Rsqr u)/4.
    have-> : (Rsqr u)/4 =   / 2 * pow (- p) * pow (-p -1).
      rewrite /Rsqr /u bpow_plus powm1; lra.
    rewrite /e2; apply/error_le_half_ulp'_flx =>//.
    by rewrite bpow_plus powm1 -/u /Rdiv.
  rewrite  /rel_err_sqrtDW2 /SQRTDWToDW /err_sqrtDWFP /rel_err_sqrtDWFP   /SQRTDWToFP.
  rewrite F2SC; ring_simplify ((F2Sum_sum (sh xh) (sl xh xl) +
      (sh xh + sl xh xl - F2Sum_sum (sh xh) (sl xh xl)))).
  move: h19; rewrite   !Rabs_mult R4_6_1 -/x e1E Rabs_R0 Rmult_0_l Rplus_0_r.
  move=> h19.
  have sx0:  0 < sqrt x    by  lra.
  have sxi: 0 < /(sqrt x) by apply/Rinv_0_lt_compat.
  have h19_2: Rabs  (sh xh + sl xh xl - sqrt x) <= 11/8* Rsqr u.
    apply/(Rle_trans _  (/ (2 * sh xh) * (sh xh - sqrt x)² + Rabs e2))=>//.
    have ->:  11/8 * Rsqr u = 9/8 * Rsqr u + (Rsqr u)/4 by lra.
    apply/Rplus_le_compat=>//.
    rewrite sh1 Rmult_1_r.
    suff:  ( 1- sqrt x)² <= 9 * / 4 * u² by lra.
    have-> :  9 * / 4 * u² = Rsqr (3/2 *u).
      rewrite /Rsqr ; field. 
    apply/Rsqr_le_abs_1.
    apply/(Rle_trans _ (Rabs xl))=>//.
    apply/(Rle_trans _ u) =>//.
    by rewrite Rabs_pos_eq; lra.
  split.
    apply/(Rle_trans _ ( (11 / 8 * u²)/(sqrt x))).
      apply/Rmult_le_compat =>//; try apply/Rabs_pos.
       rewrite Rabs_pos_eq; lra.
    rewrite Rsqr_pow2.
    suff : 11/sqrt x <= 25 by clear -upos; nra.
    apply/(Rmult_le_reg_r (sqrt x))=>//.
    by rewrite Rmult_assoc Rinv_l ?Rmult_1_r;lra.
(* algo 7*)
  have usqrtx: u <= ulp (sqrt x).
    have ->: u = ulp (/2) by rewrite -powm1 ulp_bpow /fexp /u; congr bpow.
    apply/ulp_le; rewrite !Rabs_pos_eq; try apply/sqrt_ge_0; last lra.
    apply/Rsqr_incr_0;  try apply/sqrt_ge_0; last lra.
    rewrite Rsqr_sqrt; last by  lra.
    apply/(Rle_trans _ (1-u))=>//.
    suff: u <= /2 by  rewrite /Rsqr; lra.
      rewrite /u -powm1; apply/bpow_le; lia.
    lra.
  rewrite Rmult_1_l.
  rewrite -Rmult_plus_distr_r.
  have  :  Rabs (sh xh + sl xh xl - sqrt x) <= 11 / 8 * u* (ulp (sqrt x)).
    apply/(Rle_trans _  (11 / 8 * u²))=>//.
    by rewrite /Rsqr -Rmult_assoc; apply/Rmult_le_compat_l=>//; lra.
  have h00: Rabs (RN (sh xh + sl xh xl) - (sh xh + (sl xh xl)))<=
          /2 * ulp (sh xh + sl xh xl).
    by   apply/error_le_half_ulp.
  have slE': sl xh xl = xl/2 .
    rewrite  ro2E sh1 round_generic Rmult_1_r //.
    by rewrite /Rdiv -powm1; apply/formatS.
  move=>h.
  have ->: 1 - sqrt x = 1 - (sh xh + sl xh xl) +
         (sh xh + sl xh xl - sqrt x) by ring.
  have haux':   Rabs
    (1 - (sh xh + sl xh xl) +
     (sh xh + sl xh xl - sqrt x)) <= (/2 + 11/8*u)*ulp (sqrt x).
    apply/(Rle_trans _ _  _  (Rabs_triang _ _)).
    rewrite Rmult_plus_distr_r.
    apply/Rplus_le_compat=>//.
    rewrite slE' sh1.
    ring_simplify (1 - (1 + xl / 2)).
    by rewrite Rabs_Ropp Rabs_mult (Rabs_pos_eq (/2)); lra.
  have haux:   Rabs
    (1  - (sh xh + sl xh xl) +
     (sh xh + sl xh xl - sqrt x)) <=
       1 * / 2 * ulp (sqrt x) + 175 * / 100 * u * ulp (sqrt x).
    apply/(Rle_trans _ ((/2 + 11/8*u)*ulp (sqrt x)))=>//.
    rewrite -Rmult_plus_distr_r.
    apply/Rmult_le_compat_r; first by  apply/ulp_ge_0.
    rewrite Rmult_1_l; apply/Rplus_le_compat_l.
    by lra.
  split; first  lra.
  rewrite slE' sh1 .
  field_simplify  (1 - (1 + xl / 2) + (1 + xl / 2 - sqrt x)).
  have->: ((-2 * sqrt x + 2) / 2 = 1 - sqrt x ) by field.
  rewrite (Rabs_pos_eq  (/ sqrt _)); last  by lra.
  apply/(Rle_trans _ (u /sqrt x)).
    by apply/Rmult_le_compat; try apply/Rabs_pos; lra.
  suff: /sqrt x <= 1 + 17/8* u + 33/8 * u^2 by clear -upos; nra.
  apply/(Rle_trans _ (/ (1-u))).
    apply/Rinv_le;  lra.
  apply/(Rmult_le_reg_r   (1 -u)); try lra.
  rewrite Rinv_l ; try lra.
  field_simplify.
  suff: 33 * u ^ 2  <=  16 * u  + 9  by  clear -upos; nra.
  suff : u <= /2 by clear -upos; nra.
  by rewrite /u -powm1 ; apply/bpow_le; lia.
(* 1 < xh *)
have xhge: 1 + 2*u <= xh.
  have ->: 1+2*u = succ two fexp 1 by rewrite succ_FLX_1 bpow_plus -/u pow1; ring.
  by apply/succ_le_lt=>// ; rewrite -pow0; apply/format_bpow.
move/Rabs_le_inv:(dw_ulp_flx DWx)=> xlb.
have xge: 1 + u <= x.
  rewrite /x.
  case: (Rlt_le_dec xh 2)=> xh2.
    move:xlb; rewrite ulp_neq_0 /cexp/fexp;try (clear - xh1; lra).
    by rewrite (mag_unique_pos _ _ 1) ?pow0 ?bpow_plus -/u pow1;
       clear -   xhge xh1 xh2; lra.
  move:xlb; rewrite ulp_neq_0 /cexp/fexp;try (clear - xh1; lra).
  rewrite (mag_unique_pos _ _ 2) ?pow1  ?bpow_plus -/u pow2;
  try (clear -   xhge xh1 xh2 upos xh4; lra).
  move=>h.
  apply/(Rle_trans _ (2 - 2*u)); try (clear -h xh2 upos; lra).
  suff : u <= /4 by clear -upos;lra.
  by change( pow (-p) <= pow (-2)); apply/bpow_le; lia.
rewrite  /rel_err_sqrtDW2 /SQRTDWToDW /err_sqrtDWFP /rel_err_sqrtDWFP   /SQRTDWToFP.
rewrite F2SC; ring_simplify ((F2Sum_sum (sh xh) (sl xh xl) +
      (sh xh + sl xh xl - F2Sum_sum (sh xh) (sl xh xl)))).
have h20: xh <= 2 -2*u ->   Rabs (sh xh + sl xh xl - sqrt x) <= 25/8 * Rsqr u.
  move=> xh2.
  apply/(Rle_trans _ (/2 * Rsqr (3/2 *u) + 2*Rsqr u)).
    apply/(Rle_trans _ _ _ h19).
    have ->:   / 2 * (3 / 2 * u)² + 2 * u² =   / 2 * (3 / 2 * u)² + Rsqr u + Rsqr u 
     by ring.
    apply/Rplus_le_compat=>//.
    rewrite Rabs_mult Rmult_comm (Rabs_pos_eq ( /_)); 
  last by apply/Rlt_le/Rinv_0_lt_compat; try (clear - shlb ; lra).
    rewrite -Rmult_plus_distr_r.
    apply/(Rle_trans _  (((sh xh - sqrt x)² + Rabs e1) */2)).
      apply/Rmult_le_compat_l.
        apply/Rplus_le_le_0_compat.
          by apply/Rle_0_sqr.
        by apply/Rabs_pos.
      by apply/Rinv_le;  clear -shlb ; lra.
    suff :  ((sh xh - sqrt x)² + Rabs e1)  <= (3 / 2 * u)² +  2*u² by  clear -upos; lra.
    case:h15; try (clear -xh2; lra); move =>[_ h15].
    apply/Rplus_le_compat=>//.
    apply/Rsqr_le_abs_1.
    move:h12; try (clear -xh2; lra).
    by rewrite -Rabs_Ropp Ropp_minus_distr (Rabs_pos_eq (_*_)); (clear -upos  xh2;  lra).
  by rewrite /Rsqr; field_simplify; clear -upos; lra.
have hi2sh: 0 <= /(2* sh xh) 
     by apply/Rlt_le/Rinv_0_lt_compat; try (clear - shlb ; lra).
have sqrt2gt1: 1 < sqrt 2 by  rewrite -sqrt_1; apply/sqrt_lt_1_alt; lra.
have h21: 2 <= xh -> Rabs (sh xh + sl xh xl - sqrt x) <= 7/2 * Rsqr u.
  move=> xh2.
  have hi2sh':  / (2 * sh xh) <= / (2 * (sqrt 2 - u)).
    apply/Rinv_le.
      suff: u < sqrt 2 by clear -upos; lra.
      rewrite -(Rabs_pos_eq u ) ;last by clear -upos; lra.
      rewrite -(Rabs_pos_eq (sqrt 2)); last by apply/sqrt_ge_0.
      apply/Rsqr_lt_abs_0; rewrite Rsqr_sqrt /Rsqr /u; try (clear -upos; lra).
      rewrite -pow1 -bpow_plus.  
      by apply/bpow_lt; lia.
    suff : sqrt 2  - sh xh <= u by lra.
    apply /(Rle_trans _ (sqrt xh - sh xh)).
      apply/Rplus_le_compat_r.
      by apply/sqrt_le_1_alt.
    suff : -u <= sqrt xh - sh xh <= u by clear -upos; lra.
    apply/Rabs_le_inv.
    rewrite -Rabs_Ropp Ropp_minus_distr.
    suff ->: u = /2 * ulp (sqrt xh) by apply/error_le_half_ulp.
    rewrite ulp_neq_0 /cexp/fexp; last first.
      suff: 0 < sqrt xh by clear -upos; lra.
      by apply/sqrt_lt_R0; clear -xh2; lra.
    rewrite (mag_unique_pos _ _ 1); rewrite ?pow0 ?bpow_plus pow1 -/u; first field.
    rewrite -sqrt_1; split.
      by apply/sqrt_le_1_alt; clear -xh1; lra.
    suff -> : 2 = sqrt 4 by apply/sqrt_lt_1_alt; clear -xh4  xh1 upos; lra.
    have-> : 4 = Rsqr 2 by rewrite /Rsqr ; ring.
    by rewrite sqrt_Rsqr //; lra.
  apply/(Rle_trans _ (/(2 * (sqrt 2 -u)) * Rsqr (u* (1 + (sqrt 2)/2)) + 
    /(2 * (sqrt 2 -u)) * 4* Rsqr u + Rsqr u)).
    apply/(Rle_trans _ _ _ h19).
    apply/Rplus_le_compat=>//.
    apply/Rplus_le_compat; last first.
      case:h15; try (clear -upos xh2 ; lra); move=>[_ h].
      rewrite Rmult_assoc (Rmult_comm (/_)).
      rewrite Rabs_mult (Rabs_pos_eq (/_) ) //.
      by apply/Rmult_le_compat =>//; try apply/Rabs_pos.
    apply/Rmult_le_compat=>//; try apply/Rle_0_sqr.
    apply/Rsqr_le_abs_1.
    rewrite (Rabs_pos_eq (_ *_)).
      rewrite -Rabs_Ropp Ropp_minus_distr.
      by  apply/Rlt_le/h12'; clear -xh2 upos; lra.
    have : 0 <= sqrt 2 by apply/sqrt_ge_0.
    clear -upos; nra.
  set d:= /(2*_).
  rewrite Rsqr_mult (Rmult_comm (Rsqr u))  -Rmult_assoc -(Rmult_plus_distr_r ).
  have->:  (d * (1 + sqrt 2 / 2)² + d * 4) * u² + u²  =  (d * (1 + sqrt 2 / 2)² + d * 4 +1) * u² by ring.
  apply/Rmult_le_compat_r; try apply/Rle_0_sqr.
  rewrite /d /Rsqr.
  field_simplify; last by  clear -   sqrt2gt1 ult1; lra.
  rewrite -Rsqr_pow2 Rsqr_sqrt; last by lra.
  apply/(Rmult_le_reg_r (8 * sqrt 2 - 8 * u)); first by clear -  sqrt2gt1 ult1; lra.
  rewrite /Rdiv Rmult_assoc Rinv_l ?Rmult_1_r; last by  clear -   sqrt2gt1 ult1; lra.
  suff :   (22 + 12 * sqrt 2 - 8 * u)  <= 28 * sqrt 2 - 28 * u by clear -upos; lra.
  suff: 22 + 20* u <= 16 * sqrt 2 by clear -upos; lra.
  suff h: u<= /32 .
    apply/(Rle_trans _ (22+20/32)).
      lra.
    interval.
  by change (pow (-p) <= pow (-5)); apply/bpow_le; lia.
have sqxgt0: 0 < sqrt x by apply/sqrt_lt_R0;  clear - xge upos ; lra.
have usqx : ulp (sqrt x) = 2*u.
  rewrite ulp_neq_0 /cexp /fexp ;last by clear -  sqxgt0; lra.
  rewrite (mag_unique_pos _ _ 1) ?pow0 ?bpow_plus pow1 -/u //.
  rewrite -sqrt_1; have-> : 2 = sqrt 4.
    have -> : 4 = Rsqr 2 by rewrite /Rsqr; ring.
    by rewrite sqrt_Rsqr //; lra.
  split.
    by apply/sqrt_le_1_alt; clear -upos xge; lra.
  apply/sqrt_lt_1; try (clear -xge  upos; lra).
  case: (Rle_lt_dec 4 x)=>//h.
  have : RN 4 <= RN x by apply/round_le.
  rewrite /x -xE round_generic.
    clear -xh4 upos; lra.
  by rewrite -pow2; apply/format_bpow.
have sqxh1  : 1< sqrt xh.
  by rewrite -sqrt_1; apply/sqrt_lt_1;  (clear -xge  upos xh1; lra).
have shb: 1 <= sh xh <= 2 -2*u .
  have ->: 1 = RN 1 by rewrite -pow0 round_generic //; apply/format_bpow.
  split ; first by apply/round_le/Rlt_le.
  apply/round_N_le_midp.
    by rewrite p2; apply/generic_format_pred/format_bpow.
  rewrite {2}p2 succ_pred; last by apply/format_bpow.
  by rewrite pow1; clear -  sqrt_ub; lra.
have shpsl : 1 - 2*u + Rsqr u <= sh xh + sl xh xl <= 2 - Rsqr u.
  move/Rabs_le_inv:slub.
  by clear - shb ; rewrite /Rsqr; nra.
move: shpsl.
have->: 1 -2*u + Rsqr u = Rsqr (1 -u) by rewrite /Rsqr ; ring.
move=> shpsl.
have h4_5:   Rabs (RN (sh xh + sl xh xl) - sqrt (xh + xl)) <=
              1 * / 2 * ulp (sqrt x) + 175 * / 100 * u * ulp (sqrt x).
  rewrite usqx; field_simplify.
  have->: (350 * u ^ 2 + 100 * u) / 100 = 7/2 * Rsqr u + u by rewrite /Rsqr; field.
  have->:  RN (sh xh + sl xh xl) - sqrt (xh + xl) =  
          RN (sh xh + sl xh xl) -  (sh xh + sl xh xl) +
       ((sh xh + sl xh xl)- sqrt (xh + xl)) by ring.
  apply/(Rle_trans _ _  _  (Rabs_triang _ _)).
  rewrite Rplus_comm.
  apply/Rplus_le_compat.
    case: (Rle_lt_dec 2 xh )=> xh2.
      by apply/h21.
    apply/(Rle_trans _  (25 / 8 * u²)).
      apply/h20.
      have->: 2 -2*u = pred two fexp (pow 1).
        by rewrite pred_bpow /fexp bpow_plus pow1.
      by apply/pred_ge_gt =>//; apply/format_bpow.
    clear -upos; rewrite /Rsqr; nra.
  apply/(Rle_trans _ (/2* ulp (sh xh + sl xh xl))).
    by apply/error_le_half_ulp.
  case: (Rlt_le_dec (sh xh + sl xh xl) 1)=>s1; rewrite ulp_neq_0 /cexp;
    try (move:shpsl; clear -upos ult1 ; rewrite /Rsqr ; nra).
    rewrite (mag_unique_pos _ _ 0)  bpow_plus pow0 -/u ?powm1;
       first by (clear -upos; lra).
    split=>//.
    apply/(Rle_trans _ (1 - u)²); last by clear - shpsl; lra.
    rewrite Rmult_1_l.
    rewrite /Rsqr; ring_simplify.
    suff : u <= /4 by clear -upos u2pos; lra.
    by change (pow (-p) <= pow (-2));apply/bpow_le; lia.
  by rewrite bpow_plus -/u (mag_unique_pos _ _ 1) ?pow0 pow1;
   move: shpsl;  clear -upos s1  upos u2pos;  rewrite /Rsqr; try nra.
have x1: 1 < x by clear -xge upos; lra.
have sqx1: 1 < sqrt x .
  rewrite -sqrt_1 ; apply/sqrt_lt_1_alt; clear -x1; lra.
have isqx1: 0 < /sqrt x < 1.
  split; first by  apply/Rinv_0_lt_compat; clear -sqx1; lra.
  rewrite -Rinv_1; apply/Rinv_lt; clear -sqx1; lra.
have h4_4:
  Rabs ((sh xh + sl xh xl - sqrt (xh + xl)) / sqrt (xh + xl)) <=
  25 * / 8 * u ^ 2.
(* Algo 6*)
  rewrite Rabs_mult ; rewrite (Rabs_pos_eq (/_));
    last  by clear -isqx1; rewrite  -/x; lra.
  case: (Rle_lt_dec xh (2-2*u))=>xh2.
    rewrite -[X in _ <= X]Rmult_1_r -/x; apply/Rmult_le_compat; try apply/Rabs_pos.
    + clear -   isqx1; lra.
    + have->:  25 * / 8 * u ^ 2 =  25 / 8 * u² by rewrite /Rsqr; field.
      by apply/h20.
    clear -isqx1; lra.
  have xh2': 2 <= xh.
    have->: 2 = succ two fexp (2 -2*u).
      by rewrite p2 succ_pred; try  apply/format_bpow. 
    by apply/succ_le_lt=>//; rewrite p2; apply/generic_format_pred/format_bpow.
  have x_lb: 2-u <= x.
    case: (Rle_lt_dec (2 -u) x) =>// h.
    suff: RN x <= 2 - 2*u by rewrite /x -xE; clear -upos xh2'; lra.
    rewrite p2; apply/round_N_le_midp; 
      try by apply/generic_format_pred/format_bpow.
    rewrite succ_pred ; last by apply/format_bpow.
    by rewrite -p2 pow1; clear -h; lra.
  have sqx_lb :  sqrt (2 -u) <= sqrt x by apply/sqrt_le_1_alt.
  have isqxb: /(sqrt x) <= /(sqrt (2-u)).
    apply/Rinv_le=>//.
    apply/sqrt_lt_R0; clear -upos ult1; lra.
  apply/(Rle_trans _( (7/2* Rsqr u)*/(sqrt (2 -u)))).
    apply/Rmult_le_compat; rewrite -/x //; try apply/Rabs_pos.
      by clear - isqx1; lra.
    by apply/h21.
    rewrite /Rsqr.
  have sq2pos:  0 < sqrt (2 - u) by apply/sqrt_lt_R0; clear -ult1; lra.
  suff:   7 / 2 *  / sqrt (2 - u) <= 25 * / 8 by clear -upos; nra.
  suff h :  u <= /2.
    apply/(Rmult_le_reg_r (sqrt (2 -u)))=>//.
    rewrite Rmult_assoc Rinv_l ?Rmult_1_r.
      suff: 56/50 <= sqrt (2 -u) by lra.
      apply/Rsqr_incr_0_var.
        rewrite Rsqr_sqrt /Rsqr; lra.
      by apply/sqrt_ge_0.
    lra.
  rewrite -powm1; apply/bpow_le; lia.
split=>//; split=>//.
(* Algo 7*)
move:h4_4; rewrite !Rabs_mult (Rabs_pos_eq (/_)); 
   last by rewrite -/x;  clear - isqx1; lra.
rewrite -/x=> h4_4.
have {} h4_4: Rabs (sh xh + sl xh xl - sqrt x) <=  25/ 8 * Rsqr u * sqrt x.
  apply/(Rmult_le_reg_r (/sqrt x)) ; first by clear - isqx1; lra.
  rewrite Rmult_assoc Rinv_r ?Rmult_1_r //; last by clear - sqx1; lra.
  suff ->:  25 / 8 * u² = 25 * / 8 * u ^ 2 by [].
   by rewrite /Rsqr ; field.
apply/(Rmult_le_reg_r (sqrt x)); first by  (clear -sqx1; lra).
rewrite  Rmult_assoc Rinv_l ?Rmult_1_r; 
    last by rewrite -/x ; clear - sqx1; lra.
have haux:  Rabs (RN (sh xh + sl xh xl) - (sh xh + sl xh xl)) <=
     u/(1+u)*(sh xh + sl xh xl).
  rewrite -{3}(Rabs_pos_eq (sh xh + sl xh xl)); 
   last by  move:shpsl; rewrite /Rsqr ; clear -ult1; try  nra.
  have ->: u =   u_ro two p by rewrite /u_ro bpow_plus -/u pow1; field.
  by apply/relative_error_N_FLX';  lia.
have->:  RN (sh xh + sl xh xl) - sqrt (xh + xl) =  
          RN (sh xh + sl xh xl) -  (sh xh + sl xh xl) +
       ((sh xh + sl xh xl)- sqrt (xh + xl)) by ring.
apply/(Rle_trans _ _  _  (Rabs_triang _ _)).
apply/(Rle_trans _ (   u / (1 + u) * (sh xh + sl xh xl) +
                    25 / 8 * u² * sqrt x)).
  by apply/Rplus_le_compat.
apply/(Rle_trans _ ( u / (1 + u) * (1 + 25/8* Rsqr u) * (sqrt x) +
 25 / 8 * u² * sqrt x)).
  apply/Rplus_le_compat_r.
  rewrite Rmult_assoc; apply/Rmult_le_compat_l.
    by apply/Stdlib.Rdiv_pos_compat; clear -upos ult1; lra.
  by move/Rabs_le_inv:h4_4; clear -upos; lra.
rewrite -Rmult_plus_distr_r.
apply/Rmult_le_compat_r; first  by clear -sqx1; lra.
rewrite /Rsqr ; field_simplify; try (clear -upos ult1;  nra).
have->: (8* u + 8) = (u +1)*8 by ring.
rewrite /Rdiv Rinv_mult_distr -?Rmult_assoc; try (clear -upos ult1; lra).
apply/Rmult_le_compat_r; first by lra.
apply/(Rmult_le_reg_r (u + 1)); first by (clear -upos ult1; lra).
rewrite Rmult_assoc Rinv_l ?Rmult_1_r; last by (clear -upos ult1; lra).
by clear -upos ult1; nra.
Qed.

Section Th_3_10.

Variable nu x xh xl: R.
Hypothesis DWx : (double_word xh xl).
Notation Rx := (SQRTDWToFP xh xl).
Hypothesis nu1: nu * Rsqr u < 1.
(* Hypothesis x_approx: Rabs ((xh + xl) -x ) < nu* Rsqr u * x. *)
Hypothesis x_approx: Rabs ((xh + xl) -x ) <= nu* Rsqr u * x.

Hypothesis xpos: 0 < x.
Hypothesis xhpos: 0 < xh.
Hypothesis Hp5 : (5 <= p)%Z.

(* Eq 38 *)
Theorem Th_3_10_aux :
  Rabs (sqrt (xh + xl) - sqrt x) <= nu*(Rsqr u)/(1 - nu * Rsqr u) * sqrt x.
Proof.
have upos:= (upos p).
rewrite -/u in upos.
have u2pos: 0 < Rsqr u by rewrite /Rsqr; nra.
have nuupos: 0 <= nu * Rsqr u.
  apply/( Rmult_le_reg_r x) =>//.
  apply/(Rle_trans _  (Rabs (xh + xl - x)))=> //.
  by rewrite Rmult_0_l; apply/Rabs_pos.
case:(MVT_abs (fun x =>  sqrt x) (fun  x  => / (2* sqrt x)) x (xh + xl)  ).
  move=> c hc;  apply/derivable_pt_lim_sqrt.
  apply/(Rlt_le_trans _ (Rmin x (xh + xl)  )); last  lra.
  by apply/Rmin_glb_lt; move:(DWpos DWx); lra.
  move=> zeta.
move=>[h11 hzeta].
have z0 : 0 < zeta.
  apply/(Rlt_le_trans _ (Rmin x (xh + xl) )); last  lra.
  by apply/Rmin_glb_lt;  move:(DWpos DWx);lra.
have sqrtz0 : 0 < sqrt zeta by apply/sqrt_lt_R0.
have hsqrtzi:  0 < / (2* (sqrt zeta)).
  apply/Rinv_0_lt_compat.
  suff: 0<sqrt zeta  by lra.
  by apply/sqrt_lt_R0.
have sqrtx0:=(sqrt_lt_R0 _ xpos).
rewrite h11 (Rabs_pos_eq (/_)); last first.
  apply/Rlt_le/Rinv_0_lt_compat.
  suff: 0<sqrt zeta  by lra.
  by apply/sqrt_lt_R0.
apply/(Rle_trans _  (( nu * u² * x)/(2 * sqrt zeta))).
  rewrite Rmult_comm; apply Rmult_le_compat_r; lra.
have hx: x*(1- nu * Rsqr u) <= zeta.
  move: hzeta  x_approx; case:(Rle_lt_dec (xh + xl) x) => hx.
    rewrite Rmin_right ?Rmax_left; try lra.
    by rewrite -Rabs_Ropp Ropp_minus_distr Rabs_pos_eq;  lra.
  by rewrite Rabs_pos_eq ?Rmin_left ?Rmax_right; lra.
have hx2: sqrt x * (1 - nu*Rsqr u) <= sqrt zeta.
  apply/(Rle_trans _ ((sqrt x) * sqrt (1 - nu*Rsqr u))).
    apply/Rmult_le_compat_l.
      by apply/sqrt_ge_0.
    case: nuupos=>h; first last.
      rewrite -h Rminus_0_r sqrt_1; lra.
    apply/Rlt_le/sqrt_more; lra.
  rewrite -sqrt_mult; try lra.
  by apply/sqrt_le_1_alt.
have->:  nu * u² * x / (2 * sqrt zeta) =  (nu * u² )* (x / (2 * sqrt zeta)) by field; lra.
have->: nu * u² / (1 - nu * u²) * sqrt x = nu * u² *(/ (1 - nu * u²) * sqrt x) by field; lra.
apply/Rmult_le_compat_l; try lra.
have -> :  x / (2 * sqrt zeta) =  / (2 * sqrt zeta) * sqrt x * sqrt x.
rewrite -{1}(sqrt_sqrt x); try lra.
apply/Rmult_le_compat_r; try lra.
apply/(Rmult_le_reg_r (/sqrt x)).
  by apply/Rinv_0_lt_compat.
rewrite Rmult_assoc Rinv_r; try lra.
rewrite Rmult_1_r.
rewrite -Rinv_mult_distr; try lra.
apply/Rinv_le; try nra.
Qed.


Theorem Th_3_10:
  Rabs (Rx - sqrt x) <= 
        ((u + 17/8*(Rsqr u) + 33/8* (u^3))*(1 + nu*(Rsqr u)/(1 - nu * Rsqr u))
         + nu*(Rsqr u)/(1 - nu * Rsqr u))* sqrt x.
Proof.
have xhlpos:=(DWpos DWx xhpos).
have sxhl:=(sqrt_lt_R0 _ xhlpos).
have sxhli:=(Rinv_0_lt_compat _ sxhl).
have upos:= (upos p).
rewrite -/u in upos.
have u2pos: 0 < Rsqr u by rewrite /Rsqr; nra.
have nuupos: 0 <= nu * Rsqr u.
  apply/( Rmult_le_reg_r x) =>//.
  apply/(Rle_trans _  (Rabs (xh + xl - x)))=> //.
  by rewrite Rmult_0_l; apply/Rabs_pos.
have->: (Rx -sqrt x )  = (Rx - (sqrt (xh + xl)) )+ ((sqrt (xh + xl) - sqrt x)) by ring.
apply/(Rle_trans _ _  _  (Rabs_triang _ _)).
have h33: Rabs (Rx - (sqrt (xh + xl))) <=   
            (u + 17 / 8 * u² + 33 / 8 * u ^ 3)* (sqrt (xh + xl)).
  case:(SQRTDW_bounds DWx  xhpos)=>//[ h1 [h2 h3]].
  apply/(Rmult_le_reg_r (/ sqrt (xh + xl)))=>//.
  rewrite Rmult_assoc Rinv_r; try lra.
  move:(h3).
  rewrite / rel_err_sqrtDWFP /Rdiv /Rsqr Rmult_1_r Rabs_mult (Rabs_pos_eq (/_)); lra.
have h36:=Th_3_10_aux.
apply/(Rle_trans _ ((u + 17 / 8 * u² + 33 / 8 * u ^ 3) * 
                       sqrt (xh + xl)+ nu * u² / (1 - nu * u²) * sqrt x)).
  by apply/Rplus_le_compat.
move/Rabs_le_inv: (h36).
set A:= (u + 17 / 8 * u² + 33 / 8 * u ^ 3).
set B :=   nu * u² / (1 - nu * u²).
move=> h.
apply/(Rle_trans _ (  A * (1+B)* sqrt x + B * sqrt x)).
  apply/Rplus_le_compat_r.
  rewrite Rmult_assoc.
  apply/Rmult_le_compat_l; try lra; rewrite /A;  nra.
lra.
Qed.



(* 1.1 *)
Lemma rel_err_ulp t (t0: t <> 0):
Rabs (RN t  -t) <= (Rabs ((RN t -t)/t))/u * ulp t.
Proof.
have upos : 0 < u    by  apply/bpow_gt_0.
rewrite Rabs_mult !Rmult_assoc -[X in X <= _]Rmult_1_r.
apply/Rmult_le_compat_l; first by apply/Rabs_pos.
apply/(Rmult_le_reg_l (Rabs t)).
  by apply/Rabs_pos_lt.
rewrite -!Rmult_assoc Rmult_1_r -Rabs_mult Rinv_r // (Rabs_pos_eq 1)?Rmult_1_l; try lra.
rewrite Rmult_comm; apply/(Rmult_le_reg_r u)=>//.
rewrite Rmult_assoc Rinv_l ?Rmult_1_r; try lra.
by  apply/ulp_FLX_ge.
Qed.



Theorem Th_3_10_bis (hunu: 2*u*nu <1):
  Rabs (Rx - sqrt x) <= (/2 + u *(7/4 + nu/(1-nu*Rsqr u)))* ulp(sqrt x).
Proof.
case:DWx =>[[Fxh Fxl] xE].
have upos : 0 < u    by  apply/bpow_gt_0.
have ult1: u < 1 by rewrite -pow0 /u; apply/bpow_lt; lia.
have nuupos: 0 <= nu .
  apply/( Rmult_le_reg_r (Rsqr u)).
    rewrite /Rsqr; nra.
  apply/( Rmult_le_reg_r x); try  nra.
  rewrite !Rmult_0_l.
  by apply/(Rle_trans _  (Rabs (xh + xl - x)))=>//; apply/Rabs_pos.
case:(Req_dec Rx (RN (sqrt x)))=> RxRN.
(* case for the need of the proof *)
  rewrite  RxRN; apply/(Rle_trans _ (/2 * ulp (sqrt x))).
    by apply/error_le_half_ulp.
  apply/Rmult_le_compat_r; first by apply/ulp_ge_0.
  suff:  0<=  nu / (1 - nu * u²) by nra.
  by apply/Rcomplements.Rdiv_le_0_compat; lra.
have->: (Rx -sqrt x )  = (Rx - (sqrt (xh + xl)) )+ ((sqrt (xh + xl) - sqrt x)) by ring.
apply/(Rle_trans _ _  _  (Rabs_triang _ _)).
have h3_6: Rabs (Rx - (sqrt (xh + xl))) <=  (/2 + 7/4*u)* ulp (sqrt (xh + xl)).
  case:(SQRTDW_bounds DWx  xhpos)=>//[ h1 [h2 h3]].
  move:(h2).
  rewrite /err_sqrtDWFP; lra.
have h36:=Th_3_10_aux.
have h:  Rabs (sqrt (xh + xl) - sqrt x) <= (nu * u)/(1 - nu * Rsqr u)* ulp (sqrt x).
  apply/(Rle_trans _ (nu * u² / (1 - nu * u²) * sqrt x))=>//.
  have->:  nu * u² / (1 - nu * u²) =  (nu * u / (1 - nu * u²))  * u.
    rewrite /Rsqr; field.
    by move: nu1; rewrite /Rsqr; lra.
  rewrite Rmult_assoc.
  apply/Rmult_le_compat_l.
    apply/Rmult_le_pos; try nra.
    apply/Rlt_le/Rinv_0_lt_compat; lra.
  have->:   u * sqrt x = Rabs(sqrt x) * u.
    by rewrite Rabs_pos_eq; try ring; apply/sqrt_ge_0. 
  by apply/ulp_FLX_ge.
case:(Rle_lt_dec (ulp (sqrt (xh + xl))) (ulp (sqrt x)))=>hulp.
  have h1: Rabs (Rx - sqrt (xh + xl)) <= (/ 2 + 7 / 4 * u) * ulp (sqrt x).
    apply/(Rle_trans _ ((/ 2 + 7 / 4 * u) * ulp (sqrt (xh + xl))))=>//.
    by apply/(Rmult_le_compat_l);  lra.
  suff: (/ 2 + 7 / 4 * u) * ulp (sqrt x) + nu * u / (1 - nu * u²) * ulp (sqrt x) = 
 (/ 2 + u * (7 / 4 + nu / (1 - nu * u²))) * ulp (sqrt x) by lra.
  field; lra.
case:(Rle_lt_dec (xh + xl) x)=>hx_xhxl.
  suff : ulp ( sqrt (xh + xl))  <= ulp(sqrt x) by lra.
  apply/ulp_le; rewrite !Rabs_pos_eq; try apply/sqrt_pos.
  by apply/sqrt_le_1_alt.
(* cas sqrt in <> bin et x < xh + xl *)
have sx0:=(sqrt_lt_R0 _ xpos).
have xhl0: 0 < xh + xl by lra.
have sxhl0:=(sqrt_lt_R0 _ xhl0).
pose ks:= (mag two (sqrt x)).
have xhlltpowk:sqrt  x< pow ks by case:(mag_pos_bounds  sx0).
have h1: (  ks<  (mag two (sqrt(xh + xl))))%Z.
  move:hulp.
  rewrite !ulp_neq_0; try lra.
  move/lt_bpow.
  rewrite /cexp /fexp; try lia.
  move=> h1.
  rewrite /ks; lia.
have {}h1 :(ks +1 <= mag two (sqrt (xh + xl)))%Z by lia.
have: pow ks <= sqrt (xh +xl).
  apply/(Rle_trans _ (pow ( (mag two (sqrt (xh + xl))) -1))).
    apply/bpow_le; lia.
  rewrite -{3}(Rabs_pos_eq (sqrt (xh + xl))); try lra.
  apply/bpow_mag_le; lra.
move=>h2.
have: Rsqr( sqrt x) <  Rsqr (pow ks).
  by apply/Rsqr_incrst_1; try apply/sqrt_ge_0; try apply/bpow_ge_0.
rewrite Rsqr_sqrt /Rsqr; try lra.
rewrite -bpow_plus.
set k := (ks +ks)%Z.
move =>xpk.
have:  Rsqr (pow ks) <= Rsqr (sqrt (xh + xl)).
  by apply/Rsqr_incr_1; try apply/sqrt_ge_0; try apply/bpow_ge_0.
rewrite Rsqr_sqrt /Rsqr; try lra.
rewrite -bpow_plus -/k.
move =>pkxhl.
have h3: xh + xl <= (1 + nu*Rsqr u) * x.
  by move:(x_approx); rewrite Rabs_pos_eq; lra.
have {} h3: xh + xl < (1 + u/2)*x.
  apply/(Rle_lt_trans _  ((1 + nu * u²) * x))=>//.
  apply/Rmult_lt_compat_r=>//.
  by rewrite /Rsqr; nra.
have {} h3: xh + xl < ( 1 + u/2) * pow k by nra.
have Fpowk: format (pow k) by  apply/generic_format_bpow;  rewrite /fexp; lia.
have xhE: xh = pow k.
  case: pkxhl =>  pkxhl; last first.
    rewrite xE round_generic //  -pkxhl; apply/generic_format_bpow; rewrite /fexp; lia.
  have succpk: succ two fexp (pow k) = (1 + 2* u)* pow k.
    rewrite succ_eq_pos; last by apply/bpow_ge_0.
    rewrite ulp_bpow /fexp.
    rewrite (bpow_plus _ _ (-p)) (bpow_plus _ k) pow1 -/u; ring.
  have hfloor:  round two fexp Zfloor (xh + xl) = pow k.
    apply/        round_DN_eq =>//.
    rewrite succpk; lra.
  rewrite xE.
  rewrite -hfloor.
  apply/        round_N_eq_DN.
  pose up:=  (1 + 2 * u) * pow k.
  have h4: pred two  fexp up < xh + xl  <= up.
   rewrite /up - {1}   succpk pred_succ //; lra.
  have hceil:  round two fexp Zceil (xh + xl) = up.
    apply/round_UP_eq=>//.
    by rewrite /up -succpk;apply/generic_format_succ.
  rewrite hfloor hceil /up;  lra.
have xlpos: 0 <= xl by lra.
have RxE: Rx = pow ks.
  rewrite (R4_6 ks) // xhE.
  congr bpow; rewrite /k; ring.
have h4: pow k /( 1 + u/2) < x.
  apply/(Rmult_lt_reg_r (1 + u/2)); try lra.
  rewrite Rmult_assoc Rinv_l ?Rmult_1_r; try lra.
  apply/(Rle_lt_trans _ (( 1 + nu * Rsqr u )*x)).
    move: x_approx.
    rewrite Rabs_pos_eq; lra.
  have: (1 + nu * u²)< ( 1 + u/2) by rewrite/Rsqr ;  nra.
  nra.
have hlsx: sqrt x <= pow ks.
  apply/Rsqr_incr_0_var; try apply/bpow_ge_0.
  by rewrite Rsqr_sqrt /Rsqr-?bpow_plus -/k ; lra.
have haux: ( 0 < sqrt (1 + u/2)) by apply/sqrt_lt_R0; lra.
have hauxi:= (Rinv_0_lt_compat _  haux).
have hgxsx: pow ks / (sqrt (1 + u/2)) < sqrt x.
  apply/Rsqr_incrst_0; try apply/sqrt_ge_0; try nra.
  rewrite Rsqr_sqrt   ?Rsqr_mult ?Rsqr_inv; try  lra.
  by rewrite Rsqr_sqrt /Rsqr -?bpow_plus -/k; try lra.
have haux1: pow ks * (1 - u/2) < pow ks / sqrt (1 + u / 2).
  apply/Rmult_lt_compat_l; first by apply/bpow_gt_0.
  apply/Rsqr_incrst_0; try lra.
  rewrite Rsqr_inv ?Rsqr_sqrt ; try lra.
  apply/(Rmult_lt_reg_r (1 + u / 2)); try lra.
  rewrite Rinv_l; try lra.
  rewrite /Rsqr ; field_simplify.
  suff:  u ^ 3 - 2 * u ^ 2 - 4 * u  < 0 by lra.
  suff:  u ^ 2<   4 by nra.
  nra.
have Fpowks: format (pow ks) by  apply/generic_format_bpow;  rewrite /fexp; lia.
have pbp:  pred two fexp (pow ks) = pow ks * (1 - u ).
  by rewrite pred_bpow /fexp bpow_plus -/u; ring.
have hZ: round two fexp Zceil (sqrt x) = pow ks.
  apply/round_UP_eq=>//; rewrite pbp; lra.
have hF: round two fexp Zfloor (sqrt x) = pow ks * ( 1 - u).
  apply/round_DN_eq.
    by rewrite -pbp; apply/generic_format_pred.
   rewrite -pbp succ_pred //; lra. 
have h37_1: RN (sqrt x) = pow ks by rewrite round_N_eq_UP //;lra.
lra.
Qed.

Notation RxDW := (SQRTDW2_1 xh xl + (SQRTDW2_2 xh xl)). 

Theorem Th_3_11:
  Rabs (RxDW - sqrt x) <= 
    (Rsqr u)* (25/8 + nu/(1-nu*Rsqr u) + 25/8 * (nu * Rsqr u) / (1 - nu*Rsqr u)) * sqrt x.
Proof.
have xhlpos:=(DWpos DWx xhpos).
have hxaux: ( 0 < sqrt x) by apply/sqrt_lt_R0; lra.
have hxauxi:= (Rinv_0_lt_compat _  hxaux).
have hxhlaux: ( 0 < sqrt (xh + xl)) by apply/sqrt_lt_R0; lra.
have hxhlauxi:= (Rinv_0_lt_compat _  hxhlaux).
have upos : 0 < u    by  apply/bpow_gt_0.
have nuupos: 0 <= nu .
  apply/( Rmult_le_reg_r (Rsqr u)).
    rewrite /Rsqr; nra.
  apply/( Rmult_le_reg_r x); try  nra.
  rewrite !Rmult_0_l.
  by apply/(Rle_trans _  (Rabs (xh + xl - x)))=>//; apply/Rabs_pos.
have->: (RxDW -sqrt x )  = (RxDW - (sqrt (xh + xl)) )+ ((sqrt (xh + xl) - sqrt x)) 
   by ring.
apply/(Rle_trans _ _  _  (Rabs_triang _ _)).
have h36:=Th_3_10_aux.
have->:  u² * (25 / 8 + nu /(1 - nu * u²) + 25 / 8 * (nu * u²) / (1 - nu * u²)) * sqrt x =
  u² * (25 / 8 +  25 / 8 * (nu * u²) / (1 - nu * u²)) * sqrt x +  
  nu * u² / (1 - nu * u²) * sqrt x.
  by field; lra.
apply/Rplus_le_compat=>//.
have h:  Rabs (RxDW - sqrt (xh + xl)) <= 25 / 8 * u ^ 2 * sqrt (xh + xl).
  apply/(Rmult_le_reg_r (/(sqrt (xh + xl))))=>//.
  rewrite Rmult_assoc Rinv_r; last lra.
  case:(SQRTDW_bounds DWx  xhpos)=>//[ h1 [h2 h3]].
  move:(h1).
  rewrite /rel_err_sqrtDW2 Rabs_mult (Rabs_pos_eq (/_)); lra.
apply/(Rle_trans _ (25 / 8 * u ^ 2 * sqrt (xh + xl)))=>//.
have->:  u² * (25 / 8 + 25 / 8 * (nu * u²) / (1 - nu * u²)) *sqrt x = 
        25/8 * u^2 * ((1 + (nu*(Rsqr u)) /(1-nu*Rsqr u)) * sqrt x).
  rewrite /Rsqr; field.
  move:nu1 ; rewrite /Rsqr; lra.
apply/Rmult_le_compat_l; first  nra.
move/Rabs_le_inv: x_approx=>[h1 h2].
have {} h2: xh + xl <= (1 +  nu * u² )* x by lra.
apply/(Rle_trans _ (sqrt ((1 + nu * Rsqr u) *  x))).
  apply/sqrt_le_1_alt; try lra.
rewrite sqrt_mult; try nra.
apply/Rmult_le_compat_r; first lra.
apply/(Rle_trans _ (1 + nu * Rsqr u)).
  case: nuupos=>hh; first last.
    rewrite -hh Rmult_0_l Rplus_0_r sqrt_1; lra.
  apply/Rlt_le/sqrt_less_alt.
  have : 0 < Rsqr u by rewrite /Rsqr; nra.
  nra.
suff:  nu * u² <=  nu * u² / (1 - nu * u²)by lra.
rewrite -[X in X <= _]Rmult_1_r.
apply/Rmult_le_compat_l; try  nra.
apply (Rmult_le_reg_r (1 - nu* Rsqr u)); first lra.
rewrite Rmult_1_l Rinv_l.
  rewrite /Rsqr ; nra.
lra.
Qed.
End  Th_3_10.

Section exampleDW_sqrt.

Variable X: R.

Notation xh := (1 + 6*u).
Notation xl :=  (u -6* u^2).


Notation sh  := (RN (sqrt xh)).
Notation ro1    := (RN (xh - (sh  )^2)).
Notation ro2  := (RN (xl + (ro1 ))).
Notation sl  := (RN ((ro2 )/ (2*(sh)))).

Hypothesis ZNE : choice = fun n => negb (Z.even n).

Notation SQRTDW2_1 xh xl := (fst (SQRTDWToDW xh xl)).
Notation SQRTDW2_2 xh xl := (snd (SQRTDWToDW xh xl)).

Definition DW_sqrt_res a b := SQRTDW2_1 a b +  SQRTDW2_2 a b.

Lemma exDW_sqrt (Hp4 : (4<= p)%Z): 
  DW_sqrt_res xh xl = 1 + 7/2 *u -6*Rsqr u.
Proof.
have upos:= (upos p).
rewrite -/u in upos.
have ule16: u <= /16 by change (pow (-p) <= pow (-4)); apply/bpow_le; lia.
rewrite / DW_sqrt_res /SQRTDWToDW.
have  shb: 1+ 2*u < sqrt xh < 1+3*u.
  rewrite -(Rabs_pos_eq (sqrt _)); last by apply/sqrt_ge_0.
  rewrite -(Rabs_pos_eq (1 + 3*u)); last by lra.
  rewrite -(Rabs_pos_eq (1 + 2*u)); last by lra.
  split;apply/  Rsqr_lt_abs_0; rewrite !Rsqr_sqrt; try  lra;
  rewrite /Rsqr;ring_simplify; nra.
have s1: 1+2*u = succ two fexp 1.
  by rewrite succ_FLX_1 bpow_plus  -/u pow1; ring.
have ss1: 1+4*u = succ two fexp (1+ 2*u).
  rewrite succ_eq_pos; try lra.
  rewrite ulp_neq_0 /cexp  ?bpow_plus -/u; try lra.
  rewrite (mag_unique_pos _ _ 1) ?pow0 ?bpow_plus pow1; try split;lra.
have xhE: xh = succ two fexp (1+ 4*u).
  rewrite succ_eq_pos; try lra.
  rewrite ulp_neq_0 /cexp  ?bpow_plus -/u; try lra.
  rewrite (mag_unique_pos _ _ 1) ?pow0 ?bpow_plus pow1; try split;lra.
have shE: sh = 1+2*u.
  have rZ: round two fexp Zfloor (sqrt xh) = 1 + 2 * u.
    apply/round_DN_eq.
      rewrite s1;apply/generic_format_succ; rewrite -pow0; apply/format_bpow =>//.
    rewrite succ_eq_pos; try lra.
    split; try lra.
    rewrite ulp_neq_0 /cexp /fexp; last lra.
    rewrite (mag_unique_pos _ _ 1) ?pow0 ?bpow_plus pow1 -/u;try lra; 
     split; try lra.
  have rC: round two fexp Zceil (sqrt xh) = 1 + 4 * u.
    apply/round_UP_eq.
      rewrite ss1 s1; apply/generic_format_succ/generic_format_succ.
      by rewrite -pow0 ; apply/format_bpow.
    rewrite {1}ss1 pred_succ; try lra.
    by rewrite s1; apply/generic_format_succ; rewrite -pow0; apply/format_bpow.
  rewrite -rZ; apply/round_N_eq_DN; rewrite rZ rC; lra.
have ro1E: ro1 = xh - (Rsqr sh) .
  rewrite round_generic // -Rsqr_pow2 //.
  apply/ sqrt_error_FLX_N =>//.
  rewrite xhE; apply/generic_format_succ; rewrite ss1; apply/generic_format_succ.
  rewrite s1;  apply/generic_format_succ; rewrite -pow0; apply/format_bpow; lia.
have ro1E': ro1 = 2*u -4*Rsqr u.
  by rewrite ro1E shE /Rsqr; ring.
have F3: format 3.
  apply:(@ FLX_format_Rabs_Fnum _  _ (Float two 3 0)).
  rewrite /F2R //=; ring.
  rewrite /=; apply:(Rlt_le_trans _ (pow 2)).
    rewrite Rabs_pos_eq; last lra.
    by change (3 < 4); lra.
  by apply:bpow_le; lia.
have s3m4: succ two fexp (3 - 4*u) = 3.
  rewrite succ_eq_pos ?ulp_neq_0 /cexp ?bpow_plus;  try lra.
  rewrite (mag_unique_pos _ _ 2) ?pow1 -/u pow2;  lra.
have p3: pred two fexp 3 = 3 -4*u.
  rewrite pred_eq_pos /  Ulp.pred_pos; try lra.
  rewrite 
   (mag_unique_pos _ _ 2)  ?Req_bool_false ?pow1  ?pow2 1?bpow_plus ?pow1 -/u; 
   try lra.
  rewrite ulp_neq_0 /cexp ?bpow_plus -/u; try lra.
  rewrite (mag_unique_pos _ _ 2) ?pow1 pow2; lra.
have F3m4u: format (3 -4*u).
  by rewrite -p3; apply/generic_format_pred.
have pp3: pred two fexp (3 -4*u) = 3 -8*u.
  rewrite pred_eq_pos /  Ulp.pred_pos; try lra.
rewrite 
(mag_unique_pos _ _ 2)  ?Req_bool_false ?pow1  ?pow2 1?bpow_plus ?pow1 -/u; 
   try lra.
  rewrite ulp_neq_0 /cexp ?bpow_plus -/u; try lra.
  by rewrite (mag_unique_pos _ _ 2) ?pow1 pow2; lra.
have F3m8u: format (3 -8*u).
  by rewrite -pp3; apply/generic_format_pred.
have ppp3: pred two fexp (3 -8*u) = 3 -12*u.
  rewrite pred_eq_pos /  Ulp.pred_pos; try lra.
rewrite 
(mag_unique_pos _ _ 2)  ?Req_bool_false ?pow1  ?pow2 1?bpow_plus ?pow1 -/u; 
   try lra.
  rewrite ulp_neq_0 /cexp ?bpow_plus -/u; try lra.
  by rewrite (mag_unique_pos _ _ 2) ?pow1 pow2; lra.
have F3m12u: format (3 -12*u).
  by rewrite -ppp3; apply/generic_format_pred.
have ro2E: ro2= 3*u -8*Rsqr u.
  rewrite ro1E'.
  have->:  (xl + (2 * u - 4 * u²))= (3 - 10*u)*u by rewrite /Rsqr; ring.
  rewrite round_bpow -/u /Rsqr -Rmult_assoc -Rmult_minus_distr_r.
  congr (_*_).
  have rZ: round two fexp Zfloor (3 -10*u) = 3-12*u.
    apply/round_DN_eq=>//.
    by rewrite -{2}ppp3 succ_pred //; lra.
  have rC: round two fexp Zceil (3 -10*u) = 3-8*u.
    apply/round_UP_eq=>//.
    by rewrite ppp3; lra.
  rewrite round_N_middle; first last.
    by rewrite  rZ rC; lra.
  rewrite ZNE /= /scaled_mantissa/cexp (mag_unique_pos _ _ 2) 
     ?pow1?bpow_plus ?pow2; try lra.
  rewrite /fexp.
  rewrite Z.opp_sub_distr /u.
  ring_simplify ((3 - 10 * pow (- p)) * pow (- (2) + p)).
  rewrite Rmult_assoc  -bpow_plus.
  ring_simplify(- p + (- (2) + p))%Z.
  have->:(-10 * pow (-2) + 3 * pow (- (2) + p)) = 3*(pow (p -2)) - 10 /4.
    have->: pow (-2) = /4 by [].
    have ->: ((- (2) + p) = p-2)%Z by ring.
    by rewrite /Rdiv; ring.
  have ->:Zfloor (3 * pow (p - 2) - 10 / 4) =  (3 * (2 ^ (p -2)) - 3)%Z.
    apply/Zfloor_imp.
    rewrite plus_IZR minus_IZR mult_IZR (IZR_Zpower two); last lia.
    lra.
  have ->: (3 * 2 ^ (p - 2) - 3   = ( 1+ 3 * (2 * 2 ^ (p-3)) -4) )%Z.
    rewrite -Z.pow_succ_r  -?Z.add_1_r; last lia.
    ring_simplify(p - 3 + 1) %Z; ring.
  have->: (1 + 3 * (2 * 2 ^ (p - 3)) - 4 = 1 + 2 * (3 *  2 ^ (p - 3) -2))%Z 
    by ring.
  by rewrite Z.even_add_mul_2 -/u .
have slE: sl  = (3 - 12 *u) * (u*/2).
  have  -> : (ro2 )/ (2*(sh)) = (3/2 * u  - 7 *u^2 * /(1 + 2*u)).
    rewrite ro2E shE.  
    apply/(Rmult_eq_reg_r (2*(1 +  2*u))); try lra.
    rewrite /Rdiv Rmult_assoc Rinv_l ?Rmult_1_r; try lra.
    rewrite /Rsqr; field; lra.
  have->: (3 / 2 * u - 7 * u ^ 2 * / (1 + 2 * u)) =
        (3 - 14 * u /(1+2*u)) * (u/2) by field; lra.
  have -> : u/2 = pow ( -p -1).
    by rewrite bpow_plus powm1 -/u; field.
  rewrite round_bpow bpow_plus -/u powm1.
  apply/Rmult_eq_compat_r.
  have p3m16u : pred two fexp ( (3 - 12 * u)) = 3 -16*u.
    rewrite pred_eq_pos /  Ulp.pred_pos; try lra.
    rewrite (mag_unique_pos _ _ 2)  ?Req_bool_false ?pow1  ?pow2 1?bpow_plus ?pow1 -/u; 
   try lra.
    rewrite ulp_neq_0 /cexp ?bpow_plus -/u; try lra.
    by rewrite (mag_unique_pos _ _ 2) ?pow1 pow2; lra.
  have rZ: round  two fexp Zfloor  (3 - 14 * u / (1 + 2 * u)) = 3 - 16 * u.
    apply/round_DN_eq=>//.
      by rewrite -p3m16u; apply/generic_format_pred.
    rewrite -{2} p3m16u succ_pred //.
    split.
      suff:  14 * u * / (1 + 2 * u)  <= (16*u) * 1 by lra.
      apply/Rmult_le_compat; try lra.
        by apply/Rlt_le/Rinv_0_lt_compat; lra.
      by rewrite -{2}Rinv_1; apply/Rinv_le; lra.
    suff : 12 * u < (14 * u )  / (1 + 2 * u) by lra.
    apply/(Rmult_lt_reg_r (1+2*u)); try  lra.
    rewrite  (Rmult_assoc (14*u))  Rinv_l ?Rmult_1_r; try lra.
    nra.
  have rC: round  two fexp Zceil  (3 - 14 * u / (1 + 2 * u)) = 3 - 12 * u.
    apply/round_UP_eq=>//.
    rewrite   p3m16u.
    split.
      apply/Rplus_lt_compat_l/Ropp_lt_contravar.
      apply/(Rmult_lt_reg_r (1 + 2*u)) ; try lra.
      rewrite  2!Rmult_assoc    Rinv_l  ?Rmult_1_r; try lra.
      nra.
    apply/Rplus_le_compat_l/Ropp_le_contravar.
    apply/(Rmult_le_reg_r (1 + 2*u)) ; try lra.
    rewrite  (Rmult_assoc (14*u))  Rinv_l  ?Rmult_1_r; try lra.
    nra.
  rewrite -rC; apply/round_N_eq_UP.
  rewrite rZ rC; try lra.
  have->:   (3 - 16 * u + (3 - 12 * u)) / 2 = 3 - 14*u by field.
  apply/Rplus_lt_compat_l/Ropp_lt_contravar.
  rewrite -{2}(Rmult_1_r (14*u)).
  rewrite /Rdiv; apply/Rmult_lt_compat_l; try lra.
  rewrite -{2}Rinv_1; apply/Rinv_lt; lra.
suff F2SC: (Fast2Sum_correct  p choice sh sl).
  rewrite F2SC.
  ring_simplify.
  rewrite slE shE /Rsqr; lra.
apply/F2Sum_correct_abs; try apply/generic_format_round; try (rewrite /=; lia).
  by apply/RN_sym.
rewrite slE shE !Rabs_pos_eq ; nra.
Qed.

Notation x := (xh + xl).
(* Variable O: R. *)

Notation O := 332.


Lemma  exDW_sqrt_rel   (Hp4: (4<= p)%Z): 
  Rabs ((sqrt x -  DW_sqrt_res xh xl)/ sqrt x)  <=
       25/8* (u^2) - 343/8 * (u^3) +O *(u ^4).
Proof.
have upos: 0 < u by apply/(upos p); rewrite -/u.
have ult1: u <1 by rewrite /u -pow0 ; apply/bpow_lt; lia.
have u16: u <= /16 by change (pow (-p) <= pow (-4)); apply/bpow_le; lia.
have xpo: 0 < x by nra.
have sqrtx_pos: 0 < sqrt x by apply/sqrt_lt_R0 ; nra.
have isqrtx_pos: 0 < /(sqrt x) by apply/Rinv_0_lt_compat.
rewrite Rabs_mult (Rabs_pos_eq (/_)); last lra.
rewrite -Rabs_Ropp Ropp_minus_distr.
rewrite Rabs_pos_eq; last first.
  suff:  sqrt x <= (1 + 7 / 2 * u - 6 * u²) by rewrite  exDW_sqrt //; lra.
  rewrite - (Rabs_pos_eq (sqrt x)); last by apply/sqrt_ge_0.
  rewrite -(Rabs_pos_eq  (1 + 7 / 2 * u - 6 * u²)); last by rewrite/Rsqr; nra.
  apply/Rsqr_le_abs_0.
  rewrite Rsqr_sqrt; last lra.
  suff: 0<=  - x +(1 + 7 / 2 * u - 6 * u²)² by lra.
  rewrite /Rsqr; field_simplify.
  suff: 0<= (144 * u ^ 2 - 168 * u  + 25 ) by nra.
  nra.
rewrite Rmult_minus_distr_r Rinv_r; last lra.
have h1: 1 < sqrt x < x.
  rewrite -{1}sqrt_1;split; try apply/sqrt_lt_1; try lra; try nra.
  apply/sqrt_less; nra.
have h2: /sqrt x < 1.
  rewrite -{2}Rinv_1; apply/Rinv_lt; lra.
rewrite exDW_sqrt //.
suff:   (1 + 7 / 2 * u - 6 * u²) * / sqrt x <= 1 + 
  25 / 8 * u ^ 2 - 343 / 8 * u ^ 3 + O* u ^ 4 by lra.
apply/(Rmult_le_reg_r (sqrt x)); first  lra.
rewrite Rmult_assoc Rinv_l ?Rmult_1_r; last lra.
rewrite -(Rabs_pos_eq (_ -_)); last by rewrite /Rsqr ; nra.
rewrite -(Rabs_pos_eq (_ * (sqrt _))); last first.
  apply/Rmult_le_pos; try apply/sqrt_ge_0.
  suff:   0 <= 1 + 25 / 8 * u ^ 2 - 343 / 8 * u ^ 3.
    nra.
  nra.
apply/Rsqr_le_abs_0.
rewrite Rsqr_mult Rsqr_sqrt; try lra.
rewrite /Rsqr ; field_simplify.
suff : 0 <=   -384 * u ^ 10 * O ^ 2 + 448 * u ^ 9 * O ^ 2 + 32928 * u ^ 9 * O +
  64 * u ^ 8 * O ^ 2 - 40816 * u ^ 8 * O - 705894 * u ^ 8 - 
  2688 * u ^ 7 * O + 926443 * u ^ 7 - 368 * u ^ 6 * O - 
  6151 * u ^ 6 + 896 * u ^ 5 * O + 20153 * u ^ 5 + 128 * u ^ 4 * O -
  40191 * u ^ 4 - 2304 * u^4 by lra.
ring_simplify.
nra.
Qed.
End exampleDW_sqrt.

Section Overflow.

Variables xh xl:R.
Hypothesis DWx : double_word  xh xl.
Hypothesis xhpos: 0 < xh.

Variable emin emax:Z.

Hypothesis emin_max_p:emin = (1 - emax)%Z /\ (emin < - p - 1)%Z /\ (1 < p)%Z.
Hypothesis Hp5: ( 5 <= p)%Z.
Notation Omega := (square_root_aux.Omega p emax).

Hypothesis no_ovrflow_xh: xh <= Omega.


Notation sh xh := (RN (sqrt xh)).
Notation ro1  xh  := (RN (xh - (sh xh )^2)).
Notation ro2 xh xl := (RN (xl + (ro1 xh))).
Notation sl xh xl := (RN ((ro2 xh xl)/ (2*(sh xh)))).

Lemma Omega4 : 4 < Omega.
Proof.
rewrite /Omega  -(Rmult_1_r 4).
apply/Rmult_lt_compat; try  lra.
  by rewrite -pow2; apply/bpow_lt; lia.
rewrite bpow_plus pow1.
suff:pow (-p) </2 by lra.
rewrite -powm1; apply/bpow_lt ; lia.
Qed.

Fact sqrtO: sqrt Omega < Omega/2.
Proof.
have O4:= Omega4.
suff : 2 * sqrt Omega < Omega by lra.
rewrite -{2}(Rsqr_sqrt Omega); try lra.
rewrite /Rsqrt; apply/Rmult_lt_compat_r.
  apply/sqrt_lt_R0; lra.
apply/Rsqr_incrst_0; try apply/sqrt_ge_0;try lra.
by rewrite Rsqr_sqrt /Rsqr; try lra.
Qed.

Lemma no_overf_sh :  sh xh <= Omega.
Proof.
apply/(Rle_trans _ (RN (sqrt Omega))).
  by apply/round_le/sqrt_le_1_alt.
apply/(Rle_trans _ (RN (Omega/2))).
apply/round_le/Rlt_le/sqrtO.
rewrite /Rdiv  -powm1 round_bpow round_generic.
  rewrite powm1; lra.
by apply/generic_format_FLX_FLT/(@square_root_aux.FpOmega p emin).
Qed.

Lemma no_overf_ro1_aux1:  Rabs (ro1 xh)  <= (2*u + u^2) *xh.
Proof.
have upos:= (upos p).
rewrite -/u in upos.
case:DWx =>[[Fxh Fxl] xE].
rewrite round_generic; last first.
  rewrite -Rsqr_pow2;apply/ sqrt_error_FLX_N=>//.
have->:  (xh - sh xh ^ 2) = (sqrt xh -sh xh) * (sqrt xh + sh xh).
  ring_simplify.
  rewrite  -(Rsqr_pow2 (sqrt xh)) Rsqr_sqrt; try lra;  ring.
rewrite Rabs_mult.
rewrite -Rabs_Ropp Ropp_minus_distr.
have he: Rabs (sh xh  -sqrt xh) <= u * (sqrt xh).
  have->:  u * (sqrt xh) = / 2 * pow (- p+ 1) * Rabs (sqrt xh).
    rewrite Rabs_pos_eq; last  apply/sqrt_ge_0.
    rewrite bpow_plus pow1 -/u; lra.
  by apply/relative_error_N_FLX.
apply/(Rle_trans _ ( u* (sqrt xh) * (sqrt xh + (sqrt xh * (1 + u))))).
  apply/Rmult_le_compat=>//; try apply/Rabs_pos.
  rewrite Rabs_pos_eq; last first.
    apply/Rplus_le_le_0_compat; try apply/sqrt_ge_0.
    have->: 0 = RN 0 by rewrite round_0.
    by apply/round_le/sqrt_ge_0.
  apply/Rplus_le_compat_l.
  by move/Rabs_le_inv: he; lra.
ring_simplify; rewrite -(Rsqr_pow2 (sqrt xh)) Rsqr_sqrt //;lra.
Qed.


Lemma no_overf_ro1_aux:  Rabs (ro1 xh)  <= 65/1024* Omega.
Proof.
have upos:= (upos p).
rewrite -/u in upos.
have u32:  u <=/32.
  by change (pow (-p) <= pow (-5));apply/bpow_le; lia.
apply/(Rle_trans _ ((2*u + u^2) *xh)); first by apply/ no_overf_ro1_aux1.
by apply/Rmult_le_compat;nra.
Qed.

Lemma no_overf_ro1:  Rabs (ro1 xh)  <= Omega.
Proof.
apply/(Rle_trans _ (65/1024*Omega)); try lra.
by apply/no_overf_ro1_aux.
Qed.


Fact a_1pu a :  RN (Rabs a)  <= (1+u)* (Rabs a).
Proof.
have:   Rabs (RN a - a) <=
  / 2 * pow (- p + 1) * Rabs a.
  by apply/relative_error_N_FLX.
rewrite bpow_plus -/u pow1.
move/Rabs_le_inv.
case : (Rle_lt_dec 0 a)=>a0.
  rewrite Rabs_pos_eq //; lra.
rewrite -Rabs_Ropp Rabs_pos_eq; try lra.
suff : - RN (- a) =  RN a by lra.
by rewrite round_N_opp Ropp_involutive /round  /Znearest -choice_sym.
Qed.

Lemma no_overf_ro2_aux_1:  Rabs (ro2 xh xl)  <= (3*u + 4*u^2 + u^3)*xh.
Proof.
have upos:= (upos p).
rewrite -/u in upos.
have xh_ge_0: 0 <= xh by lra.
case:DWx =>[[Fxh Fxl] xE].
have haux : Rabs (xl + ro1 xh) <= (3*u + u^2) * xh.
  apply/(Rle_trans _ _  _  (Rabs_triang _ _)).
  have ->:  (3 * u + u ^ 2) * xh = u*xh + (2*u + u^2)*xh by ring.
  apply/Rplus_le_compat.
    have->: xl = xh + xl - RN(xh + xl) by rewrite -xE; ring. 
    rewrite -Rabs_Ropp Ropp_minus_distr.
    have->: u*xh = /2 * pow(-p +1)* Rabs(RN (xh + xl)).
      rewrite -xE Rabs_pos_eq // bpow_plus pow1 -/u; field.
apply/relative_error_N_FLX_round=>//.
  by apply/no_overf_ro1_aux1.
rewrite -RN_abs //.
apply/(Rle_trans _ ((1+u)* (Rabs (xl + ro1 xh)))).
apply/a_1pu.
apply/(Rle_trans _ ((1+u)*((3 * u + u ^ 2) * xh))).
  apply/Rmult_le_compat_l=>//; lra.
rewrite -Rmult_assoc.
apply/Rmult_le_compat; nra.
Qed.


Lemma no_overf_ro2_aux:  Rabs (ro2 xh xl)  <= 3201/32768* Omega.
Proof.
have upos:= (upos p).
rewrite -/u in upos.
apply/(Rle_trans _ ((3*u + 4*u^2 + u^3)*xh)).
  by apply/ no_overf_ro2_aux_1.
apply/Rmult_le_compat; try nra.
suff :u <= /32 by nra.
change (pow (-p) <= pow (-5)); apply/bpow_le; lia.
Qed.



Lemma no_overf_ro2:  Rabs (ro2 xh xl)  <= Omega.
Proof.
apply/(Rle_trans _  (3201/32768*Omega)); try lra.
by apply/no_overf_ro2_aux.
Qed.

Lemma no_overf_sl_aux2: Rabs ( ro2 xh xl /(2 * sh xh)) <= 
                         (3*u + 4*u^2 + u^3)/(2 -2*u) * (sqrt xh).
Proof.
have upos:= (upos p);rewrite -/u in upos.
have ult1: u < 1 by change (pow(-p) < pow 0); apply/bpow_lt; lia.

have xhge_0: 0 <= xh by lra.
rewrite Rabs_mult.
have sqrtpos: 0 < sqrt xh by  apply/sqrt_lt_R0.
have shpos: 0 < sh xh.
  by apply/gt_0_round_gt_0_FLX.
have isqrtpos: 0 < / sqrt xh by apply/Rinv_0_lt_compat.
have ishpos : 0 < /sh xh  by apply/Rinv_0_lt_compat. 
have->: (3 * u + 4 * u ^ 2 + u ^ 3) / (2 - 2 * u) * sqrt xh =  ((3 * u + 4 * u ^ 2 + u ^ 3)*xh) /(2* (1-u)*sqrt xh).
  rewrite -{2}(Rsqr_sqrt xh) // /Rsqr;field; lra.
apply/Rmult_le_compat; try apply/Rabs_pos.
  by apply/no_overf_ro2_aux_1.
rewrite Rabs_pos_eq ; last by apply/Rlt_le/Rinv_0_lt_compat; lra.
apply/Rinv_le; try nra.
suff :  - (u * sqrt xh ) <= sqrt xh - sh xh <= u * sqrt xh by lra.
apply/Rabs_le_inv.
rewrite -Rabs_Ropp Ropp_minus_distr.
have->:  u * sqrt xh = /2 * pow (- p + 1) * Rabs (sqrt xh).
  by rewrite bpow_plus -/u pow1  Rabs_pos_eq ; try apply/sqrt_ge_0; field.
by apply/relative_error_N_FLX.
Qed.

Lemma no_overf_sl_aux1: Rabs ( ro2 xh xl /(2 * sh xh)) <= 3201/63488 * (sqrt xh).
Proof.
have upos:= (upos p);rewrite -/u in upos.
have ult1: u < 1 by change (pow(-p) < pow 0); apply/bpow_lt; lia.
have u32 : u <= /32 by change (pow (-p) <= pow (-5)); apply/bpow_le; lia.
apply/(Rle_trans  _ _ _  no_overf_sl_aux2).
apply/Rmult_le_compat_r; try apply/sqrt_ge_0.
have->: 3201 / 63488 = 3201 /32768 * (32768/ 63488) by field.
apply/Rmult_le_compat; try nra.
  apply/Rlt_le/Rinv_0_lt_compat; lra.
have ->: 32768 / 63488 = / (63488/ 32768) by field.
apply/Rinv_le; lra.
Qed.



Lemma no_overf_sl_aux: Rabs ( ro2 xh xl /(2 * sh xh)) < Omega.

Proof.
apply/(Rle_lt_trans _ _ _  no_overf_sl_aux1).
apply/(Rlt_trans _ (sqrt xh)).
  suff : 0 < sqrt xh by lra.
  by apply/sqrt_lt_R0.
apply/(Rle_lt_trans _ (sqrt Omega)).
  apply/sqrt_le_1; lra.
apply/(Rlt_trans _  (Omega/2)); try apply/sqrtO; lra.
Qed.

Lemma no_overf_sl :  Rabs (sl  xh xl) < Omega.
Proof.
have upos:= (upos p);rewrite -/u in upos.
have u32 : u <= /32 by change (pow (-p) <= pow (-5)); apply/bpow_le; lia.

apply/(Rle_lt_trans _ ((1+u)* (Rabs (ro2 xh xl /(2 * sh xh))))).
  by rewrite -RN_abs //;apply/ a_1pu.
apply/(Rle_lt_trans _ ((1+u) * (3201/63488 * (sqrt xh)))).
  apply/Rmult_le_compat_l; try lra.
  by apply/no_overf_sl_aux1.
apply/(Rlt_trans _ (sqrt xh)).
  rewrite  -Rmult_assoc -{2}(Rmult_1_l (sqrt xh)).
  apply/Rmult_lt_compat_r; try nra.
  by apply/sqrt_lt_R0.
apply/(Rle_lt_trans _ (sqrt Omega)).
  apply/sqrt_le_1; lra.
apply/(Rlt_trans _  (Omega/2)); try apply/sqrtO; lra.
Qed.

Lemma no_overflow_res: sh xh  + sl  xh  xl< Omega.
Proof.
have upos:= (upos p);rewrite -/u in upos.
have u32 : u <= /32 by change (pow (-p) <= pow (-5)); apply/bpow_le; lia.
apply/(Rle_lt_trans _ (sqrt xh * (1 + u +  (3*u + 4*u^2 + u^3)/(2 -2*u)* (1+u)))).
  rewrite Rmult_plus_distr_l.
  apply/Rplus_le_compat.
    rewrite Rmult_comm.
    rewrite -(Rabs_pos_eq (sqrt xh)); last by apply/sqrt_ge_0.
    by apply/a_1pu.
  apply/ (Rle_trans _ ((1 + u) * Rabs (ro2 xh xl / (2 * sh xh)))).
     move :  (a_1pu ( ro2 xh xl /(2 * sh xh))).
    rewrite RN_abs //.
    move/Rabs_le_inv; lra.
  rewrite Rmult_comm -Rmult_assoc; apply/Rmult_le_compat_r; first lra.
  by rewrite (Rmult_comm  (sqrt xh)); apply/no_overf_sl_aux2.
have->:  sqrt xh * (1 + u + (3 * u + 4 * u ^ 2 + u ^ 3) / (2 - 2 * u) * (1 + u)) =
  (2 + 3*u + 5* u^2 + 5* u^3 + u^4)/(2 -2*u)* sqrt xh.
  field; lra.
have haux: 0 < (2 -2*u) by lra.
have ihaux: 0 < /(2 -2*u) by apply/Rinv_0_lt_compat.
apply/(Rle_lt_trans _ (2200737/2031616 * sqrt Omega)).
  apply/Rmult_le_compat; try apply/sqrt_ge_0.
  + by rewrite /Rdiv; apply/Rmult_le_pos;  nra.
  + apply/(Rmult_le_reg_r (2 - 2*u)); try lra.
    rewrite Rmult_assoc Rinv_l ?Rmult_1_r; try nra.
    apply/(Rle_trans _ ( 2 + 3 * /32 + 5 * (/32) ^ 2 + 5 *( /32) ^ 3 + (/32) ^ 4)).
      apply/Rplus_le_compat;try nra.
      apply/pow_incr; lra.
    nra.
  apply/sqrt_le_1; lra.
apply/(Rlt_trans _   (2200737 / 2031616 * (Omega/2))); try lra.
apply/Rmult_lt_compat_l; try lra.
apply/sqrtO; lra.
Qed.

Fact shpsl_pos: 0 <= sh xh + sl xh xl.
Proof.
suff : Rabs (sl xh xl) <= Rabs (sh xh).
  rewrite (Rabs_pos_eq (sh xh)).
    move/Rabs_le_inv; lra.
  have->: 0 = RN 0 by rewrite round_0.
  by apply/round_le/sqrt_ge_0.
have shpos: 0 < sh xh.
  by apply/gt_0_round_gt_0_FLX/sqrt_lt_R0.
rewrite -!RN_abs //; apply/round_le.
rewrite (Rabs_pos_eq (sqrt _)); last apply/sqrt_ge_0.
rewrite Rabs_mult (Rabs_pos_eq (/_)); last first.
  apply/Rlt_le/Rinv_0_lt_compat; lra.
apply/(Rmult_le_reg_r (2*sh xh)); try lra.
rewrite Rmult_assoc Rinv_l ?Rmult_1_r; try lra.
apply/(Rle_trans _ (sqrt xh * 2 * (1-u)*(sqrt xh))); last first.
  rewrite (Rmult_comm (sqrt _) (2* _)).
  apply/Rmult_le_compat_r; try apply/sqrt_ge_0.
  rewrite (Rmult_comm (sqrt _)) Rmult_assoc.
  apply/Rmult_le_compat_l; try lra.
  suff: -(u * sqrt xh) <= sqrt xh - sh xh <= u* sqrt xh by lra.
  apply/Rabs_le_inv.
  rewrite -Rabs_Ropp Ropp_minus_distr.
  have ->:  u * sqrt xh =  / 2 * pow (- p + 1) * Rabs (sqrt xh).
    by rewrite bpow_plus -/u pow1 Rabs_pos_eq; try apply/sqrt_ge_0; field.
  by apply/relative_error_N_FLX.
have ->: sqrt xh * 2 * (1 - u) * sqrt xh = 2* (1-u) * xh.
  suff: sqrt xh * sqrt xh = xh by nra.
  by rewrite sqrt_sqrt; lra.
apply/(Rle_trans _ _ _  no_overf_ro2_aux_1).
apply/Rmult_le_compat_r; try lra.
have u32 : u <= /32 by change (pow (-p) <= pow (-5)); apply/bpow_le; lia.
have upos:= (upos p);rewrite -/u in upos.
 nra.
Qed.

End Overflow.

End DW_sqrt.

End Main.

Theorem Th_3_10'
     : forall p : Z,
       forall choice : Z -> bool,
       (forall x : Z, choice x = negb (choice (- (x + 1))%Z)) ->
       forall nu x xh xl : R,
       0 <= nu->
       (generic_format radix2 (FLX_exp p) xh /\
        generic_format radix2 (FLX_exp p) xl) /\
       xh = round radix2 (FLX_exp p) (Znearest choice) (xh + xl) ->
       Rabs (xh + xl - x) <= nu * (bpow radix2 (- p))² * x ->
       0 <= x ->
       0 <= xh ->
       (5 <= p)%Z ->
       2 * bpow radix2 (- p) * nu < 1 ->
       Rabs (SQRTDWToFP p choice xh xl - sqrt x) <=
       (/ 2 +
        bpow radix2 (- p) * (7 / 4 + nu / (1 - nu * (bpow radix2 (- p))²))) *
       ulp radix2 (FLX_exp p) (sqrt x).
Proof.
move=>p choice choice_sym nu  x xh xl nu0 dwxhl xhxlmx xpos  xhpos   p5 nu1.
have p0:     Prec_gt_0 p by rewrite /Prec_gt_0; lia.
case:(dwxhl)=>[[Fxh Fxl]] xE.
have hh:  nu * (bpow radix2 (- p))² <=  2 * bpow radix2 (- p) * nu.
  rewrite Rmult_comm.
  apply/Rmult_le_compat_r=>//.
  rewrite /Rsqr.
  apply/Rmult_le_compat_r.
    by apply/bpow_ge_0.
  rewrite -pow1.
  apply/bpow_le; lia.
case:xpos =>h0 ; last first.
  move: xhxlmx; rewrite -h0 sqrt_0 Rminus_0_r ulp_FLX_0 !Rmult_0_r.
  case; first by move:(Rabs_pos (xh +xl)); lra.
  move /Rabs_eq_R0=> xhxl0.
  have xh0 : xh = 0 by rewrite xE xhxl0 round_0.
  have xl0: xl = 0 by lra.
  rewrite /SQRTDWToFP xh0 xl0 !(sqrt_0, round_0, Rplus_0_l, pow_i); last lia.
  rewrite !Rminus_0_r !round_0 /Rdiv Rmult_0_l !round_0 Rabs_R0; lra.
case:xhpos =>xh0 ; last first.
  case:(Req_dec xl 0)=>[xl0|]; last first.
    move:xE; rewrite - xh0 Rplus_0_l round_generic //;  lra.
  move: xhxlmx.  
  rewrite xl0 -xh0 Rplus_0_l Rminus_0_l   Rabs_Ropp  Rabs_pos_eq; last lra.
  move=> h.
  have: False;move =>//.
  have h' : 1 <= nu * (bpow radix2 (- p))².
    apply/(Rmult_le_reg_r x) => //; lra.
  rewrite /Rsqr;lra.
apply/Th_3_10_bis=>//; try lia; lra.
Qed.


